/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cache_manager.cc,v 1.20 2008/09/26 11:58:22 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/cache_manager.h"

#include "libochusha/misc_utils.h"
#include "libochusha/repository.h"
#include "libochusha/uri_utils.h"

#include <algorithm>
#include <list>

#include <cassert>
#include <climits>
#include <cstdlib>

#ifdef HAVE_STDINT_H
#include <stdint.h>
#else
#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif
#endif


#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_UTIME_H
#include <utime.h>
#endif


namespace Ochusha
{

#define NUMBER_OF_CACHE_SLOTS	31

CacheManager::CacheManager(Repository &repository,
			   const char *temporary_dir, const char *tenured_dir,
			   uintmax_t capacity, unsigned int max_num_files)
  : repository(repository),
    cache_capacity(capacity * 1024 * 1024), current_cache_size(0),
    maximum_number_of_files(max_num_files), current_number_of_files(0),
    temporary_cache_table(101)
{
  if (temporary_dir != NULL)
    temporary_cache_directory = strdup(temporary_dir);
  else
    temporary_cache_directory = strdup("temporary_cache");

  if (tenured_dir != NULL)
    tenured_cache_directory = strdup(tenured_dir);
  else
    tenured_cache_directory = strdup("tenured_cache");

  update_cache_info();
}


#if 1	// for debugging
static void prepare_for_cleanup(RefCount *entry, void *data, const char *key);
#endif


CacheManager::~CacheManager()
{
#if 1	// for debugging
  temporary_cache_table.for_each(prepare_for_cleanup, NULL);
#endif
  free(temporary_cache_directory);
  free(tenured_cache_directory);
}


void
CacheManager::set_cache_capacity(uintmax_t capacity)
{
  cache_capacity = capacity * 1024 * 1024;
  ensure_limits();
}


void
CacheManager::set_maximum_number_of_files(unsigned int max_num_files)
{
  maximum_number_of_files = max_num_files;
}


void
CacheManager::set_limits(uintmax_t capacity, unsigned int max_num_files)
{
  cache_capacity = capacity * 1024 * 1024;
  maximum_number_of_files = max_num_files;
  ensure_limits();
}


class CacheEntry : public RefCount
{
private:
  char *filename;
  time_t mtime;
  uintmax_t size;
#if 1	// for debugging
  bool in_hashtable;
#endif

public:
  CacheEntry(const char *filename, time_t mtime, uintmax_t size)
#if 0
    : filename(strdup(filename)), mtime(mtime), size(size) {}
#else	// for debugging
    : filename(strdup(filename)), mtime(mtime), size(size), in_hashtable(false) {}
#endif
  ~CacheEntry()
  {
    assert(!in_hashtable);
    free(filename);
  }


  inline const char *get_filename() const { return filename; }
  inline time_t get_mtime() const { return mtime; }
  inline void set_mtime(time_t mt) { mtime = mt; }
  inline uintmax_t get_size() const { return size; }
#if 1	// for debugging
  inline void set_in_hashtable() { assert(!in_hashtable); in_hashtable = true; }
  inline void unset_in_hashtable() { assert(in_hashtable); in_hashtable = false; }
#endif
};


#if 1	// for debugging
static void
prepare_for_cleanup(RefCount *entry, void *data, const char *key)
{
  smart_ptr<CacheEntry> cache_entry = dynamic_cast<CacheEntry *>(entry);
  cache_entry->unset_in_hashtable();
}
#endif


struct CacheEntryCompare
{
  inline bool operator()(CacheEntry *x, CacheEntry *y)
  { return y->get_mtime() > x->get_mtime(); }
};


static unsigned int
string_hash(const char *str)
{
  int h = 0;
  while (*str != '\0')
    h = 31 * h + *str++;
  return h;
}


char *
CacheManager::get_cache_filename_from_url(const char *url) const
{
  smart_ptr<CacheEntry> entry
    = dynamic_cast<CacheEntry *>(temporary_cache_table.lookup(url));
  if (static_cast<CacheEntry *>(entry) == NULL
      && cache_capacity == 0 && maximum_number_of_files == 0)
    return get_tenured_file_path(url);

  char *enc = URI::encode_string(url);
  if (enc == NULL)
    return NULL;	// Out of Memory?

  unsigned int h = string_hash(enc);
  char filename[PATH_MAX];
  if (snprintf(filename, PATH_MAX,
	       "%s/%02u/%s", temporary_cache_directory,
	       h % NUMBER_OF_CACHE_SLOTS, enc) < PATH_MAX)
    {
      free(enc);
      return strdup(filename);
    }
  free(enc);
  return NULL;		// ե̾Ĺ
}


char *
CacheManager::get_url_from_cache_filename(const char *filename) const
{
  assert(filename != NULL);

  size_t len = strlen(temporary_cache_directory);
  assert(strncmp(filename, temporary_cache_directory, len) == 0);
  assert(strlen(filename + len) > 5);

  unsigned int h = string_hash(filename + len + 4);
  char slot[5];
  snprintf(slot, 5, "/%02u/", h % NUMBER_OF_CACHE_SLOTS);
  assert(strncmp(filename + len, slot, 4) == 0);

  return URI::decode_string(filename + len + 4);
}


char *
CacheManager::find_cache(const char *url) const
{
  char *filename = get_tenured_file_path(url);
  if (filename != NULL)
    {
      char *fullpath = repository.find_file(filename);
      free(filename);
      if (fullpath != NULL)
	return fullpath;
    }

  filename = get_cache_filename_from_url(url);
  if (filename == NULL)
    return NULL;

  char *fullpath = repository.find_file(filename);
  free(filename);
  return fullpath;
}


char *
CacheManager::find_cache_directory(const char *url) const
{
  char *dirname = get_tenured_file_path(url);
  if (dirname == NULL)
    return NULL;

  char *fullpath = repository.find_directory(dirname);
  free(dirname);

  return fullpath;
}


int
CacheManager::open_cache(const char *url, int flags) const
{
  char *filename = get_tenured_file_path(url);
  if (filename != NULL)
    {
      int fd = repository.open_file(filename, flags);
      free(filename);
      if (fd >= 0)
	return fd;
    }

  filename = get_cache_filename_from_url(url);
  if (filename == NULL)
    return -1;

  int fd = repository.open_file(filename, flags);
  free(filename);
  return fd;
}


bool
CacheManager::cache_exist(const char *url) const
{
  char *filename = get_tenured_file_path(url);
  if (filename != NULL)
    {
      bool result = repository.file_exist(filename);
      free(filename);
      if (result)
	return true;
    }

  filename = get_cache_filename_from_url(url);
  if (filename == NULL)
    return false;

  bool result = repository.file_exist(filename);
  free(filename);
  return result;
}


bool
CacheManager::is_tenured(const char *url) const
{
  char *filename = get_tenured_file_path(url);
  if (filename == NULL)
    return false;

  bool result = repository.file_exist(filename);
  free(filename);
  return result;
}


bool
CacheManager::cache(const char *url, Tmpfile *tmpfile)
{
  const struct stat *sb = tmpfile->stat();
  if (sb == NULL)
    return false;	// ɤˤʤʤ顼

  if (static_cast<uintmax_t>(sb->st_size) > cache_capacity)
    return false;	// Ǥ

  char *filename = get_cache_filename_from_url(url);
  if (filename == NULL)
    return false;	// Out of Memory?

  bool result = false;
  if (temporary_cache_table.lookup(url) != NULL)
    {
#if 1	// for debugging
      fprintf(stderr, "cache file for key=\"%s\" duplicated!\n", url);
#endif
    }
  else if (repository.rename(tmpfile->get_fullpath(), filename))
    {
      smart_ptr<CacheEntry> entry
      	= new CacheEntry(filename, sb->st_mtime, sb->st_size);

#if 1	// for debugging
      entry->set_in_hashtable();
#endif
      temporary_cache_table.insert(url, entry);

      list<CacheEntry *> new_entry;
      new_entry.push_back(entry);
      cache_entry_list.merge(new_entry, CacheEntryCompare());

      current_cache_size += sb->st_size;
      current_number_of_files++;

      ensure_limits();
      result = true;
    }

  if (!result)
    repository.unlink_file(tmpfile->get_fullpath());

  free(filename);
  return result;
}


bool
CacheManager::recache(const char *filename, const char *url)
{
  if (filename == NULL || url == NULL)
    return false;

  char *new_path = get_cache_filename_from_url(url);
  if (new_path == NULL)
    return false;

  bool result = repository.rename(filename, new_path);
  free(new_path);
  return result;
}


void
CacheManager::touch(const char *url)
{
  smart_ptr<CacheEntry> entry
    = dynamic_cast<CacheEntry *>(temporary_cache_table.lookup(url));
  if (static_cast<CacheEntry *>(entry) == NULL)
    return;	// åLRUå¸ߤʤ

  char *filename = get_cache_filename_from_url(url);
  if (filename == NULL)
    return;	// Out of Memory?

  char *fullpath = repository.find_file(filename);
  if (fullpath != NULL)
    {
      struct stat sb;
      if (utime(fullpath, NULL) == 0 && stat(fullpath, &sb) == 0)
	{
	  entry->set_mtime(sb.st_mtime);
	  cache_entry_list.sort(CacheEntryCompare());
	}
      free(fullpath);
    }

  free(filename);
}


void
CacheManager::invalidate(const char *url)
{
  smart_ptr<CacheEntry> entry
    = dynamic_cast<CacheEntry *>(temporary_cache_table.lookup(url));
  if (static_cast<CacheEntry *>(entry) != NULL)
    {
      // å夬LRUå¸ߤ硣
      cache_entry_list.erase(find(cache_entry_list.begin(),
				  cache_entry_list.end(),
				  static_cast<CacheEntry *>(entry)));
      invalidate(entry);
      return;
    }

  char *path = get_tenured_file_path(url);
  repository.unlink_file(path);
  free(path);
}


void
CacheManager::invalidate(CacheEntry *entry)
{
  char *url = get_url_from_cache_filename(entry->get_filename());

  repository.unlink_file(entry->get_filename());
  current_cache_size -= entry->get_size();
  current_number_of_files--;

  temporary_cache_table.remove(url);
  free(url);

#if 1	// for debugging
  entry->unset_in_hashtable();
#endif
}


bool
CacheManager::tenure(const char *url)
{
  smart_ptr<CacheEntry> entry
    = dynamic_cast<CacheEntry *>(temporary_cache_table.lookup(url));
  if (static_cast<CacheEntry *>(entry) == NULL)
    return false;

  char *tenured_path = get_tenured_file_path(url);
  bool result = repository.rename(entry->get_filename(), tenured_path);
  free(tenured_path);

  if (result)
    {
      current_cache_size -= entry->get_size();
      current_number_of_files--;
      cache_entry_list.erase(find(cache_entry_list.begin(),
				  cache_entry_list.end(),
				  static_cast<CacheEntry *>(entry)));

      temporary_cache_table.remove(url);
#if 1	// for debugging
      entry->unset_in_hashtable();
#endif
    }

  return result;
}


char *
CacheManager::get_tenured_file_path(const char *url) const
{
  if (url == NULL)
    return NULL;

  URI uri(url);
  if (!uri.is_valid())
    return NULL;

  string tmp_name(tenured_cache_directory);

  tmp_name.append("/");
#if 0	// ̯ġġ
  if (uri.get_user() != NULL)
    tmp_name.append(uri.get_user()).append("@");
#endif

  if (uri.get_server() != NULL)
    tmp_name.append(uri.get_server());

  if (uri.get_port() != 0)
    {
      char buf[32];
      snprintf(buf, 32, ":%u", uri.get_port());
      tmp_name.append(buf);
    }

  if (uri.get_path() != NULL)
    tmp_name.append(uri.get_path());
  else
    tmp_name.append("/");

  if (uri.get_query() != NULL)
    tmp_name.append("?").append(uri.get_query());

#if 0	// ϴޤ٤ʤ
  if (uri.get_fragment() != NULL)
    tmp_name.append("#").append(uri.get_fragment());
#endif

  return strdup(tmp_name.c_str());
}


void
CacheManager::register_file_as_cache(const char *fullpath, const char *name,
				     struct stat *sb, void *user_data)
{
  CacheManager *manager = static_cast<CacheManager *>(user_data);
  unsigned int h = string_hash(name);
  char filename[PATH_MAX];
  snprintf(filename, PATH_MAX,
	   "%s/%02u/%s", manager->temporary_cache_directory,
	   h % NUMBER_OF_CACHE_SLOTS, name);
  if (strstr(fullpath, filename) == NULL)
    {
      fprintf(stderr, "\"%s\" isn't a valid cache file\n", fullpath);
      return;
    }

  smart_ptr<CacheEntry> entry
    = new CacheEntry(filename, sb->st_mtime, sb->st_size);
  char *url = URI::decode_string(name);
  if (url == NULL)
    return;	// Out of Memory?

#if 1	// for debugging
  entry->set_in_hashtable();
#endif
  manager->temporary_cache_table.insert(url, entry);
  free(url);

  list<CacheEntry *> new_entry;
  new_entry.push_front(entry);
  manager->cache_entry_list.merge(new_entry, CacheEntryCompare());

  manager->current_cache_size += sb->st_size;
  manager->current_number_of_files++;
}


void
CacheManager::relocate_cache_file(const char *fullpath, const char *name,
				  struct stat *sb, void *user_data)
{
  CacheManager *manager = static_cast<CacheManager *>(user_data);
  char *url = URI::decode_string(name);
  if (url == NULL)
    return;	// Out of Memory?

  char *filename = manager->get_cache_filename_from_url(url);
  if (filename == NULL)
    {
      free(url);
      return;	// Out of Memory?
    }

  if (!manager->repository.rename(fullpath, filename))
    {
      free(url);
      free(filename);
      return;
    }

  smart_ptr<CacheEntry> entry
    = new CacheEntry(filename, sb->st_mtime, sb->st_size);
#if 1	// for debugging
  entry->set_in_hashtable();
#endif
  manager->temporary_cache_table.insert(url, entry);
  free(url);
  free(filename);

  list<CacheEntry *> new_entry;
  new_entry.push_front(entry);
  manager->cache_entry_list.merge(new_entry, CacheEntryCompare());

  manager->current_cache_size += sb->st_size;
  manager->current_number_of_files++;
}


void
CacheManager::update_cache_info()
{
  if (repository.get_home_path() == NULL)
    return;	// Ȥʪˤʤʤ

  if (!repository.directory_exist(temporary_cache_directory))
    return;

  current_cache_size = 0;
  current_number_of_files = 0;
  temporary_cache_table.clear();
  cache_entry_list.erase(cache_entry_list.begin(), cache_entry_list.end());

  char dirname[PATH_MAX];
  for (int i = 0; i < NUMBER_OF_CACHE_SLOTS; i++)
    {
      snprintf(dirname, PATH_MAX, "%s/%02u", temporary_cache_directory, i);
      repository.for_each_file(dirname, register_file_as_cache, this);
    }
  repository.for_each_file(temporary_cache_directory,
			   relocate_cache_file, this);
  ensure_limits();
}


void
CacheManager::ensure_limits()
{
  if (maximum_number_of_files > 0)
    {
      // ե¤н
      while (maximum_number_of_files < current_number_of_files)
	{
	  smart_ptr<CacheEntry> entry = cache_entry_list.front();
	  cache_entry_list.pop_front();
	  invalidate(entry);
	}
    }

  if (cache_capacity > 0)
    {
      // å¤н
      while (cache_capacity < current_cache_size)
	{
	  smart_ptr<CacheEntry> entry = cache_entry_list.front();
	  cache_entry_list.pop_front();
	  invalidate(entry);
	}
    }
}

}
