/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cache_manager.h,v 1.17 2008/09/24 11:49:14 fuyu Exp $
 */

#ifndef _LIBOCHUSHA_CACHE_MANAGER_H_
#define _LIBOCHUSHA_CACHE_MANAGER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

#include "libochusha/buffer.h"
#include "libochusha/hashtable.h"

#include <list>

#if defined(HAVE_STDINT_H) || !defined(HAVE_CONFIG_H)
#include <stdint.h>
#else
#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif
#endif

#include <sys/types.h>
#include <sys/stat.h>


#define DEFAULT_CACHE_CAPACITY		32		// 32MB
#define DEFAULT_MAX_NUMBER_OF_FILES	1024


using namespace std;
namespace Ochusha
{
  class CacheEntry;
  class Repository;
  class Tmpfile;

  class CacheManager
  {
  private:
    Repository &repository;

    char *temporary_cache_directory;
    char *tenured_cache_directory;

    uintmax_t cache_capacity;
    uintmax_t current_cache_size;

    unsigned int maximum_number_of_files;
    unsigned int current_number_of_files;

    RefCountHashtable temporary_cache_table;
    list<CacheEntry *> cache_entry_list;


    char *get_tenured_file_path(const char *url) const;
    void invalidate(CacheEntry *entry);
    void update_cache_info();
    void ensure_limits();

    char *get_url_from_cache_filename(const char *filename) const;

    static void register_file_as_cache(const char *fullpath, const char *name,
				       struct stat *sb, void *user_data);
    static void relocate_cache_file(const char *fullpath, const char *name,
				    struct stat *sb, void *user_data);


  protected:
    char *get_cache_filename_from_url(const char *url) const;


  public:
    CacheManager(Repository &repository,
		 const char *temporary_dir, const char *tenured_dir,
		 uintmax_t capacity = DEFAULT_CACHE_CAPACITY,
		 unsigned max_num_files = DEFAULT_MAX_NUMBER_OF_FILES);
    virtual ~CacheManager();

    inline uintmax_t get_cache_capacity()
    { return cache_capacity / (1024 * 1024); }
    void set_cache_capacity(uintmax_t capacity);

    inline unsigned int get_maximum_number_of_files()
    { return maximum_number_of_files; }
    void set_maximum_number_of_files(unsigned int max_num_files);

    void set_limits(uintmax_t capacity, unsigned int max_num_files);

    char *find_cache(const char *url) const;
    char *find_cache_directory(const char *url) const;
    int open_cache(const char *url, int flags) const;
    bool cache_exist(const char *url) const;
    bool is_tenured(const char *url) const;

    bool cache(const char *url, Tmpfile *tmpfile);
    bool recache(const char *filename, const char *url);
    void touch(const char *url);
    void invalidate(const char *url);

    bool tenure(const char *url);
  };
}
#endif

#endif	/* _LIBOCHUSHA_CACHE_MANAGER_H_ */
