/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: repository.cc,v 1.18 2008/10/10 15:55:14 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libochusha/repository.h"

#include "libochusha/bulletin_board.h"
#include "libochusha/misc_utils.h"

#include <cassert>
#include <cerrno>
#include <climits>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


namespace Ochusha
{

Tmpfile::Tmpfile(const char *fullpath, int fd)
{
  assert(fullpath != NULL && *fullpath != '\0');
  assert(fd >= 0);

  this->fullpath = strdup(fullpath);
  this->fd = fd;
}


Tmpfile::~Tmpfile()
{
  if (fd >= 0)
    {
      ::close(fd);
      unlink(fullpath);
    }

  free(fullpath);
}


void
Tmpfile::close()
{
  if (fd >= 0)
    ::close(fd);
  fd = -1;
}


const struct stat *
Tmpfile::stat()
{
  int result;
  if (fd >= 0)
    result = ::fstat(fd, &buf);
  else
    result = ::stat(fullpath, &buf);

  if (result == 0)
    return &buf;

  ErrorMessage error_message(errno);
  fprintf(stderr, LIB_("stat(2) failed due to: %s\n"), error_message.get_message());
  return NULL;
}


Repository::Repository(const char *home_path)
  : home_path(prepare_home_path(home_path)),
    bbs_data(*this)
{
}


Repository::~Repository()
{
  free(home_path);
}


char *
Repository::prepare_home_path(const char *path)
{
  if (path == NULL)
    return NULL;

  size_t len = strlen(path);
  char path_buf[PATH_MAX];
  if (path[len - 1] != '/')
    snprintf(path_buf, PATH_MAX, "%s/", path);
  else
    snprintf(path_buf, PATH_MAX, "%s", path);

  if (!mkdir_p(path_buf))
    return strdup(path_buf);
  else
    return NULL;
}


const char *
Repository::get_home_path()
{
  return home_path;
}


/*
 * Repository::file_exist
 *
 * ochushahomeǥ쥯ȥФpath̾filenameĥե뤬
 * ¸ߤ뤫ɤ򸡺¸ߤˤtrueʤˤfalse
 * ֤
 */
bool
Repository::file_exist(const char *filename) const
{
  struct stat sb;
  int len;
  char pathname[PATH_MAX];

  assert(home_path != NULL);
  assert(filename != NULL && *filename != '\0');

  if (*filename == '/')
    { // absolute pathξ
      pathname[PATH_MAX - 1] = '\0';
      strncpy(pathname, filename, PATH_MAX);
      if (pathname[PATH_MAX - 1] != '\0')
	return false;
    }
  else if (snprintf(pathname, PATH_MAX,
		    "%s%s", home_path, filename) >= PATH_MAX)
    return false;

  if (stat(pathname, &sb) == 0 && S_ISREG(sb.st_mode)
      && (sb.st_mode & S_IRUSR) != 0)
    return true;

  len = strlen(pathname);
  if (len > 3 && strcmp(pathname + len - 3, ".gz") == 0)
    {
      pathname[len - 3] = '\0';
      if (stat(pathname, &sb) == 0 && S_ISREG(sb.st_mode)
	  && (sb.st_mode & S_IRUSR) != 0)
	return true;
    }

  return false;
}


/*
 * Repository::directory_exist
 *
 * ochushahomeǥ쥯ȥФpath̾dirnameĥǥ쥯ȥ꤬
 * ¸ߤ뤫ɤ򸡺¸ߤˤtrueʤˤfalse
 * ֤
 */
bool
Repository::directory_exist(const char *dirname) const
{
  struct stat sb;
  char pathname[PATH_MAX];

  assert(home_path != NULL);
  assert(dirname != NULL && *dirname != '\0');

  if (*dirname == '/')
    { // absolute pathξ
      pathname[PATH_MAX - 1] = '\0';
      strncpy(pathname, dirname, PATH_MAX);
      if (pathname[PATH_MAX - 1] != '\0')
	return false;
    }
  else if (snprintf(pathname, PATH_MAX,
		    "%s%s", home_path, dirname) >= PATH_MAX)
    return false;

  if (stat(pathname, &sb) == 0 && S_ISDIR(sb.st_mode)
      && (sb.st_mode & (S_IRUSR|S_IWUSR|S_IXUSR)) == (S_IRUSR|S_IWUSR|S_IXUSR))
    return true;

  return false;
}


/*
 * Repository::find_file
 *
 * ochushahomeǥ쥯ȥФpath̾filenameĥե뤬
 * ¸ߤ뤫ɤ򸡺¸ߤˤϥѥ̾ɽʸ֤
 * callerϥѥ̾ʸפˤʤäfree()̳餦
 *
 * ե뤬ĤʤˤNULL֤
 */
char *
Repository::find_file(const char *filename) const
{
  struct stat sb;
  int len;
  char pathname[PATH_MAX];

  assert(home_path != NULL);
  assert(filename != NULL && *filename != '\0');

  if (*filename == '/')
    { // absolute pathξ
      pathname[PATH_MAX - 1] = '\0';
      strncpy(pathname, filename, PATH_MAX);
      if (pathname[PATH_MAX - 1] != '\0')
	return NULL;
    }
  else if (snprintf(pathname, PATH_MAX,
		    "%s%s", home_path, filename) >= PATH_MAX)
    return NULL;

  if (stat(pathname, &sb) == 0 && S_ISREG(sb.st_mode)
      && (sb.st_mode & S_IRUSR) != 0)
    return strdup(pathname);

  len = strlen(pathname);
  if (len > 3 && strcmp(pathname + len - 3, ".gz") == 0)
    {
      pathname[len - 3] = '\0';
      if (stat(pathname, &sb) == 0 && S_ISREG(sb.st_mode)
	  && (sb.st_mode & S_IRUSR) != 0)
	return strdup(pathname);
    }

  return NULL;
}


/*
 * Repository::find_directory
 *
 * ochushahomeǥ쥯ȥФpath̾dirnameĥǥ쥯ȥ꤬
 * ¸ߤ뤫ɤ򸡺¸ߤˤϥѥ̾ɽʸ֤
 * callerϥѥ̾ʸפˤʤäfree()̳餦
 * NULLdirͿ줿硢
 *
 * ǥ쥯ȥ꤬ĤʤˤNULL֤
 */
char *
Repository::find_directory(const char *dirname) const
{
  struct stat sb;
  char pathname[PATH_MAX];

  assert(home_path != NULL);
  assert(dirname != NULL && *dirname != '\0');

  if (*dirname == '/')
    { // absolute pathξ
      pathname[PATH_MAX - 1] = '\0';
      strncpy(pathname, dirname, PATH_MAX);
      if (pathname[PATH_MAX - 1] != '\0')
	return NULL;
    }
  else if (snprintf(pathname, PATH_MAX,
		    "%s%s", home_path, dirname) >= PATH_MAX)
    return NULL;

  if (stat(pathname, &sb) == 0 && S_ISDIR(sb.st_mode)
      && (sb.st_mode & (S_IRUSR|S_IWUSR|S_IXUSR)) == (S_IRUSR|S_IWUSR|S_IXUSR))
    return strdup(pathname);

  return NULL;
}


/*
 * Repository::open_file
 *
 * ochushahomeǥ쥯ȥФpath̾filenameĥե
 * flagsͿ줿⡼ɤǳ
 *
 * ե뤬ˤfile descriptor򡢳ʤäˤ-1֤
 */
int
Repository::open_file(const char *filename, int flags) const
{
  int result;
  int len = 0;
  char pathname[PATH_MAX];

  assert(home_path != NULL);
  assert(filename != NULL && *filename != '\0');

  if (*filename == '/')
    { // absolute pathξ
      pathname[PATH_MAX - 1] = '\0';
      strncpy(pathname, filename, PATH_MAX);
      if (pathname[PATH_MAX - 1] != '\0')
	return -1;
      len = strlen(pathname);
    }
  else if ((len = snprintf(pathname, PATH_MAX,
			   "%s%s", home_path, filename)) >= PATH_MAX)
    return -1;

  result = open(pathname, flags, 0x180);
  if (result >= 0 || (flags & O_CREAT) == 0)
    return result;

  /* O_CREATե饰դǳʤäˤϥǥ쥯ȥ򷡤äƤĩ */
  while (true)
    {
      if (pathname[len] == '/')
	break;
      len--;
    }
  pathname[len] = '\0';

  if (mkdir_p(pathname))
    return -1;

  pathname[len] = '/';

  return open(pathname, flags, 0x180);
}


/*
 * Repository::create_tmpfile
 *
 * ochushahomeǥ쥯ȥʲ˰եꡢбTmpfile
 * ֥Ȥ֤ԻˤNULL֤
 */
Tmpfile *
Repository::create_tmpfile(int flags) const
{
  assert(home_path != NULL);

  char path_template[PATH_MAX];
  if (snprintf(path_template, PATH_MAX, "%stmp", home_path) < PATH_MAX)
    if (mkdir_p(path_template))
      return NULL;

  bool retry = true;
 try_again:
  if (snprintf(path_template, PATH_MAX,
	       "%stmp/ochu-XXXXXX", home_path) < PATH_MAX)
    {
      int fd = mkstemp(path_template);
      if (fd >= 0)
	return new Tmpfile(path_template, fd);
      if (errno == EEXIST && retry)
	{
	  retry = false;
	  goto try_again;
	}
    }
  return NULL;
}


/*
 * Repository::unlink_file
 *
 * ochushahomeǥ쥯ȥФpath̾filenameĥե
 * unlink롣
 */
void
Repository::unlink_file(const char *filename) const
{
  struct stat sb;
  char pathname[PATH_MAX];

  assert(home_path != NULL);
  assert(filename != NULL && *filename != '\0');

  if (*filename == '/')
    { // absolute pathξ
      pathname[PATH_MAX - 1] = '\0';
      strncpy(pathname, filename, PATH_MAX);
      if (pathname[PATH_MAX - 1] != '\0')
	return;
    }
  else if (snprintf(pathname, PATH_MAX,
		    "%s%s", home_path, filename) >= PATH_MAX)
    return;

  if (stat(pathname, &sb) == 0
      && (S_ISREG(sb.st_mode) || S_ISLNK(sb.st_mode)))
    unlink(pathname);
}


/*
 * Repository::unlink_directory
 *
 * ochushahomeǥ쥯ȥФpath̾dirnameĥǥ쥯ȥ
 * unlink롣
 * ǥ쥯ȥ꤬Ǥʤˤϡǥ쥯ȥΥեǥ쥯ȥ
 * ƵŪunlink롣
 */
void
Repository::unlink_directory(const char *dirname) const
{
  struct stat sb;
  char pathname[PATH_MAX];
  DIR *dir_list;

  assert(home_path != NULL);

  if (*dirname == '/')
    { // absolute pathξ
      pathname[PATH_MAX - 1] = '\0';
      strncpy(pathname, dirname, PATH_MAX);
      if (pathname[PATH_MAX - 1] != '\0')
	return;
    }
  else if (snprintf(pathname, PATH_MAX, "%s%s", home_path, dirname) >= PATH_MAX)
    return;

  if (stat(pathname, &sb) != 0 || !S_ISDIR(sb.st_mode))
    return;

  dir_list = opendir(pathname);
  if (dir_list != NULL)
    {
      struct dirent *entry;
      char tmp_path[PATH_MAX];

      while ((entry = readdir(dir_list)) != NULL)
	{
	  const char *name = entry->d_name;
	  if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0)
	    continue;

	  if (snprintf(tmp_path, PATH_MAX, "%s/%s", pathname, name) < PATH_MAX)
	    {
	      if (stat(tmp_path, &sb) == 0)
		{
		  if (S_ISDIR(sb.st_mode))
		    {
		      if (snprintf(tmp_path, PATH_MAX,
				   "%s/%s", dirname, name) < PATH_MAX)
			unlink_directory(tmp_path);
		    }
		  else
		    {
		      if (snprintf(tmp_path, PATH_MAX,
				   "%s/%s", dirname, name) < PATH_MAX)
			unlink_file(tmp_path);
		    }
		}
	    }
	}
      closedir(dir_list);
    }

  rmdir(pathname);
}


/*
 * pathnameͿ줿ʸϢ³'/'Ĥ֤롣
 * internalʴؿʤΤNULLåϤʤ
 */
static void
canon_pathname(char *pathname)
{
  char *src_pos = pathname;
  char *dst_pos = pathname;

  while (*src_pos != '\0')
    {
      *dst_pos = *src_pos++;
      if (*dst_pos != '/' || *src_pos != '/')
	dst_pos++;
    }
  *dst_pos = '\0';
}


/*
 * home_path/fromnamehome_path/tonamerename롣
 * fromnametonameХѥǤˤϡͿ줿ХѥѤ롣
 */
bool
Repository::rename(const char *fromname, const char *toname) const
{
  char from_path[PATH_MAX];
  char to_path[PATH_MAX];

  assert(home_path != NULL);
  assert(fromname != NULL && *fromname != '\0');
  assert(toname != NULL && *toname != '\0');

  if (*fromname == '/')
    { // absolute pathξ
      from_path[PATH_MAX - 1] = '\0';
      strncpy(from_path, fromname, PATH_MAX);
      if (from_path[PATH_MAX - 1] != '\0')
	return false;
    }
  else if (snprintf(from_path, PATH_MAX,
		    "%s%s", home_path, fromname) < PATH_MAX)
    canon_pathname(from_path);
  else
    return false;

  if (*toname == '/')
    { // absolute pathξ
      to_path[PATH_MAX - 1] = '\0';
      strncpy(to_path, toname, PATH_MAX);
      if (to_path[PATH_MAX - 1] != '\0')
	return false;
    }
  else if (snprintf(to_path, PATH_MAX, "%s%s", home_path, toname) < PATH_MAX)
    canon_pathname(to_path);
  else
    return false;

  char *tmp_pos = strrchr(to_path, '/');
  if (tmp_pos != NULL)
    {
      *tmp_pos = '\0';
      if (mkdir_p(to_path) != 0)
	return false;
      *tmp_pos = '/';
    }

  return ::rename(from_path, to_path) == 0;
}


/*
 * home_path/dir/ľΤ줾ΥեФfuncŬѤ롣
 * dirabsolute pathξdir/ľΤ줾ΥեФfuncŬѤ롣
 */
void
Repository::for_each_file(const char *dir, ForEachFunc func,
			  void *user_data) const
{
  char dirpath[PATH_MAX];
  if (*dir != '/')
    snprintf(dirpath, PATH_MAX, "%s%s", home_path, dir);
  else
    snprintf(dirpath, PATH_MAX, "%s", dir);

  canon_pathname(dirpath);
  size_t len = strlen(dirpath);
  if (dirpath[len - 1] == '/')
    dirpath[len - 1] = '\0';	// '/'ϼΤơ

  struct stat sb;
  if (stat(dirpath, &sb) != 0 || !S_ISDIR(sb.st_mode))
    return;	// ɤʤǥ쥯ȥꤸʤ

  DIR *dir_list = opendir(dirpath);
  if (dir_list == NULL)
    return;	// ɤˤǤ͡

  struct dirent *entry;
  char fullpath[PATH_MAX];

  while ((entry = readdir(dir_list)) != NULL)
    {
      const char *name = entry->d_name;
      if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0)
	continue;

      if (snprintf(fullpath, PATH_MAX, "%s/%s", dirpath, name) < PATH_MAX)
	{
	  if (stat(fullpath, &sb) == 0 && !S_ISDIR(sb.st_mode))
	    (*func)(fullpath, name, &sb, user_data);
	}
    }

  closedir(dir_list);
}


/*
 * home_path/dir/ľΤ줾Υǥ쥯ȥФfuncŬѤ롣
 * dirabsolute pathξdir/ľΤ줾Υǥ쥯ȥФfunc
 * ŬѤ롣
 */
void
Repository::for_each_dir(const char *dir, ForEachFunc func,
			 void *user_data) const
{
  char dirpath[PATH_MAX];
  if (*dir != '/')
    snprintf(dirpath, PATH_MAX, "%s%s", home_path, dir);
  else
    snprintf(dirpath, PATH_MAX, "%s", dir);

  canon_pathname(dirpath);
  size_t len = strlen(dirpath);
  if (dirpath[len - 1] == '/')
    dirpath[len - 1] = '\0';	// '/'ϼΤơ

  struct stat sb;
  if (stat(dirpath, &sb) != 0 || !S_ISDIR(sb.st_mode))
    return;	// ɤʤǥ쥯ȥꤸʤ

  DIR *dir_list = opendir(dirpath);
  if (dir_list == NULL)
    return;	// ɤˤǤ͡

  struct dirent *entry;
  char fullpath[PATH_MAX];

  while ((entry = readdir(dir_list)) != NULL)
    {
      const char *name = entry->d_name;
      if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0)
	continue;

      if (snprintf(fullpath, PATH_MAX, "%s/%s", dirpath, name) < PATH_MAX)
	{
	  if (stat(fullpath, &sb) == 0 && S_ISDIR(sb.st_mode))
	    (*func)(fullpath, name, &sb, user_data);
	}
    }

  closedir(dir_list);
}


}
