/*
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: icon_label.c,v 1.11 2008/09/24 11:25:13 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"
#include "libochushagtk/icon_label.h"

#include <gtk/gtk.h>

#include <string.h>


static void ochusha_icon_label_class_init(OchushaIconLabelClass *klass);

static void ochusha_icon_label_init(OchushaIconLabel *icon_label);
static void ochusha_icon_label_set_property(GObject *object,
					    guint prop_id,
					    const GValue *value,
					    GParamSpec *pspec);
static void ochusha_icon_label_get_property(GObject *object,
					    guint prop_id,
					    GValue *value,
					    GParamSpec *pspec);
static void ochusha_icon_label_finalize(GObject *object);

static void ochusha_icon_label_destroy(GtkObject *object);
static void ochusha_icon_label_size_request(GtkWidget *widget,
					    GtkRequisition *requisition);
static void ochusha_icon_label_size_allocate(GtkWidget *widget,
					     GtkAllocation *allocation);
static int ochusha_icon_label_expose(GtkWidget *widget, GdkEventExpose *event);


GType
ochusha_icon_label_get_type(void)
{
  static GType icon_label_type = 0;

  if (icon_label_type == 0)
    {
      static const GTypeInfo icon_label_info =
	{
	  sizeof(OchushaIconLabelClass),
	  NULL,	/* base_init */
	  NULL,	/* base_finalize */
	  (GClassInitFunc)ochusha_icon_label_class_init,
	  NULL,	/* class_finalize */
	  NULL,	/* class_data */
	  sizeof(OchushaIconLabel),
	  0,	/* n_preallocs */
	  (GInstanceInitFunc)ochusha_icon_label_init,
	};

      icon_label_type = g_type_register_static(GTK_TYPE_LABEL,
					       "OchushaIconLabel",
					       &icon_label_info, 0);
    }

  return icon_label_type;
}


enum {
  PROP_0 = 0,
  PROP_PIXBUF,
  PROP_ANIMATION,
  PROP_STOCK
};


static GtkLabelClass *parent_class = NULL;
static GtkMiscClass *parent_parent_class = NULL;


#if !GLIB_CHECK_VERSION(2,8,0)
#define G_PARAM_STATIC_NAME	G_PARAM_PRIVATE
#define G_PARAM_STATIC_NICK	0
#define G_PARAM_STATIC_BLURB	0
#endif


#define OCHUSHA_PARAM_READWRITE G_PARAM_READWRITE|G_PARAM_STATIC_NAME|G_PARAM_STATIC_NICK|G_PARAM_STATIC_BLURB


static void
ochusha_icon_label_class_init(OchushaIconLabelClass *klass)
{
  GObjectClass *o_class = (GObjectClass *)klass;
  GtkObjectClass *object_class = (GtkObjectClass *)klass;
  GtkWidgetClass *widget_class = (GtkWidgetClass *)klass;

  parent_class = g_type_class_peek_parent(klass);
  parent_parent_class = g_type_class_peek_parent(parent_class);

  /* as a GObject */
  o_class->set_property = ochusha_icon_label_set_property;
  o_class->get_property = ochusha_icon_label_get_property;
  o_class->finalize = ochusha_icon_label_finalize;

  g_object_class_install_property(o_class,
				  PROP_PIXBUF,
				  g_param_spec_object("pixbuf",
						      PROP_("Pixbuf"),
						      PROP_("GdkPixbuf icon that will be displayed alongside of label text"),
						      GDK_TYPE_PIXBUF,
						      OCHUSHA_PARAM_READWRITE));

  g_object_class_install_property(o_class,
				  PROP_ANIMATION,
				  g_param_spec_object("pixbuf-animation",
						      PROP_("Animation"),
						      PROP_("GdkPixbufAnimation icon that will be displayed alongside of label text"),
						      GDK_TYPE_PIXBUF_ANIMATION,
						      OCHUSHA_PARAM_READWRITE));

  g_object_class_install_property(o_class,
				  PROP_STOCK,
				  g_param_spec_string("stock",
						      PROP_("Stock ID"),
						      PROP_("Stock ID for a stock icon to display"),
						      NULL,
						      OCHUSHA_PARAM_READWRITE));

  /* as a GtkObject */
  object_class->destroy = ochusha_icon_label_destroy;

  /* as a GtkWidget */
  widget_class->size_request = ochusha_icon_label_size_request;
  widget_class->size_allocate = ochusha_icon_label_size_allocate;
  widget_class->expose_event = ochusha_icon_label_expose;
}


static void
ochusha_icon_label_init(OchushaIconLabel *icon_label)
{
  icon_label->pixbuf = NULL;
  icon_label->animation = NULL;
  icon_label->animation_iter = NULL;
  icon_label->stock_id = NULL;

  icon_label->frame_timeout = 0;
  icon_label->icon_width = 0;
  icon_label->icon_height = 0;
  icon_label->icon_y_adjustment = 0;
}


static void
ochusha_icon_label_set_property(GObject *object, guint prop_id,
				const GValue *value, GParamSpec *pspec)
{
  OchushaIconLabel *icon_label = OCHUSHA_ICON_LABEL(object);

  switch (prop_id)
    {
    case PROP_PIXBUF:
      ochusha_icon_label_set_icon_from_pixbuf(icon_label,
				(GdkPixbuf *)g_value_get_object(value));
      break;

    case PROP_ANIMATION:
      ochusha_icon_label_set_icon_from_animation(icon_label,
			(GdkPixbufAnimation *)g_value_get_object(value));
      break;

    case PROP_STOCK:
      ochusha_icon_label_set_icon_from_stock(icon_label,
					     g_value_get_string(value));
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
    }
}


static void
ochusha_icon_label_get_property(GObject *object, guint prop_id,
				GValue *value, GParamSpec *pspec)
{
  OchushaIconLabel *icon_label = OCHUSHA_ICON_LABEL(object);

  switch (prop_id)
    {
    case PROP_PIXBUF:
      g_value_set_object(value, (GObject *)icon_label->pixbuf);
      break;

    case PROP_ANIMATION:
      g_value_set_object(value, (GObject *)icon_label->animation);
      break;

    case PROP_STOCK:
      g_value_set_string(value, icon_label->stock_id);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
      break;
    }
}


static void
ochusha_icon_label_finalize(GObject *object)
{
#if 0
  OchushaIconLabel *icon_label = OCHUSHA_ICON_LABEL(object);
#endif

  if (G_OBJECT_CLASS(parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS(parent_class)->finalize)(object);
}


static void
ochusha_icon_label_invalidate_animation(OchushaIconLabel *icon_label)
{
  if (icon_label->animation != NULL)
    {
      g_object_unref(icon_label->animation);
      icon_label->animation = NULL;
    }

  if (icon_label->frame_timeout != 0)
    {
      g_source_remove(icon_label->frame_timeout);
      icon_label->frame_timeout = 0;
    }

  if (icon_label->animation_iter != NULL)
    {
      g_object_unref(icon_label->animation_iter);
      icon_label->animation_iter = NULL;
    }
}


static void
ochusha_icon_label_destroy(GtkObject *object)
{
  OchushaIconLabel *icon_label = OCHUSHA_ICON_LABEL(object);

  if (icon_label->pixbuf != NULL)
    {
      g_object_unref(icon_label->pixbuf);
      icon_label->pixbuf = NULL;
    }

  ochusha_icon_label_invalidate_animation(icon_label);

  if (icon_label->stock_id != NULL)
    {
      g_free(icon_label->stock_id);
      icon_label->stock_id = NULL;
    }

  if (GTK_OBJECT_CLASS(parent_class)->destroy != NULL)
    (*GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}


static gint
animation_timeout_cb(gpointer data)
{
  OchushaIconLabel *icon_label;
  int delay;

  gdk_threads_enter();

  icon_label = OCHUSHA_ICON_LABEL(data);
  icon_label->frame_timeout = 0;

  if (icon_label->animation_iter != NULL)
    {
      gdk_pixbuf_animation_iter_advance(icon_label->animation_iter, NULL);
      delay = gdk_pixbuf_animation_iter_get_delay_time(icon_label->animation_iter);
      if (delay >= 0)
	icon_label->frame_timeout
	  = g_timeout_add(delay, animation_timeout_cb, icon_label);

      gtk_widget_queue_draw(GTK_WIDGET(icon_label));

      if (GTK_WIDGET_DRAWABLE(icon_label))
	gdk_window_process_updates(GTK_WIDGET(icon_label)->window, TRUE);
    }

  gdk_threads_leave();

  return FALSE;
}


static GdkPixbuf *
ochusha_icon_label_get_current_pixbuf(OchushaIconLabel *icon_label)
{
  if (icon_label->pixbuf != NULL)
    return icon_label->pixbuf;

  if (icon_label->animation != NULL)
    {
      if (icon_label->animation_iter == NULL)
	{
	  int delay;
	  icon_label->animation_iter
	    = gdk_pixbuf_animation_get_iter(icon_label->animation, NULL);

	  delay = gdk_pixbuf_animation_iter_get_delay_time(icon_label->animation_iter);
	  if (delay >= 0)
	    icon_label->frame_timeout
	      = g_timeout_add(delay, animation_timeout_cb, icon_label);
	}

      return gdk_pixbuf_animation_iter_get_pixbuf(icon_label->animation_iter);
    }

  if (icon_label->stock_id != NULL)
    return gtk_widget_render_icon(GTK_WIDGET(icon_label), icon_label->stock_id,
				  GTK_ICON_SIZE_MENU, NULL);

  return NULL;
}


static void
ochusha_icon_label_update_size(OchushaIconLabel *icon_label,
			       gint icon_width, gint icon_height)
{
  icon_label->icon_width = icon_width;
  icon_label->icon_height = icon_height;
}


static void
ochusha_icon_label_size_request(GtkWidget *widget, GtkRequisition *requisition)
{
  OchushaIconLabel *icon_label = OCHUSHA_ICON_LABEL(widget);
  gint width = icon_label->icon_width;
  gint height = icon_label->icon_height;

  (*GTK_WIDGET_CLASS(parent_class)->size_request)(widget, requisition);

  if (icon_label->stock_id != NULL)
    {
      GdkPixbuf *pixbuf = ochusha_icon_label_get_current_pixbuf(icon_label);
      if (pixbuf != NULL)
	{
	  width += gdk_pixbuf_get_width(pixbuf);
	  height += gdk_pixbuf_get_height(pixbuf);
	  g_object_unref(pixbuf);
	}
    }

  requisition->width += width;	/* iconʬä롣*/
  requisition->height = MAX(height, requisition->height);
}


static void
ochusha_icon_label_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
{
  OchushaIconLabel *icon_label = OCHUSHA_ICON_LABEL(widget);
  int icon_width = icon_label->icon_width;
  int icon_height = icon_label->icon_height;
  int y_adjustment;
  int height_adjustment;

  if (icon_label->stock_id != NULL)
    {
      GdkPixbuf *pixbuf = ochusha_icon_label_get_current_pixbuf(icon_label);
      if (pixbuf != NULL)
	{
	  icon_width += gdk_pixbuf_get_width(pixbuf);
	  icon_height += gdk_pixbuf_get_height(pixbuf);
	  g_object_unref(pixbuf);
	}
    }

  if (icon_width > 0 && icon_height > 0)
    {
      GtkRequisition requisition;
      
      (*GTK_WIDGET_CLASS(parent_class)->size_request)(widget, &requisition);
      if (requisition.height < icon_height)
	{
	  height_adjustment = icon_height - requisition.height;
	  y_adjustment = height_adjustment / 2;
	  icon_label->icon_y_adjustment = 0;
	}
      else
	{
	  icon_label->icon_y_adjustment
	    = (requisition.height - icon_height) / 2;
	  y_adjustment = 0;
	  height_adjustment = 0;
	}
    }
  else
    {
      y_adjustment = 0;
      height_adjustment = 0;
    }

  icon_label->fake_allocation.x = allocation->x + icon_width;
  icon_label->fake_allocation.y = allocation->y + y_adjustment;
  icon_label->fake_allocation.width = allocation->width - icon_width;
  icon_label->fake_allocation.height = allocation->height - height_adjustment;
  (*GTK_WIDGET_CLASS(parent_class)->size_allocate)(widget,
						&icon_label->fake_allocation);

  /* Hack: GtkLabelؤ٤αƶǤä*/
  (*GTK_WIDGET_CLASS(parent_parent_class)->size_allocate)(widget, allocation);

}


static int
ochusha_icon_label_expose(GtkWidget *widget, GdkEventExpose *event)
{
  OchushaIconLabel *icon_label = OCHUSHA_ICON_LABEL(widget);
  GtkAllocation allocation = widget->allocation;

  /* ʪallocationȸGtkLabelʬ褵 */
  widget->allocation = icon_label->fake_allocation;
  (*GTK_WIDGET_CLASS(parent_class)->expose_event)(widget, event);

  /* ᤹ */
  widget->allocation = allocation;

  if (GTK_WIDGET_VISIBLE(widget) && GTK_WIDGET_MAPPED(widget))
    {
      GdkPixbuf *pixbuf = ochusha_icon_label_get_current_pixbuf(icon_label);

      if (pixbuf != NULL)
	{
	  /* icon褹롣*/
	  gdk_draw_pixbuf(widget->window,
			  widget->style->black_gc,	/* Τ */
			  pixbuf,
			  0, 0,				/* src_x, src_y */
			  allocation.x,			/* dst_x */
			  allocation.y + icon_label->icon_y_adjustment,
			  				/* dst_y */
			  -1, -1,			/* use pixbuf's size */
			  GDK_RGB_DITHER_NONE,
			  0, 0);
	  if (icon_label->stock_id != NULL)
	    g_object_unref(pixbuf);
	}
    }

  return FALSE;
}


GtkWidget *
ochusha_icon_label_new(const gchar *label_text)
{
  OchushaIconLabel *icon_label = g_object_new(OCHUSHA_TYPE_ICON_LABEL, NULL);

  if (label_text != NULL && *label_text != '\0')
    gtk_label_set_text(GTK_LABEL(icon_label), label_text);

  return GTK_WIDGET(icon_label);
}


GtkWidget *
ochusha_icon_label_new_from_stock(const gchar *label_text,
				  const gchar *stock_id)
{
  OchushaIconLabel *icon_label = g_object_new(OCHUSHA_TYPE_ICON_LABEL, NULL);

  if (label_text != NULL && *label_text != '\0')
    gtk_label_set_text(GTK_LABEL(icon_label), label_text);

  if (stock_id != NULL)
    ochusha_icon_label_set_icon_from_stock(icon_label, stock_id);

  return GTK_WIDGET(icon_label);
}


GtkWidget *
ochusha_icon_label_new_with_mnemonic(const gchar *mnemonic)
{
  OchushaIconLabel *icon_label = g_object_new(OCHUSHA_TYPE_ICON_LABEL, NULL);

  if (mnemonic != NULL && *mnemonic != '\0')
    gtk_label_set_text_with_mnemonic(GTK_LABEL(icon_label), mnemonic);

  return GTK_WIDGET(icon_label);
}


GtkWidget *
ochusha_icon_label_new_with_mnemonic_from_stock(const gchar *mnemonic,
						const gchar *stock_id)
{
  OchushaIconLabel *icon_label = g_object_new(OCHUSHA_TYPE_ICON_LABEL, NULL);

  if (mnemonic != NULL && *mnemonic != '\0')
    gtk_label_set_text_with_mnemonic(GTK_LABEL(icon_label), mnemonic);

  if (stock_id != NULL)
    ochusha_icon_label_set_icon_from_stock(icon_label, stock_id);

  return GTK_WIDGET(icon_label);
}


void
ochusha_icon_label_set_icon_from_pixbuf(OchushaIconLabel *icon_label,
					GdkPixbuf *pixbuf)
{
  g_return_if_fail(OCHUSHA_IS_ICON_LABEL(icon_label));

  if (pixbuf != NULL && icon_label->pixbuf == pixbuf)
    return;

  g_object_freeze_notify(G_OBJECT(icon_label));

  if (icon_label->pixbuf != NULL)
    g_object_unref(icon_label->pixbuf);

  if (pixbuf != NULL)
    g_object_ref(pixbuf);

  icon_label->pixbuf = pixbuf;

  ochusha_icon_label_invalidate_animation(icon_label);

  if (icon_label->stock_id != NULL)
    {
      g_free(icon_label->stock_id);
      icon_label->stock_id = NULL;
    }

  if (pixbuf != NULL)
    ochusha_icon_label_update_size(icon_label,
				   gdk_pixbuf_get_width(pixbuf),
				   gdk_pixbuf_get_height(pixbuf));
  else
    ochusha_icon_label_update_size(icon_label, 0, 0);

  g_object_notify(G_OBJECT(icon_label), "pixbuf");

  g_object_thaw_notify(G_OBJECT(icon_label));

  gtk_widget_queue_draw(GTK_WIDGET(icon_label));
}


void
ochusha_icon_label_set_icon_from_animation(OchushaIconLabel *icon_label,
					   GdkPixbufAnimation *animation)
{
  g_return_if_fail(OCHUSHA_IS_ICON_LABEL(icon_label));

  if (animation != NULL && icon_label->animation == animation)
    return;

  g_object_freeze_notify(G_OBJECT(icon_label));

  ochusha_icon_label_invalidate_animation(icon_label);

  if (animation != NULL)
    g_object_ref(animation);

  icon_label->animation = animation;

  if (icon_label->pixbuf != NULL)
    {
      g_object_unref(icon_label->pixbuf);
      icon_label->pixbuf = NULL;
    }

  if (icon_label->stock_id != NULL)
    {
      g_free(icon_label->stock_id);
      icon_label->stock_id = NULL;
    }

  if (animation != NULL)
    ochusha_icon_label_update_size(icon_label,
				   gdk_pixbuf_animation_get_width(animation),
				   gdk_pixbuf_animation_get_height(animation));
  else
    ochusha_icon_label_update_size(icon_label, 0, 0);

  g_object_notify(G_OBJECT(icon_label), "pixbuf-animation");

  g_object_thaw_notify(G_OBJECT(icon_label));

  gtk_widget_queue_draw(GTK_WIDGET(icon_label));
}


void
ochusha_icon_label_set_icon_from_stock(OchushaIconLabel *icon_label,
				       const gchar *stock_id)
{
  gchar *prev_id;
  g_return_if_fail(OCHUSHA_IS_ICON_LABEL(icon_label));

  if (stock_id != NULL && *stock_id == '\0')
    stock_id = NULL;

  prev_id = icon_label->stock_id;
  if (prev_id == stock_id
      || (prev_id != NULL && stock_id != NULL
	  && strcmp(prev_id, stock_id) == 0))
    return;

  g_object_freeze_notify(G_OBJECT(icon_label));

  icon_label->stock_id = stock_id != NULL ? g_strdup(stock_id) : NULL;

  if (icon_label->pixbuf != NULL)
    {
      g_object_unref(icon_label->pixbuf);
      icon_label->pixbuf = NULL;
    }

  ochusha_icon_label_invalidate_animation(icon_label);

  ochusha_icon_label_update_size(icon_label, 0, 0);

  g_object_notify(G_OBJECT(icon_label), "stock");

  g_object_thaw_notify(G_OBJECT(icon_label));

  if (prev_id != NULL)
    g_free(prev_id);

  gtk_widget_queue_draw(GTK_WIDGET(icon_label));
}


GdkPixbuf *
ochusha_icon_label_get_icon_pixbuf(OchushaIconLabel *icon_label)
{
  g_return_val_if_fail(OCHUSHA_IS_ICON_LABEL(icon_label), NULL);

  return icon_label->pixbuf;
}


GdkPixbufAnimation *
ochusha_icon_label_get_icon_animation(OchushaIconLabel *icon_label)
{
  g_return_val_if_fail(OCHUSHA_IS_ICON_LABEL(icon_label), NULL);

  return icon_label->animation;
}


const gchar *
ochusha_icon_label_get_icon_stock(OchushaIconLabel *icon_label)
{
  g_return_val_if_fail(OCHUSHA_IS_ICON_LABEL(icon_label), NULL);

  return icon_label->stock_id;
}
