/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: virtual_machine.cc,v 1.18 2008/10/25 09:20:59 fuyu Exp $
 */

#include "config.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "libtscore/virtual_machine.h"

#include "libtscore/bytecode_buffer.h"
#include "libtscore/cell.h"
#include "libtscore/tscore.h"

#include <cassert>
#include <stdint.h>


namespace LibTSCore
{

Register::Register(VirtualMachine &vm, Cell *proc, Cell *caller_frame,
		   Cell *cpool, Cell *inport, Cell *outport)
  : memory_system(vm.memory_system),
    procedure(proc), frame(caller_frame), constant_pool(cpool),
    input_port(inport), output_port(outport), value(Cell::unspecified()),
    pc(NULL), saved_pc(NULL), trampoline_frame(Cell::nil()), nargs(0)
{
  vm.memory_system.stack_roots.push_back(&procedure);
  vm.memory_system.stack_roots.push_back(&frame);
  vm.memory_system.stack_roots.push_back(&constant_pool);
  vm.memory_system.stack_roots.push_back(&input_port);
  vm.memory_system.stack_roots.push_back(&output_port);
  vm.memory_system.stack_roots.push_back(&value);
  vm.memory_system.stack_roots.push_back(&trampoline_frame);
}


Register::~Register()
{
  assert(memory_system.stack_roots.back() == &trampoline_frame);
  memory_system.stack_roots.pop_back();
  assert(memory_system.stack_roots.back() == &value);
  memory_system.stack_roots.pop_back();
  assert(memory_system.stack_roots.back() == &output_port);
  memory_system.stack_roots.pop_back();
  assert(memory_system.stack_roots.back() == &input_port);
  memory_system.stack_roots.pop_back();
  assert(memory_system.stack_roots.back() == &constant_pool);
  memory_system.stack_roots.pop_back();
  assert(memory_system.stack_roots.back() == &frame);
  memory_system.stack_roots.pop_back();
  assert(memory_system.stack_roots.back() == &procedure);
  memory_system.stack_roots.pop_back();
}


const VirtualMachine::OpcodeInfo VirtualMachine::opcode_info[] = {
  { "*not used*", 1 },
  { "iconst_m1", 1 },
  { "iconst_0", 1 },
  { "iconst_1", 1 },
  { "iconst_2", 1 },
  { "iconst_3", 1 },
  { "iconst_4", 1 },
  { "iconst_i1", 2 },
  { "iconst_i2", 3 },
  { "iconst_i4", 5 },
  { "iconst_i8", 9 },
  { "iconst_null", 1 },
  { "iconst_t", 1 },
  { "iconst_f", 1 },
  { "const_p_0", 1 },
  { "const_p_1", 1 },
  { "const_p_2", 1 },
  { "const_p_3", 1 },
  { "const_p1", 2 },
  { "const_p2", 3 },
  { "const_p4", 5 },
  { "load_l_0", 1 },
  { "load_l_1", 1 },
  { "load_l_2", 1 },
  { "load_l_3", 1 },
  { "load_l_4", 1 },
  { "load_l_5", 1 },
  { "load_l_6", 1 },
  { "load_l_7", 1 },
  { "store_l_0", 1 },
  { "store_l_1", 1 },
  { "store_l_2", 1 },
  { "store_l_3", 1 },
  { "store_l_4", 1 },
  { "store_l_5", 1 },
  { "store_l_6", 1 },
  { "store_l_7", 1 },
  { "load_l1", 2 },
  { "load_l2", 3 },
  { "load_l4", 5 },
  { "store_l1", 2 },
  { "store_l2", 3 },
  { "store_l4", 5 },
  { "load_g1", 3 },
  { "load_g2", 4 },
  { "load_g4", 6 },
  { "store_g1", 3 },
  { "store_g2", 4 },
  { "store_g4", 6 },
  { "load_t1", 2 },
  { "load_t2", 3 },
  { "load_t4", 5 },
  { "store_t1", 2 },
  { "store_t2", 3 },
  { "store_t4", 5 },
  { "load_f1", 2 },
  { "load_f2", 3 },
  { "load_f4", 5 },
  { "store_f1", 2 },
  { "store_f2", 3 },
  { "store_f4", 5 },
  { "if_t1", 2 },
  { "if_t2", 3 },
  { "if_t4", 5 },
  { "if_f1", 2 },
  { "if_f2", 3 },
  { "if_f4", 5 },
  { "goto1", 2 },
  { "goto2", 3 },
  { "goto4", 5 },
  { "car", 1 },
  { "cdr", 1 },
  { "apply1", 3 },
  { "apply2", 5 },
  { "apply4", 9 },
  { "call_cc1", 3 },
  { "call_cc2", 5 },
  { "call_cc4", 9 },
  { "apply_continuation0", 1 },
  { "lambda", 1 },
  { "trampoline", 1 },
  { "meta_return", 1 },
  { "nop", 1 }
};


VirtualMachine::VirtualMachine(TSCore &tscore, long temporary_frame_size)
  : core(tscore), memory_system(tscore.memory_system), reader(*this),
    tracing(false), bytecompile_tracing(false), bytecode_tracing(false)
{
  continuation_cache = Cell::nil();
  toplevel_constant_pool = Cell::nil();
  temporary_frame = Cell::nil();

  memory_system.register_static_root(&continuation_cache);
  memory_system.register_static_root(&toplevel_constant_pool);
  memory_system.register_static_root(&temporary_frame);

  toplevel_constant_pool = Cell::constant_pool_new(memory_system);
  temporary_frame = frame_new(Cell::nil(), temporary_frame_size)
    ->set_frame_closed();
}


VirtualMachine::~VirtualMachine()
{
  continuation_cache = Cell::nil();
  toplevel_constant_pool = Cell::nil();

  memory_system.unregister_static_root(&continuation_cache);
  memory_system.unregister_static_root(&toplevel_constant_pool);
  memory_system.unregister_static_root(&temporary_frame);
}


Cell *
VirtualMachine::unbound_variable_error(const char *name)
{
  error_buffer.clear();
  error_buffer.append("undefined variable: ")->append(name);
  return NULL;
}


    // This is only for car/cdr at this time.
Cell *
VirtualMachine::type_mismatch_error(const char *name, Cell *val)
{
  error_buffer.clear();
  error_buffer.append(name);
  error_buffer.append(": argument #1 must be of pair : ");
  Writer::write_datum(val, &error_buffer, false);
  return NULL;
}


Cell *
VirtualMachine::application_error(Cell *val)
{
  error_buffer.clear();
  error_buffer.append("bad procedure: ");
  Writer::write_datum(val, &error_buffer, false);
  return NULL;
}


Cell *
VirtualMachine::mk_regexp(Cell *rx_string, bool ignore_case)
{
  Cell *x = memory_system.get_cell(rx_string);
  return x->mk_regexp(rx_string, ignore_case);
}


Cell *
VirtualMachine::mk_input_file_port(const char *name)
{
  FILE *file = fopen(name, "r");
  if (file == NULL)
    return Cell::nil();

  Port *port = new InputFilePort(file, true);
  return mk_port(port);
}


Cell *
VirtualMachine::mk_output_file_port(const char *name)
{
  FILE *file = fopen(name, "w");
  if (file == NULL)
    return Cell::nil();

  Port *port = new OutputFilePort(file, true);
  return mk_port(port);
}


Cell *
VirtualMachine::mk_input_string_port(Cell *str)
{
  Port *port = new InputStringPort(str->get_string(), str->get_strlen());
  if (port == NULL)
    return Cell::nil();
  return mk_port(port);
}


Cell *
VirtualMachine::mk_output_string_port()
{
  Port *port = new OutputStringPort();
  if (port == NULL)
    return Cell::nil();
  return mk_port(port);
}


#ifndef WORDS_BIGENDIAN
# if !defined(__i386) && !defined(__i386__) && !defined(_M_IX86) && \
  !(defined(__ppc__) && defined(__APPLE__)) &&			   \
  !defined(__x86_64) && !defined(__x86_64__) &&			   \
  !defined(__mc68020__)
#  define WORDS_BIGENDIAN 1
# endif
#endif


static inline long
read_signed_8(const char *pc)
{
  return static_cast<long>(*pc);
}


static inline long
read_signed_16(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_SHORT != 2
  long value = static_cast<long>(pc[1]) << 8;
  return value | reinterpret_cast<const unsigned char *>(pc)[0];
#else
  return static_cast<long>(reinterpret_cast<const short *>(pc)[0]);
#endif
}


static inline long
read_signed_32(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_INT != 4
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<long>(pc[3]) << 24
    | static_cast<long>(upc[2]) << 16
    | static_cast<long>(upc[1]) << 8
    | static_cast<long>(upc[0]);
#else
  return static_cast<long>(reinterpret_cast<const int *>(pc)[0]);
#endif
}


static inline long
read_signed_64(const char *pc)
{
#if SIZEOF_LONG >= 8
# if defined(WORDS_BIGENDIAN) || SIZEOF_LONG != 8
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<long>(pc[7]) << 56
    | static_cast<long>(upc[6]) << 48
    | static_cast<long>(upc[5]) << 40
    | static_cast<long>(upc[4]) << 32
    | static_cast<long>(upc[3]) << 24
    | static_cast<long>(upc[2]) << 16
    | static_cast<long>(upc[1]) << 8
    | static_cast<long>(upc[0]);
# else
  return static_cast<long>(reinterpret_cast<const long *>(pc)[0]);
# endif
#else
  // We just ignore higher 32 bits of 64 bits integer here.
  return read_signed_32(pc);
#endif
}


static inline long
read_unsigned_8(const char *pc)
{
  return static_cast<long>(reinterpret_cast<const unsigned char *>(pc)[0]);
}


static inline long
read_unsigned_16(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_SHORT != 2
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<long>(upc[1]) << 8 | static_cast<long>(upc[0]);
#else
  return static_cast<long>(reinterpret_cast<const short *>(pc)[0]);
#endif
}


static inline long
read_unsigned_32(const char *pc)
{
#if defined(WORDS_BIGENDIAN) || SIZEOF_INT != 4
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<long>(upc[3]) << 24
    | static_cast<long>(upc[2]) << 16
    | static_cast<long>(upc[1]) << 8
    | static_cast<long>(upc[0]);
#else
  return static_cast<long>(reinterpret_cast<const unsigned int *>(pc)[0]);
#endif
}


static inline long
read_unsigned_64(const char *pc)
{
#if SIZEOF_LONG >= 8
# if defined(WORDS_BIGENDIAN) || SIZEOF_LONG != 8
  const unsigned char *upc = reinterpret_cast<const unsigned char *>(pc);
  return static_cast<long>(upc[7]) << 56
    | static_cast<long>(upc[6]) << 48
    | static_cast<long>(upc[5]) << 40
    | static_cast<long>(upc[4]) << 32
    | static_cast<long>(upc[3]) << 24
    | static_cast<long>(upc[2]) << 16
    | static_cast<long>(upc[1]) << 8
    | static_cast<long>(upc[0]);
# else
  return static_cast<long>(reinterpret_cast<const unsigned long *>(pc)[0]);
# endif
#else
  // We just ignore higher 32 bits of 64 bits integer here.
  return read_unsigned_32(pc);
#endif
}


void
VirtualMachine::print_bytecode(Register &context)
{
  if (!context.output_port->is_output_port())
    return;
  Port *port = context.output_port->get_port();
  port->append("=> value: ");
  Writer::write_datum(context.value, port, false);
  port->append('\n');
  print_bytecode(port, context.pc, context.constant_pool);
}


void
VirtualMachine::print_bytecode(Port *port,
			       const char *from, Cell *constant_pool)
{
  char buf[256];
  switch (static_cast<TSVMOpcode>(*from))
    {
    case OP_NOT_USED:
    case OP_ICONST_M1:
    case OP_ICONST_0:
    case OP_ICONST_1:
    case OP_ICONST_2:
    case OP_ICONST_3:
    case OP_ICONST_4:
    case OP_CONST_NULL:
    case OP_CONST_T:
    case OP_CONST_F:
    case OP_APPLY_CONTINUATION0:
    case OP_LAMBDA:
    case OP_TRAMPOLINE:
    case OP_META_RETURN:
    case OP_NOP:
    case OP_CAR:
    case OP_CDR:
    case OP_LOAD_L_0:
    case OP_LOAD_L_1:
    case OP_LOAD_L_2:
    case OP_LOAD_L_3:
    case OP_LOAD_L_4:
    case OP_LOAD_L_5:
    case OP_LOAD_L_6:
    case OP_LOAD_L_7:
    case OP_STORE_L_0:
    case OP_STORE_L_1:
    case OP_STORE_L_2:
    case OP_STORE_L_3:
    case OP_STORE_L_4:
    case OP_STORE_L_5:
    case OP_STORE_L_6:
    case OP_STORE_L_7:
      snprintf(buf, 256, "%p: %s\n", from, get_opcode_name(*from));
      port->append(buf);
      return;

    case OP_ICONST_I1:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_signed_8(from + 1));
      port->append(buf);
      return;

    case OP_ICONST_I2:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_signed_16(from + 1));
      port->append(buf);
      return;

    case OP_ICONST_I4:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_signed_32(from + 1));
      port->append(buf);
      return;

    case OP_ICONST_I8:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_signed_64(from + 1));
      port->append(buf);
      return;

    case OP_CONST_P_0:
      snprintf(buf, 256, "%p: %s  // ", from, get_opcode_name(*from));
      port->append(buf);
      Writer::write_datum(constant_pool->load_constant(0), port, false);
      port->append('\n');
      return;

    case OP_CONST_P_1:
      snprintf(buf, 256, "%p: %s  // ", from, get_opcode_name(*from));
      port->append(buf);
      Writer::write_datum(constant_pool->load_constant(1), port, false);
      port->append('\n');
      return;

    case OP_CONST_P_2:
      snprintf(buf, 256, "%p: %s  // ", from, get_opcode_name(*from));
      port->append(buf);
      Writer::write_datum(constant_pool->load_constant(2), port, false);
      port->append('\n');
      return;

    case OP_CONST_P_3:
      snprintf(buf, 256, "%p: %s  // ", from, get_opcode_name(*from));
      port->append(buf);
      Writer::write_datum(constant_pool->load_constant(3), port, false);
      port->append('\n');
      return;

    case OP_CONST_P1:
    case OP_LOAD_F1:
    case OP_STORE_F1:
      snprintf(buf, 256, "%p: %s %ld  // ",
	       from, get_opcode_name(*from), read_unsigned_8(from + 1));
      port->append(buf);
      Writer::write_datum(constant_pool->load_constant(read_unsigned_8(from + 1)), port, false);
      port->append('\n');
      return;

    case OP_CONST_P2:
    case OP_LOAD_F2:
    case OP_STORE_F2:
      snprintf(buf, 256, "%p: %s %ld  // ",
	       from, get_opcode_name(*from), read_unsigned_16(from + 1));
      port->append(buf);
      Writer::write_datum(constant_pool->load_constant(read_unsigned_16(from + 1)), port, false);
      port->append('\n');
      return;

    case OP_CONST_P4:
    case OP_LOAD_F4:
    case OP_STORE_F4:
      snprintf(buf, 256, "%p: %s %ld  // ",
	       from, get_opcode_name(*from), read_unsigned_32(from + 1));
      port->append(buf);
      Writer::write_datum(constant_pool->load_constant(read_unsigned_32(from + 1)), port, false);
      port->append('\n');
      return;

    case OP_LOAD_L1:
    case OP_STORE_L1:
    case OP_LOAD_T1:
    case OP_STORE_T1:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_unsigned_8(from + 1));
      port->append(buf);
      return;

    case OP_LOAD_L2:
    case OP_STORE_L2:
    case OP_LOAD_T2:
    case OP_STORE_T2:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_unsigned_16(from + 1));
      port->append(buf);
      return;


    case OP_LOAD_L4:
    case OP_STORE_L4:
    case OP_LOAD_T4:
    case OP_STORE_T4:
      snprintf(buf, 256, "%p: %s %ld\n",
	       from, get_opcode_name(*from), read_unsigned_32(from + 1));
      port->append(buf);
      return;


    case OP_LOAD_G1:
    case OP_STORE_G1:
      snprintf(buf, 256, "%p: %s %ld %ld\n",
	       from, get_opcode_name(*from),
	       read_unsigned_8(from + 1), read_unsigned_8(from + 2));
      port->append(buf);
      return;

    case OP_LOAD_G2:
    case OP_STORE_G2:
      snprintf(buf, 256, "%p: %s %ld %ld\n",
	       from, get_opcode_name(*from),
	       read_unsigned_8(from + 1), read_unsigned_16(from + 2));
      port->append(buf);
      return;

    case OP_LOAD_G4:
    case OP_STORE_G4:
      snprintf(buf, 256, "%p: %s %ld %ld\n",
	       from, get_opcode_name(*from),
	       read_unsigned_8(from + 1), read_unsigned_32(from + 2));
      port->append(buf);
      return;

    case OP_IF_T1:
    case OP_IF_F1:
    case OP_GOTO1:
      snprintf(buf, 256, "%p: %s %p\n",
	       from, get_opcode_name(*from),
	       from + read_signed_8(from + 1) + 1);
      port->append(buf);
      return;

    case OP_IF_T2:
    case OP_IF_F2:
    case OP_GOTO2:
      snprintf(buf, 256, "%p: %s %p\n",
	       from, get_opcode_name(*from),
	       from + read_signed_16(from + 1) + 1);
      port->append(buf);
      return;

    case OP_IF_T4:
    case OP_IF_F4:
    case OP_GOTO4:
      snprintf(buf, 256, "%p: %s %p\n",
	       from, get_opcode_name(*from),
	       from + read_signed_32(from + 1) + 1);
      port->append(buf);
      return;

    case OP_APPLY1:
    case OP_CALL_CC1:
      snprintf(buf, 256, "%p: %s %ld %ld\n",
	       from, get_opcode_name(*from),
	       read_unsigned_8(from + 1), read_unsigned_8(from + 2));
      port->append(buf);
      return;

    case OP_APPLY2:
    case OP_CALL_CC2:
      snprintf(buf, 256, "%p: %s %ld %ld\n",
	       from, get_opcode_name(*from),
	       read_unsigned_16(from + 1), read_unsigned_16(from + 3));
      port->append(buf);
      return;

    case OP_APPLY4:
    case OP_CALL_CC4:
      snprintf(buf, 256, "%p: %s %ld %ld\n",
	       from, get_opcode_name(*from),
	       read_unsigned_32(from + 1), read_unsigned_32(from + 5));
      port->append(buf);
      return;
    }
}


const char *
VirtualMachine::get_next_pc(const char *current_pc)
{
  return current_pc + get_opcode_length(*current_pc);
}


Cell *
VirtualMachine::apply(Cell *procedure, Register &caller_context,
		      unsigned long argument_pointer,
		      unsigned long number_of_arguments)
{
  Register context(*this, Cell::nil(), caller_context.frame, Cell::nil(),
		   caller_context.input_port, caller_context.output_port);


#ifdef __GNUG__
# define LABEL(label)		label:
# ifdef PIC
#  define LABEL_OFF(label)	(reinterpret_cast<intptr_t>(&&label) - reinterpret_cast<intptr_t>(&&op_iconst_m1))
#  ifdef NDEBUG
#   define GOTO_NEXT_OPCODE() \
  { goto *reinterpret_cast<void *>(reinterpret_cast<intptr_t>(&&op_iconst_m1) + label_offsets[static_cast<long>(*context.pc++)]); }
#  else
#   define GOTO_NEXT_OPCODE()						\
  { if (bytecode_tracing)						\
      print_bytecode(context);						\
    goto *reinterpret_cast<void *>(reinterpret_cast<intptr_t>(&&op_iconst_m1) + label_offsets[static_cast<long>(*context.pc++)]); }
#  endif	// !NDEBUG
  static const ptrdiff_t label_offsets[] = {

# else	// non-PIC

#  define LABEL_OFF(label)	(&&label)
#  ifdef NDEBUG
#   define GOTO_NEXT_OPCODE()	{ goto *const_cast<void *>(labels[static_cast<long>(*context.pc++)]); }
#  else
#   define GOTO_NEXT_OPCODE()						\
  { if (bytecode_tracing)						\
      print_bytecode(context);						\
    goto *const_cast<void *>(labels[static_cast<long>(*context.pc++)]); }
#  endif	// !NDEBUG
  static const void *labels[] = {

# endif

    0,
    LABEL_OFF(op_iconst_m1), LABEL_OFF(op_iconst_0), LABEL_OFF(op_iconst_1),
    LABEL_OFF(op_iconst_2), LABEL_OFF(op_iconst_3), LABEL_OFF(op_iconst_4),

    LABEL_OFF(op_iconst_i1), LABEL_OFF(op_iconst_i2),
    LABEL_OFF(op_iconst_i4), LABEL_OFF(op_iconst_i8),

    LABEL_OFF(op_const_null), LABEL_OFF(op_const_t), LABEL_OFF(op_const_f),

    LABEL_OFF(op_const_p_0), LABEL_OFF(op_const_p_1),
    LABEL_OFF(op_const_p_2), LABEL_OFF(op_const_p_3),

    LABEL_OFF(op_const_p1), LABEL_OFF(op_const_p2), LABEL_OFF(op_const_p4),

    LABEL_OFF(op_load_l_0), LABEL_OFF(op_load_l_1),
    LABEL_OFF(op_load_l_2), LABEL_OFF(op_load_l_3),
    LABEL_OFF(op_load_l_4), LABEL_OFF(op_load_l_5),
    LABEL_OFF(op_load_l_6), LABEL_OFF(op_load_l_7),

    LABEL_OFF(op_store_l_0), LABEL_OFF(op_store_l_1),
    LABEL_OFF(op_store_l_2), LABEL_OFF(op_store_l_3),
    LABEL_OFF(op_store_l_4), LABEL_OFF(op_store_l_5),
    LABEL_OFF(op_store_l_6), LABEL_OFF(op_store_l_7),

    LABEL_OFF(op_load_l1), LABEL_OFF(op_load_l2), LABEL_OFF(op_load_l4),

    LABEL_OFF(op_store_l1), LABEL_OFF(op_store_l2), LABEL_OFF(op_store_l4),

    LABEL_OFF(op_load_g1), LABEL_OFF(op_load_g2), LABEL_OFF(op_load_g4),

    LABEL_OFF(op_store_g1), LABEL_OFF(op_store_g2), LABEL_OFF(op_store_g4),

    LABEL_OFF(op_load_t1), LABEL_OFF(op_load_t2), LABEL_OFF(op_load_t4),

    LABEL_OFF(op_store_t1), LABEL_OFF(op_store_t2), LABEL_OFF(op_store_t4),

    LABEL_OFF(op_load_f1), LABEL_OFF(op_load_f2), LABEL_OFF(op_load_f4),

    LABEL_OFF(op_store_f1), LABEL_OFF(op_store_f2), LABEL_OFF(op_store_f4),

    LABEL_OFF(op_if_t1), LABEL_OFF(op_if_t2), LABEL_OFF(op_if_t4),
    LABEL_OFF(op_if_f1), LABEL_OFF(op_if_f2), LABEL_OFF(op_if_f4),
    LABEL_OFF(op_goto1), LABEL_OFF(op_goto2), LABEL_OFF(op_goto4),
    LABEL_OFF(op_car), LABEL_OFF(op_cdr),

    LABEL_OFF(op_apply1), LABEL_OFF(op_apply2), LABEL_OFF(op_apply4),
    LABEL_OFF(op_call_cc1), LABEL_OFF(op_call_cc2), LABEL_OFF(op_call_cc4),
    LABEL_OFF(op_apply_continuation0),
    LABEL_OFF(op_lambda), LABEL_OFF(op_trampoline),
    LABEL_OFF(op_meta_return),
    LABEL_OFF(op_nop)
  };
#else
#define GOTO_NEXT_OPCODE()	opcode = static_cast<TSVMOpcode>(*context.pc++); continue
#define LABEL(label)
#endif

  if (procedure->is_procedure())
    {
      Procedure *invoker = procedure->get_procedure_invoker();
      Cell *result = invoker->call(*this, procedure,
				   context, context.frame,
				   argument_pointer, number_of_arguments);
      if (result == NULL)
	return NULL;	// error
    }
  else
    return application_error(procedure);

#ifndef NDEBUG
  if (bytecode_tracing)
    print_bytecode(context);
#endif

  TSVMOpcode opcode = static_cast<TSVMOpcode>(*context.pc++);

  while (true)
    {
      switch (opcode)
	{
	case OP_ICONST_M1: LABEL(op_iconst_m1);
	  context.value = mk_integer(-1);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_0: LABEL(op_iconst_0);
	  context.value = mk_integer(0);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_1: LABEL(op_iconst_1);
	  context.value = mk_integer(1);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_2: LABEL(op_iconst_2);
	  context.value = mk_integer(2);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_3: LABEL(op_iconst_3);
	  context.value = mk_integer(3);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_4: LABEL(op_iconst_4);
	  context.value = mk_integer(4);
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_I1: LABEL(op_iconst_i1);
	  context.value = mk_integer(read_signed_8(context.pc));
	  ++context.pc;
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_I2: LABEL(op_iconst_i2);
	  context.value = mk_integer(read_signed_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_I4: LABEL(op_iconst_i4);
	  context.value = mk_integer(read_signed_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_ICONST_I8: LABEL(op_iconst_i8);
	  context.value = mk_integer(read_signed_64(context.pc));
	  context.pc += 8;
	  GOTO_NEXT_OPCODE();


	case OP_CONST_NULL: LABEL(op_const_null);
	  context.value = Cell::nil();
	  GOTO_NEXT_OPCODE();


	case OP_CONST_T: LABEL(op_const_t);
	  context.value = Cell::t();
	  GOTO_NEXT_OPCODE();


	case OP_CONST_F: LABEL(op_const_f);
	  context.value = Cell::f();
	  GOTO_NEXT_OPCODE();


#define LOAD_CONSTANT(index) \
	  context.value = context.constant_pool->load_constant(index)

	case OP_CONST_P_0: LABEL(op_const_p_0);
	  LOAD_CONSTANT(0);
	  GOTO_NEXT_OPCODE();


	case OP_CONST_P_1: LABEL(op_const_p_1);
	  LOAD_CONSTANT(1);
	  GOTO_NEXT_OPCODE();


	case OP_CONST_P_2: LABEL(op_const_p_2);
	  LOAD_CONSTANT(2);
	  GOTO_NEXT_OPCODE();


	case OP_CONST_P_3: LABEL(op_const_p_3);
	  LOAD_CONSTANT(3);
	  GOTO_NEXT_OPCODE();


	case OP_CONST_P1: LABEL(op_const_p1);
	  LOAD_CONSTANT(read_unsigned_8(context.pc));
	  ++context.pc;
	  GOTO_NEXT_OPCODE();


	case OP_CONST_P2: LABEL(op_const_p2);
	  LOAD_CONSTANT(read_unsigned_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_CONST_P4: LABEL(op_const_p4);
	  LOAD_CONSTANT(read_unsigned_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


#define LOAD_LOCAL_VARIABLE(slot) \
  context.value = context.frame->load_variable(slot)

	case OP_LOAD_L_0: LABEL(op_load_l_0);
	  LOAD_LOCAL_VARIABLE(0);
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L_1: LABEL(op_load_l_1);
	  LOAD_LOCAL_VARIABLE(1);
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L_2: LABEL(op_load_l_2);
	  LOAD_LOCAL_VARIABLE(2);
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L_3: LABEL(op_load_l_3);
	  LOAD_LOCAL_VARIABLE(3);
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L_4: LABEL(op_load_l_4);
	  LOAD_LOCAL_VARIABLE(4);
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L_5: LABEL(op_load_l_5);
	  LOAD_LOCAL_VARIABLE(5);
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L_6: LABEL(op_load_l_6);
	  LOAD_LOCAL_VARIABLE(6);
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L_7: LABEL(op_load_l_7);
	  LOAD_LOCAL_VARIABLE(7);
	  GOTO_NEXT_OPCODE();


#define STORE_LOCAL_VARIABLE(slot) \
  context.frame->store_variable(memory_system, slot, context.value)

	case OP_STORE_L_0: LABEL(op_store_l_0);
	  STORE_LOCAL_VARIABLE(0);
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L_1: LABEL(op_store_l_1);
	  STORE_LOCAL_VARIABLE(1);
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L_2: LABEL(op_store_l_2);
	  STORE_LOCAL_VARIABLE(2);
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L_3: LABEL(op_store_l_3);
	  STORE_LOCAL_VARIABLE(3);
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L_4: LABEL(op_store_l_4);
	  STORE_LOCAL_VARIABLE(4);
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L_5: LABEL(op_store_l_5);
	  STORE_LOCAL_VARIABLE(5);
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L_6: LABEL(op_store_l_6);
	  STORE_LOCAL_VARIABLE(6);
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L_7: LABEL(op_store_l_7);
	  STORE_LOCAL_VARIABLE(7);
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L1: LABEL(op_load_l1);
	  LOAD_LOCAL_VARIABLE(read_unsigned_8(context.pc));
	  ++context.pc;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L2: LABEL(op_load_l2);
	  LOAD_LOCAL_VARIABLE(read_unsigned_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_L4: LABEL(op_load_l4);
	  LOAD_LOCAL_VARIABLE(read_unsigned_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L1: LABEL(op_store_l1);
	  STORE_LOCAL_VARIABLE(read_unsigned_8(context.pc));
	  ++context.pc;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L2: LABEL(op_store_l2);
	  STORE_LOCAL_VARIABLE(read_unsigned_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_L4: LABEL(op_store_l4);
	  STORE_LOCAL_VARIABLE(read_unsigned_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


#define LOAD_GLOBAL_VARIABLE(nesting, slot)		\
  do				    			\
    {							\
      long n = nesting;					\
      Cell *frame = context.frame;			\
      for (long i = 0; i < n; i++)			\
	frame = frame->get_parent_frame();		\
      context.value = frame->load_variable(slot);	\
    } while (0)

	case OP_LOAD_G1: LABEL(op_load_g1);
	  LOAD_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
			       read_unsigned_8(context.pc + 1));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_G2: LABEL(op_load_g2);
	  LOAD_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
			       read_unsigned_16(context.pc + 1));
	  context.pc += 3;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_G4: LABEL(op_load_g4);
	  LOAD_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
			       read_unsigned_32(context.pc + 1));
	  context.pc += 5;
	  GOTO_NEXT_OPCODE();


#define STORE_GLOBAL_VARIABLE(nesting, slot)				\
  do						    			\
    {									\
      long n = nesting;							\
      Cell *frame = context.frame;					\
      for (long i = 0; i < n; i++)					\
	frame = frame->get_parent_frame();				\
      frame->store_variable(memory_system, slot, context.value);	\
    } while (0)

	case OP_STORE_G1: LABEL(op_store_g1);
	  STORE_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
				read_unsigned_8(context.pc + 1));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_G2: LABEL(op_store_g2);
	  STORE_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
				read_unsigned_16(context.pc + 1));
	  context.pc += 3;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_G4: LABEL(op_store_g4);
	  STORE_GLOBAL_VARIABLE(read_unsigned_8(context.pc),
				read_unsigned_32(context.pc + 1));
	  context.pc += 5;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_T1: LABEL(op_load_t1);
	  context.value = load_toplevel_variable(read_unsigned_8(context.pc));
	  ++context.pc;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_T2: LABEL(op_load_t2);
	  context.value = load_toplevel_variable(read_unsigned_16(context.pc));
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_T4: LABEL(op_load_t4);
	  context.value = load_toplevel_variable(read_unsigned_32(context.pc));
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_T1: LABEL(op_store_t1);
	  store_toplevel_variable(read_unsigned_8(context.pc), context.value);
	  ++context.pc;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_T2: LABEL(op_store_t2);
	  store_toplevel_variable(read_unsigned_16(context.pc), context.value);
	  context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_STORE_T4: LABEL(op_store_t4);
	  store_toplevel_variable(read_unsigned_32(context.pc), context.value);
	  context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_LOAD_F1: LABEL(op_load_f1);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_8(context.pc));
	    ++context.pc;
	    const Cell *binding = lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    context.value = load_toplevel_variable(slot);
		    GOTO_NEXT_OPCODE();
		  }
	      }
	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_LOAD_F2: LABEL(op_load_f2);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_16(context.pc));
	    context.pc += 2;
	    const Cell *binding = lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    context.value = load_toplevel_variable(slot);
		    GOTO_NEXT_OPCODE();
		  }
	      }
	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_LOAD_F4: LABEL(op_load_f4);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_32(context.pc));
	    context.pc += 4;
	    const Cell *binding = lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    context.value = load_toplevel_variable(slot);
		    GOTO_NEXT_OPCODE();
		  }
	      }
	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_STORE_F1: LABEL(op_store_f1);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_8(context.pc));
	    ++context.pc;
	    const Cell *binding = lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    store_toplevel_variable(slot, context.value);
		    GOTO_NEXT_OPCODE();
		  }
	      }
	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_STORE_F2: LABEL(op_store_f2);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_16(context.pc));
	    context.pc += 2;
	    const Cell *binding = lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    store_toplevel_variable(slot, context.value);
		    GOTO_NEXT_OPCODE();
		  }
	      }
	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_STORE_F4: LABEL(op_store_f4);
	  {
	    const Cell *symbol = LOAD_CONSTANT(read_unsigned_32(context.pc));
	    context.pc += 4;
	    const Cell *binding = lookup_toplevel_binding(symbol);
	    if (binding->is_binding())
	      {
		long slot = binding->get_slot_location();
		if (slot >= 0)
		  {
		    store_toplevel_variable(slot, context.value);
		    GOTO_NEXT_OPCODE();
		  }
	      }
	    return unbound_variable_error(symbol->get_symname());
	  }


	case OP_IF_T1: LABEL(op_if_t1);
	  if (context.value != Cell::f())
	    context.pc += read_signed_8(context.pc);
	  else
	    ++context.pc;
	  GOTO_NEXT_OPCODE();


	case OP_IF_T2: LABEL(op_if_t2);
	  if (context.value != Cell::f())
	    context.pc += read_signed_16(context.pc);
	  else
	    context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_IF_T4: LABEL(op_if_t4);
	  if (context.value != Cell::f())
	    context.pc += read_signed_32(context.pc);
	  else
	    context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_IF_F1: LABEL(op_if_f1);
	  if (context.value == Cell::f())
	    context.pc += read_signed_8(context.pc);
	  else
	    ++context.pc;
	  GOTO_NEXT_OPCODE();


	case OP_IF_F2: LABEL(op_if_f2);
	  if (context.value == Cell::f())
	    context.pc += read_signed_16(context.pc);
	  else
	    context.pc += 2;
	  GOTO_NEXT_OPCODE();


	case OP_IF_F4: LABEL(op_if_f4);
	  if (context.value == Cell::f())
	    context.pc += read_signed_32(context.pc);
	  else
	    context.pc += 4;
	  GOTO_NEXT_OPCODE();


	case OP_GOTO1: LABEL(op_goto1);
	  context.pc += read_signed_8(context.pc);
	  GOTO_NEXT_OPCODE();


	case OP_GOTO2: LABEL(op_goto2);
	  context.pc += read_signed_16(context.pc);
	  GOTO_NEXT_OPCODE();


	case OP_GOTO4: LABEL(op_goto4);
	  context.pc += read_signed_32(context.pc);
	  GOTO_NEXT_OPCODE();


	case OP_CAR: LABEL(op_car);
	  if (context.value->is_pair())
	    context.value = CAR(context.value);
	  else
	    return type_mismatch_error("car", context.value);
	  GOTO_NEXT_OPCODE();


	case OP_CDR: LABEL(op_cdr);
	  if (context.value->is_pair())
	    context.value = CDR(context.value);
	  else
	    return type_mismatch_error("cdr", context.value);
	  GOTO_NEXT_OPCODE();


#define COPY_CONTINUATION(slot)						\
  context.frame->store_variable(memory_system, slot,			\
				context.frame->load_variable(0))

	case OP_APPLY1: LABEL(op_apply1);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_8(context.pc);
	      ++context.pc;
	      unsigned long nargs = read_unsigned_8(context.pc);
	      ++context.pc;
	      COPY_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();
	      return NULL;	// error
	    }
	  return application_error(context.value);


	case OP_APPLY2: LABEL(op_apply2);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_16(context.pc);
	      context.pc += 2;
	      unsigned long nargs = read_unsigned_16(context.pc);
	      context.pc += 2;
	      COPY_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();
	      return NULL;	// error
	    }
	  return application_error(context.value);


	case OP_APPLY4: LABEL(op_apply4);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_32(context.pc);
	      context.pc += 4;
	      unsigned long nargs = read_unsigned_32(context.pc);
	      context.pc += 4;
	      COPY_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();
	      return NULL;	// error
	    }
	  return application_error(context.value);


#define SET_CURRENT_CONTINUATION(slot)	\
  context.frame->store_variable(memory_system, slot, Cell::nil())

	case OP_CALL_CC1: LABEL(op_call_cc1);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_8(context.pc);
	      ++context.pc;
	      unsigned long nargs = read_unsigned_8(context.pc);
	      ++context.pc;
	      SET_CURRENT_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();
	      return NULL;	// error
	    }
	  return application_error(context.value);


	case OP_CALL_CC2: LABEL(op_call_cc2);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_16(context.pc);
	      context.pc += 2;
	      unsigned long nargs = read_unsigned_16(context.pc);
	      context.pc += 2;
	      SET_CURRENT_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();
	      return NULL;	// error
	    }
	  return application_error(context.value);


	case OP_CALL_CC4: LABEL(op_call_cc4);
	  if (context.value->is_procedure())
	    {
	      unsigned long args = read_unsigned_32(context.pc);
	      context.pc += 4;
	      unsigned long nargs = read_unsigned_32(context.pc);
	      context.pc += 4;
	      SET_CURRENT_CONTINUATION(args);
	      Cell *proc = context.value;
	      Procedure *invoker = proc->get_procedure_invoker();
	      Cell *result = invoker->call(*this, proc, context,
					   context.frame, args, nargs);
	      if (result != NULL)
		GOTO_NEXT_OPCODE();
	      return NULL;	// error
	    }
	  return application_error(context.value);


	case OP_APPLY_CONTINUATION0: LABEL(op_apply_continuation0);
	  {
	    Cell *continuation = context.frame->load_variable(0);
	    continuation->get_continuation()->apply(*this,
						    continuation, context,
						    context.value);
	    GOTO_NEXT_OPCODE();
	  }


	case OP_LAMBDA: LABEL(op_lambda);
	  context.value = mk_closure(context.value, context.frame);
	  GOTO_NEXT_OPCODE();


	case OP_TRAMPOLINE: LABEL(op_trampoline);
	  {
	    assert(context.trampoline_frame->is_scheme_frame());
	    context.pc = context.saved_pc;
	    context.saved_pc = NULL;
	    Cell *proc = context.value;
	    Procedure *invoker = proc->get_procedure_invoker();
	    Cell *result = invoker->call(*this, proc, context,
					 context.trampoline_frame, 0,
					 context.nargs);
	    if (result != NULL)
	      GOTO_NEXT_OPCODE();
	    return NULL;	// error
	  }


	case OP_META_RETURN: LABEL(op_meta_return);
	  return context.value;


	case OP_NOP: LABEL(op_nop);
	  // do nothing
	  GOTO_NEXT_OPCODE();

	default:
	  // never reached here!
	  abort();
	}
    }
}


Cell *
VirtualMachine::eval(Register &context, Cell *expression)
{
  toplevel_constant_pool->reset_constant_pool();
  error_buffer.clear();
  StackRoot exp(memory_system, mk_cons(expression, Cell::nil0(), false));
  Cell *env = environment_new(get_toplevel_environment());
  env->allocate_anonymous_slot();	// dummy
  Cell *closure_template
    = Lambda::closure_template_new(*this, context, exp, env,
				   toplevel_constant_pool,
				   Cell::nil(), Cell::nil(), exp);
  if (closure_template == Cell::nil())
    {
#if 1
      fprintf(stderr, "bytecompilation error: %s\n",
	      error_buffer.get_output_string());
#endif
      return NULL;	// bytecompilation error
    }

  temporary_frame->store_variable(memory_system, 0, core.get_meta_return());
  Cell *procedure = mk_closure(closure_template, Cell::nil());

  context.frame = temporary_frame;
  return apply(procedure, context, 0, 1);
}


void
VirtualMachine::read_eval_print(Cell *input_port, Cell *output_port,
				const char *prompt)
{
  Register repl_context(*this, Cell::nil(), Cell::nil(), Cell::nil(),
			input_port, output_port);
  if (!input_port->is_input_port())
    {
      fprintf(stderr, "Invalid input port\n");
      return;
    }

  Port *inport = input_port->get_port();
  if (prompt != NULL && output_port->is_output_port())
    {
      // interactive repl
      Port *outport = output_port->get_port();
      while (true)
	{
	  outport->append(prompt);

	  Cell *expression = reader.read_datum(inport);
	  const char *error_message = NULL;
	  if (expression != NULL)
	    {
	      Cell *result = eval(repl_context, expression);
	      if (result != NULL)
		{
		  Writer::write_datum(result, outport, true, true);
		  if (result == Cell::eof())
		    return;
		  continue;
		}
	      error_message = error_buffer.get_output_string();
	    }
	  else
	    error_message = reader.get_error_message();

	  outport->append("Error: ")->append(error_message);
	  inport->flush_remaining_input();
	}
    }

  // non interactive repl
  while (true)
    {
      Cell *expression = reader.read_datum(inport);
      if (expression != NULL)
	{
	  Cell *result = eval(repl_context, expression);
	  if (result != NULL)
	    {
	      if (result == Cell::eof())
		return;
	      continue;
	    }
	}
      return;	// error
    }
}


Cell *
VirtualMachine::apply(Cell *procedure, Cell *args)
{
  if (procedure->is_symbol())
    {
      const Cell *binding = lookup_toplevel_binding(procedure);
      if (!binding->is_binding())
	{
	  error_buffer.clear();
	  error_buffer.append("unbound variable: ");
	  error_buffer.append(procedure->get_symname());
#if 0
	  fprintf(stderr, "%s\n", error_buffer.get_output_string());
#endif
	  return NULL;
	}
      procedure = load_toplevel_variable(binding->get_slot_location());
    }
  if (!procedure->is_procedure())
    {
      error_buffer.clear();
      error_buffer.append("cannot apply: ");
      Writer::write_datum(procedure, &error_buffer, false);
#if 0
      fprintf(stderr, "%s\n", error_buffer.get_output_string());
#endif
      return NULL;
    }

  long nargs = 1;
  temporary_frame->store_variable(memory_system, 0, core.get_meta_return());
  long frame_size = temporary_frame->get_frame_size();
  while (args->is_pair())
    {
      if (frame_size <= nargs)
	{
	  char buf[128];
	  snprintf(buf, 128,
		   "Too many arguments for this VirtualMachine: %ld\n",
		   nargs);
	  error_buffer.clear();
	  error_buffer.append(buf);
#if 0
	  fprintf(stderr, "%s\n", error_buffer.get_output_string());
#endif
	  return NULL;
	}
      temporary_frame->store_variable(memory_system, nargs++, CAR(args));
      args = CDR(args);
    }

  Register context(*this, Cell::nil(), temporary_frame, Cell::nil(),
		   core.get_interaction_input_port(),
		   core.get_interaction_output_port());
  return apply(procedure, context, 0, nargs);
}


Cell *
VirtualMachine::applyf(Cell *procedure, const char *fmt, va_list ap)
{
  if (procedure->is_symbol())
    {
      const Cell *binding = lookup_toplevel_binding(procedure);
      if (!binding->is_binding())
	{
	  error_buffer.clear();
	  error_buffer.append("unbound variable: ");
	  error_buffer.append(procedure->get_symname());
#if 0
	  fprintf(stderr, "VM::applyf()@1: %s\n",
		  error_buffer.get_output_string());
#endif
	  return NULL;
	}
      procedure = load_toplevel_variable(binding->get_slot_location());
    }

  if (!procedure->is_procedure())
    {
      error_buffer.clear();
      error_buffer.append("cannot apply: ");
      Writer::write_datum(procedure, &error_buffer, false);
#if 0
      fprintf(stderr, "VM::applyf()@2: %s\n",
	      error_buffer.get_output_string());
#endif
      return NULL;
    }

  StackRoot proc(memory_system, procedure);
  long nargs = 1;
  temporary_frame->store_variable(memory_system, 0, core.get_meta_return());
  long frame_size = temporary_frame->get_frame_size();
  if (fmt != NULL)
    {
      bool post_percent_state = false;
      bool long_argument = false;

      while (*fmt != '\0')
	{
	  if (frame_size <= nargs)
	    {
	      char buf[128];
	      snprintf(buf, 128,
		      "Too many arguments for this VirtualMachine: %ld\n",
		      nargs);
	      error_buffer.clear();
	      error_buffer.append(buf);
#if 0
	      fprintf(stderr, "VM::applyf()@3: %s\n",
		      error_buffer.get_output_string());
#endif
	      return NULL;
	    }
	  switch (*fmt)
	    {
	    case '%':
	      post_percent_state = !post_percent_state;
	      break;

	    case 'l':
	      if (!post_percent_state)
		break;
	      long_argument = true;
	      break;

	    case 'd':
	      if (!post_percent_state)
		break;
	      if (long_argument)
		temporary_frame->store_variable(memory_system, nargs++,
						mk_integer(va_arg(ap, long)));
	      else
		temporary_frame->store_variable(memory_system, nargs++,
				mk_integer(static_cast<long>(va_arg(ap, int))));
	      post_percent_state = false;
	      long_argument = false;
	      break;

	    case 'f':
	      if (!post_percent_state)
		break;
	      temporary_frame->store_variable(memory_system, nargs++,
					      mk_real(va_arg(ap, double)));
	      post_percent_state = false;
	      long_argument = false;
	      break;

	    case 'c':
	      if (!post_percent_state)
		break;
	      temporary_frame->store_variable(memory_system, nargs++,
					      mk_character(va_arg(ap, int)));
	      post_percent_state = false;
	      long_argument = false;
	      break;

	    case 'p':
	      if (!post_percent_state)
		break;
	      temporary_frame->store_variable(memory_system, nargs++,
					mk_foreign_object(va_arg(ap, char *)));
	      post_percent_state = false;
	      long_argument = false;
	      break;

	    case 's':
	      if (!post_percent_state)
		break;
	      temporary_frame->store_variable(memory_system, nargs++,
					      mk_static_string(va_arg(ap,
								      char *)));
	      post_percent_state = false;
	      long_argument = false;
	      break;

	    default:
	      if (post_percent_state)
		fprintf(stderr, "unknown format '%%%c'\n", *fmt);
	      else
		fprintf(stderr, "not a format specifier '%c' ignored\n", *fmt);
	      post_percent_state = false;
	      long_argument = false;
	      break;
	    }
	  fmt++;
	}
    }

  Register context(*this, Cell::nil(), temporary_frame, Cell::nil(),
		   core.get_interaction_input_port(),
		   core.get_interaction_output_port());
  return apply(proc, context, 0, nargs);
}

}
