/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: virtual_machine.h,v 1.18 2008/09/24 11:26:46 fuyu Exp $
 */

#ifndef _LIBTSCORE_VIRTUAL_MACHINE_H_
#define _LIBTSCORE_VIRTUAL_MACHINE_H_

#ifdef __cplusplus

#if defined(__GNUG__) && !defined(__ICC)
#pragma interface
#endif

namespace LibTSCore
{
  class Cell;
  class MemorySystem;
  class VirtualMachine;
  class Register
  {
  private:
    MemorySystem &memory_system;

  public:
    Cell *procedure;
    Cell *frame;
    Cell *constant_pool;
    Cell *input_port;
    Cell *output_port;
    Cell *value;
    const char *pc;

    const char *saved_pc;
    Cell *trampoline_frame;
    long nargs;

    Register(VirtualMachine &vm,
	     Cell *procedure, Cell *frame, Cell *constant_pool,
	     Cell *input_port, Cell *output_port);
    ~Register();
#if 0
    inline Cell *get_frame() const { return frame; }
    inline Cell *&get_frame_addr() { return frame; }
    inline void set_frame(Cell *frm);
#endif
  };
}


#include "libtscore/cell.h"
#include "libtscore/port.h"
#include "libtscore/reader.h"
#include "libtscore/tscore.h"
#include "libtscore/writer.h"

#include <vector>

#include <cassert>
#include <cstdarg>


using namespace std;
namespace LibTSCore
{
  class Cell;
  class TSCore;
  class VirtualMachine
  {
    friend class BytecodeProcedure;
    friend class Continuation;
    friend class Define;
    friend class DefineSyntax;
    friend class Lambda;
    friend class LetSyntax;
    friend class LetrecSyntax;
    friend class MacroSyntax;
    friend class NativeProcedure;
    friend class Procedure;
    friend class Quasiquote;
    friend class R5RSIOProcedure;
    friend class R5RSMiscProcedure;
    friend class Reader;
    friend class Register;
    friend class Syntax;
    friend class SyntaxRules;
    friend class TSCore;
    friend class TSCoreProcedure;
    friend class Unquote;
    friend class UnquoteSplicing;
    friend class VMStackRoot;
  private:
    TSCore &core;
    MemorySystem &memory_system;
    Reader reader;
    OutputStringPort error_buffer;


    Cell *continuation_cache;

    Cell *toplevel_constant_pool;
    Cell *temporary_frame;	// used for application from C++ world.

    bool tracing;
    bool bytecompile_tracing;
    bool bytecode_tracing;


    inline Cell *get_toplevel_environment() const;
    inline Cell *get_toplevel_frame() const;

    inline Cell *load_toplevel_variable(long slot) const
    { return get_toplevel_frame()->load_toplevel_variable(slot); }
    inline void store_toplevel_variable(long slot, Cell *value)
    {
      get_toplevel_frame()->store_toplevel_variable(get_toplevel_environment(),
						    slot, value);
    }

    inline const Cell *lookup_toplevel_binding(const Cell *symbol);
    inline const Cell *bind_toplevel_variable(const Cell *symbol);

    inline Cell *environment_new(Cell *parent)
    { return Cell::list_environment_new(memory_system, parent); }
    inline Cell *frame_new(Cell *parent, long size)
    { return Cell::scheme_frame_new(memory_system, parent, size); }

    inline Cell *mk_current_continuation(Register &context)
    {
      Cell *continuation = continuation_cache;
      continuation_cache = Cell::nil();
      if (continuation == Cell::nil())
	continuation = memory_system.get_cell();
      else
	memory_system.write_barrier(continuation);
      continuation->mk_continuation(context.procedure, context.frame,
				    context.pc);
      return continuation;
    }
    inline void recycle_continuation(Cell *continuation)
    { continuation_cache = continuation; }

    inline Cell *mk_current_continuation_full(Register &context)
    {
      return Cell::continuation_new(memory_system,
				    context.procedure, context.frame,
				    context.pc,
				    0,	// reserved for `dynamic-wind'
				    context.input_port, context.output_port);
    }

    Cell *unbound_variable_error(const char *name);
    // This is only for car/cdr at this time.
    Cell *type_mismatch_error(const char *name, Cell *val);
    Cell *application_error(Cell *val);

    inline OutputStringPort *get_error_buffer() { return &error_buffer; }

    inline Cell *mk_closure(Cell *closure_template, Cell *frame)
    {
      Cell *closure = memory_system.get_cell(closure_template, frame);
      return closure->mk_closure(closure_template, frame);
    }

    Cell *apply(Cell *procedure, Register &caller_context,
		unsigned long args_pointer, unsigned long nargs);


    void print_bytecode(Register &state);

    struct OpcodeInfo
    {
      const char *name;
      long length;
    };

    static const OpcodeInfo opcode_info[];
    static inline const char *get_opcode_name(unsigned char code)
    { return opcode_info[static_cast<long>(code)].name; }
    static inline long get_opcode_length(unsigned char code)
    { return opcode_info[static_cast<long>(code)].length; }


  public:
    VirtualMachine(TSCore &tscore, long temporary_frame_size = 28);
    ~VirtualMachine();

    static void print_bytecode(Port *port,
			       const char *from, Cell *constant_pool);
    static const char *get_next_pc(const char *current_pc);

    Cell *eval(Register &context, Cell *expression);

    void read_eval_print(Cell *input_port, Cell *output_port,
			 const char *prompt);

    inline const char *get_error_message() const
    { return error_buffer.get_output_string(); }

    inline Reader *get_reader() { return &reader; }


    inline Cell *mk_cons(Cell *&a, Cell *&d, bool immutable)
    { return memory_system.get_cons(a, d, immutable); }
    inline Cell *set_car(Cell *pair, Cell *value)
    { return memory_system.set_car(pair, value); }
    inline Cell *set_cdr(Cell *pair, Cell *value)
    { return memory_system.set_cdr(pair, value); }
    inline Cell *set_vector_element(Cell *vec, long index, Cell *elm)
    { return memory_system.set_vector_element(vec, index, elm); }

    inline Cell *mk_integer(long val)
    {
      Cell *x = Cell::mk_small_integer(val);
      if (x != NULL) return x;
      return memory_system.get_cell()->mk_integer(val);
    }

    inline Cell *mk_real(double val)
    { return memory_system.get_cell()->mk_real(val); }

    inline Cell *mk_number(Number &number)
    {
      if (number.is_fixnum())
	return mk_integer(number.get_ivalue0());
      return mk_real(number.get_rvalue0());
    }

    inline Cell *mk_counted_string(const char *str, size_t len)
    { return memory_system.get_counted_string(str, len); }
    inline Cell *mk_empty_string(size_t len, char fill)
    { return memory_system.get_empty_string(len, fill); }
    inline Cell *mk_string(const char *str)
    { return mk_counted_string(str, strlen(str)); }
    inline Cell *mk_static_string(const char *str)
    { return memory_system.get_static_string(str); }
    inline Cell *mk_substring_copy(Cell *string, long start, long end)
    { return memory_system.get_substring_copy(string, start, end); }
    inline Cell *mk_substring_reference(Cell *string, long start, long end)
    { return memory_system.get_substring_reference(string, start, end); }


    // This may return string cell that explains the error instead of
    // regexp cell if compilation of regexp fails.
    Cell *mk_regexp(Cell *rx_string, bool ignore_case = false);

    inline Cell *mk_time(long time_type, long nanosecond, long second)
    { return memory_system.get_cell()->mk_time(time_type, nanosecond, second); }

#ifdef HAVE_TIME_H
    inline Cell *mk_date(const struct tm *value, long nano, long zone_offset)
    { return memory_system.get_cell()->mk_date(value, nano, zone_offset); }
#endif

    inline Cell *mk_character(long c)
    { return Cell::mk_character(c); }
    inline Cell *mk_closure_template(Cell *constant_pool, Cell *code,
				     BytecodeProcedure *bytecode)
    {
      Cell *cell = memory_system.get_cell(constant_pool, code);
      return cell->mk_closure_template(constant_pool, code, bytecode);
    }
#if 0
    inline Cell *mk_promise(Cell *&code, Cell *&env)
    { return get_cons(code, env, false)->mk_promise(); }
#endif

    inline Cell *mk_vector(size_t len)
    { return memory_system.get_vector(len); }
    inline Cell *mk_vector(size_t len, Cell *&fill)
    { return memory_system.get_vector(len, fill); }

    inline Cell *mk_port(Port *port)
    { return memory_system.get_cell()->mk_port(port); }

    Cell *mk_input_file_port(const char *name);
    Cell *mk_input_file_port(FILE *file, bool interactive = false);
    Cell *mk_output_file_port(const char *name);
    Cell *mk_output_file_port(FILE *file);

    Cell *mk_input_string_port(Cell *str);
    Cell *mk_output_string_port();

    inline Cell *mk_foreign_object(void *pointer,
				   void (*destructor)(void *) = NULL)
    {
      if (destructor == NULL)
	{
	  Cell *cell = Cell::mk_foreign_object(pointer);
	  if (cell != NULL)
	    return cell;
	}
      return memory_system.get_cell()->mk_foreign_object(pointer, destructor);
    }

    inline Cell *cons(Cell *&a, Cell *&d = Cell::nil0())
    { return mk_cons(a, d, false); }
    inline Cell *immutable_cons(Cell *&a, Cell *&d = Cell::nil0())
    { return mk_cons(a, d, true); }

    inline MemorySystem &get_memory_system() { return memory_system; }

    Cell *apply(Cell *procedure, Cell *args = Cell::nil());
    Cell *applyf(Cell *procedure, const char *fmt, va_list ap);

    inline bool set_tracing(bool enable)
    {
      bool prev = tracing;
      tracing = enable;
      return prev;
    }
    inline bool get_tracing(Register &context) const
    { return tracing && context.output_port->is_output_port(); }

    inline bool set_bytecompile_tracing(bool enable)
    {
      bool prev = bytecompile_tracing;
      bytecompile_tracing = enable;
      return prev;
    }
    inline bool get_bytecompile_tracing(Register &context) const
    { return bytecompile_tracing && context.output_port->is_output_port(); }

    inline bool set_bytecode_tracing(bool enable)
    {
      bool prev = bytecode_tracing;
      bytecode_tracing = enable;
      return prev;
    }
    inline bool get_bytecode_tracing(Register &context) const
    { return bytecode_tracing && context.output_port->is_output_port(); }
  };

#define _VIRTUAL_MACHINE_DEFINED_	1


  enum TSVMOpcode
  {
    OP_NOT_USED = 0,

    // Load very small integer, -1, 0, 1, 2, 3 or 4, into `value'.
    OP_ICONST_M1,
#define OPS_ICONST_M1	"\001"
    OP_ICONST_0,
#define OPS_ICONST_0	"\002"
    OP_ICONST_1,
#define OPS_ICONST_1	"\003"
    OP_ICONST_2,
#define OPS_ICONST_2	"\004"
    OP_ICONST_3,
#define OPS_ICONST_3	"\005"
    OP_ICONST_4,
#define OPS_ICONST_4	"\006"


    // Load small integer value given as immediate into `value'.
    // Immediates are stored in little endian order within bytecode.
    // Postfixes, I1, I2, I4 and I8 shows the byte length of immediate
    // value.  Each immediate values are treated as `signed'.
    // OP_ICONST_I8 only used on ISA having 8 bytes `long'.
    OP_ICONST_I1,	// imm8
#define OPS_ICONST_I1	"\007"
    OP_ICONST_I2,	// imm16
#define OPS_ICONST_I2	"\010"
    OP_ICONST_I4,	// imm32
#define OPS_ICONST_I4	"\011"
    OP_ICONST_I8,	// imm64
#define OPS_ICONST_I8	"\012"


    // Load well known constant, (), #t or #f, into `value'.
    OP_CONST_NULL,
#define OPS_CONST_NULL	"\013"
    OP_CONST_T,
#define OPS_CONST_T	"\014"
    OP_CONST_F,
#define OPS_CONST_F	"\015"


    // Load constant value within constant pool specified by small index,
    // 0, 1, 2 or 3, into `value'.
    OP_CONST_P_0,
#define OPS_CONST_P_0	"\016"
    OP_CONST_P_1,
#define OPS_CONST_P_1	"\017"
    OP_CONST_P_2,
#define OPS_CONST_P_2	"\020"
    OP_CONST_P_3,
#define OPS_CONST_P_3	"\021"


    // Load constant value within constant pool specified by index given as
    // immediate into `value'.  Immediates are stored in little endian
    // order within bytecode.
    // Postfixes, P1, P2 and P4 shows the byte length of immediate
    // index.  Each immediate values are treated as `unsigned'.
    // P2 and P4 will be only rarely used we think, just reserved.
    OP_CONST_P1,	// index8
#define OPS_CONST_P1	"\022"
    OP_CONST_P2,	// index16
#define OPS_CONST_P2	"\023"
    OP_CONST_P4,	// index32
#define OPS_CONST_P4	"\024"


    // Load value within local variable slot specified by very frequently
    // used slot number, 0, 1, 2, 3, 4, 5, 6 and 7 into `value'.
    OP_LOAD_L_0,
#define OPS_LOAD_L_0	"\025"
    OP_LOAD_L_1,
#define OPS_LOAD_L_1	"\026"
    OP_LOAD_L_2,
#define OPS_LOAD_L_2	"\027"
    OP_LOAD_L_3,
#define OPS_LOAD_L_3	"\030"
    OP_LOAD_L_4,
#define OPS_LOAD_L_4	"\031"
    OP_LOAD_L_5,
#define OPS_LOAD_L_5	"\032"
    OP_LOAD_L_6,
#define OPS_LOAD_L_6	"\033"
    OP_LOAD_L_7,
#define OPS_LOAD_L_7	"\034"


    // Store value within `value' into local variable slot specified by
    // very frequently used slot number, 0, 1, 2, 3, 4, 5, 6 and 7.
    OP_STORE_L_0,
#define OPS_STORE_L_0	"\035"
    OP_STORE_L_1,
#define OPS_STORE_L_1	"\036"
    OP_STORE_L_2,
#define OPS_STORE_L_2	"\037"
    OP_STORE_L_3,
#define OPS_STORE_L_3	"\040"
    OP_STORE_L_4,
#define OPS_STORE_L_4	"\041"
    OP_STORE_L_5,
#define OPS_STORE_L_5	"\042"
    OP_STORE_L_6,
#define OPS_STORE_L_6	"\043"
    OP_STORE_L_7,
#define OPS_STORE_L_7	"\044"


    // Load value within local variable slot specified by immediate
    // slot number into `value'.  Immediates are stored in little endian
    // order within bytecode.
    // Postfixes, L1, L2 and L4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // L2 and L4 will be only rarely used we think, just reserved.
    OP_LOAD_L1,		// slot8
#define OPS_LOAD_L1	"\045"
    OP_LOAD_L2,		// slot16
#define OPS_LOAD_L2	"\046"
    OP_LOAD_L4,		// slot32
#define OPS_LOAD_L4	"\047"


    // Store value within `value' into local variable slot specified by
    // immediate slot number.  Immediates are stored in little endian
    // order within bytecode.
    // Postfixes, L1, L2 and L4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // L2 and L4 will be only rarely used we think, just reserved.
    OP_STORE_L1,	// slot8
#define OPS_STORE_L1	"\050"
    OP_STORE_L2,	// slot16
#define OPS_STORE_L2	"\051"
    OP_STORE_L4,	// slot32
#define OPS_STORE_L4	"\052"


    // Load value within variable slot within outer environment specified
    // by immediate count and slot number into `value'.  Immediates are
    // stored in little endian order within bytecode.
    // `env8' specifies the environment `env8' times outer environment
    // from current environment in which the variable slot located.
    // Postfixes, G1, G2 and G4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // G2 and G4 will be only rarely used we think, just reserved.
    OP_LOAD_G1,		// env8 slot8
#define OPS_LOAD_G1	"\053"
    OP_LOAD_G2,		// env8 slot16
#define OPS_LOAD_G2	"\054"
    OP_LOAD_G4,		// env8 slot32
#define OPS_LOAD_G4	"\055"


    // Store value within `value' into variable slot within outer environment
    // specified by immediate count and slot number.  Immediates are stored
    // in little endian order within bytecode.
    // `env8' specifies the environment `env8' times outer environment
    // from current environment in which the variable slot located.
    // Postfixes, G1, G2 and G4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // G2 and G4 will be only rarely used we think, just reserved.
    OP_STORE_G1,	// env8 slot8
#define OPS_STORE_G1	"\056"
    OP_STORE_G2,	// env8 slot16
#define OPS_STORE_G2	"\057"
    OP_STORE_G4,	// env8 slot32
#define OPS_STORE_G4	"\060"


    // Load value within variable slot of top level environment specified
    // by immediate slot number into `value'.  Immediates are stored in
    // little endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // T4 will be only rarely used we think, just reserved.
    OP_LOAD_T1,		// slot8
#define OPS_LOAD_T1	"\061"
    OP_LOAD_T2,		// slot16
#define OPS_LOAD_T2	"\062"
    OP_LOAD_T4,		// slot32
#define OPS_LOAD_T4	"\063"


    // Store value within `value' into variable slot of top level environment
    // specified by immediate slot number.  Immediates are stored in little
    // endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate
    // slot number.  Each immediate slot numbers are treated as `unsigned'.
    // T4 will be only rarely used we think, just reserved.
    OP_STORE_T1,	// slot8
#define OPS_STORE_T1	"\064"
    OP_STORE_T2,	// slot16
#define OPS_STORE_T2	"\065"
    OP_STORE_T4,	// slot32
#define OPS_STORE_T4	"\066"


    // Load value within variable specified by symbol given by constant
    // pool whose indices are given as immediates into `value'.  Immediates
    // are stored in little endian order within bytecode.
    // Postfixes, F1, F2 and F4 shows the byte length of immediate index.
    // Each immediate values are treated as `unsigned'.
    // These bytecodes are for `free variable's that aren't known to be
    // bound at bytecompilation time.
    OP_LOAD_F1,		// index8
#define OPS_LOAD_F1	"\067"
    OP_LOAD_F2,		// index16
#define OPS_LOAD_F2	"\070"
    OP_LOAD_F4,		// index32
#define OPS_LOAD_F4	"\071"


    // Store value within `value' into variable specified by symbol given
    // by constant pool whose indices are given as immediates.  Immediates
    // are stored in little endian order within bytecode.
    // Postfixes, F1, F2 and F4 shows the byte length of immediate index.
    // Each immediate values are treated as `unsigned'.
    // These bytecodes are for `free variable's that aren't known to be
    // bound at bytecompilation time.
    OP_STORE_F1,	// index8
#define OPS_STORE_F1	"\072"
    OP_STORE_F2,	// index16
#define OPS_STORE_F2	"\073"
    OP_STORE_F4,	// index32
#define OPS_STORE_F3	"\074"


    // If the value within `value' is considered as `true' in Scheme semantics,
    // i.e. not #f, flow of control is jumped with immediate offsets.
    // Immediates are stored in little endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate offset.
    // Each immediate values are treated as `signed'.
    // T2 and T4 will be only rarely used we think, just reserved.
    OP_IF_T1,		// off8
#define OPS_IF_T1	"\075"
    OP_IF_T2,		// off16
#define OPS_IF_T2	"\076"
    OP_IF_T4,		// off32
#define OPS_IF_T4	"\077"


    // If the value within `value' is considered as `false' in Scheme
    // semantics, i.e. #f, flow of control is jumped with immediate offsets.
    // Immediates are stored in little endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate offset.
    // Each immediate values are treated as `signed'.
    // T2 and T4 will be only rarely used we think, just reserved.
    OP_IF_F1,		// off8
#define OPS_IF_F1	"\100"
    OP_IF_F2,		// off16
#define OPS_IF_F2	"\101"
    OP_IF_F4,		// off32
#define OPS_IF_F4	"\102"


    // Unconditionally jumps the flow of control with immediate offsets.
    // Immediates are stored in little endian order within bytecode.
    // Postfixes, T1, T2 and T4 shows the byte length of immediate offset.
    // Each immediate values are treated as `signed'.
    // T2 and T4 will be only rarely used we think, just reserved.
    OP_GOTO1,		// off8
#define OPS_GOTO1	"\103"
    OP_GOTO2,		// off16
#define OPS_GOTO2	"\104"
    OP_GOTO4,		// off32
#define OPS_GOTO4	"\105"


    // Get the value of (car `value') into `value'.
    OP_CAR,
#define OPS_CAR		"\106"


    // Get the value of (cdr `value') into `value'.
    OP_CDR,
#define OPS_CDR		"\107"


    // OP_APPLY[124] and OP_CALL_CC[124] are almost same but OP_APPLY[124]
    // refer the implicit continuation argument from caller's first
    // argument, i.e. caller's continuation.  On the other hand,
    // OP_CALL_CC[124] fill the implicit continuation argument with
    // nil to show the fact that it isn't a tail-call.  Essentially,
    // these instructions are macro that combine sequence of bytecode
    // like below:
    // { OP_LOAD_L_0, OP_STORE_L1, args_pointer,
    //   OP_LOAD_L1, proc_pointer, OP_CALL1 } for OP_APPLY[124]
    // { OP_CONST_NULL, OP_STORE_L1, args_pointer,
    //   OP_LOAD_L1, proc_pointer, OP_CALL1 } for OP_CALL_CC[124]
    //
    // NOTE: Since we don't actually have OP_CALL[124], these aren't usable.

    // Applies procedure with computed arguments on frame.
    OP_APPLY1,		// off8 nargs8
#define OPS_APPLY1	"\110"
    OP_APPLY2,		// off16 nargs16
#define OPS_APPLY2	"\111"
    OP_APPLY4,		// off32 nargs32
#define OPS_APPLY4	"\112"


    // Applies procedure with computed arguments on frame with current
    // continuation.
    OP_CALL_CC1,	// off8 nargs8
#define OPS_CALL_CC1	"\113"
    OP_CALL_CC2,	// off16 nargs16
#define OPS_CALL_CC2	"\114"
    OP_CALL_CC4,	// off32 nargs32
#define OPS_CALL_CC4	"\115"


    // Applies implicit continuation introduced by bytecompiler to the
    // current context.
    // In our convention, the `continuation' is always given as the
    // first argument of all procedures.
    OP_APPLY_CONTINUATION0,
#define OPS_APPLY_CONTINUATION0	"\116"


    // Creates a closure from closure template within `value' and the
    // current frame.
    OP_LAMBDA,
#define OPS_LAMBDA	"\117"


    // Applies procedure with prepared arguments on trampoline frame
    // as they are in current frame.
    OP_TRAMPOLINE,
#define OPS_TRAMPOLINE	"\120"


    // Returns the `value' to the caller and terminates current
    // evaluation cycle.
    OP_META_RETURN,
#define OPS_META_RETURN	"\121"


    // no operation
    OP_NOP
#define OPS_NOP		"\122"
  };


  class VMStackRoot
  {
  private:
    StackRoot stack_root;

  public:
    inline VMStackRoot(VirtualMachine &vm, Cell *cell)
      : stack_root(vm.memory_system, cell) {}
    inline VMStackRoot(VirtualMachine &vm, const Cell *cell)
      : stack_root(vm.memory_system, const_cast<Cell *>(cell)) {}
    inline ~VMStackRoot() {}
    inline operator Cell *&() { return stack_root; }
    inline Cell *operator->() { return stack_root; }
    inline VMStackRoot &operator=(Cell *cell)
    { stack_root = cell; return *this; }
    inline VMStackRoot &operator=(VMStackRoot &root)
    { stack_root = static_cast<Cell *>(root); return *this; }
  };
}

#endif

#endif	/* _LIBTSCORE_VIRTUAL_MACHINE_H_ */

#if defined(_TSCORE_DEFINED_) && defined(_VIRTUAL_MACHINE_DEFINED_) && !defined(_VIRTUAL_MACHINE_INLINE_FUNC_DEFINED_)
namespace LibTSCore
{
  inline Cell *VirtualMachine::get_toplevel_environment() const
  { return core.get_interaction_environment(); }
  inline Cell *VirtualMachine::get_toplevel_frame() const
  { return core.get_interaction_toplevel_frame(); }
  inline const Cell *VirtualMachine::lookup_toplevel_binding(const Cell *symbol)
  { return core.get_interaction_environment()->lookup(symbol, false); }
  inline const Cell *VirtualMachine::bind_toplevel_variable(const Cell *symbol)
  { return core.get_interaction_environment()->bind_variable(symbol); }
}

#define _VIRTUAL_MACHINE_INLINE_FUNC_DEFINED_	1
#endif

#if 0
#if defined(_CELL_DEFINED_) && !defined(_REGISTER_INLINE_FUNC_DEFINED_)
namespace LibTSCore
{
  inline void Register::set_frame(Cell *frm)
  {
    assert(frm->is_scheme_frame() || frm == Cell::nil());
    frame = frm;
  }
}

#define _REGISTER_INLINE_FUNC_DEFINED_ 1
#endif
#endif
