/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bookmarks_ui.cc,v 1.27 2008/10/25 08:45:56 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "bookmarks_ui.h"

#include "libochusha/bookmarks.h"

#include "controller.h"
#include "session_client.h"
#include "tree_view.h"

#include "libochushagtk/stock.h"

#include <cassert>

#include <gtk/gtk.h>


namespace Ochusha
{


BookmarksUI::BookmarksUI(Controller &controller)
  : TreeUIImpl(controller, new Bookmarks()),
    action_group(create_action_group()),
    bookmarks(dynamic_cast<Bookmarks *>(get_root_folder())),
    my_bookmark_folder(NULL)
{
  assert(bookmarks != NULL);


  GDK_THREADS_LEAVE();

  bookmarks->read_bookmarks_xml(controller.get_repository());

  GDK_THREADS_ENTER();


  GtkWidget *view_widget = master_view->get_tree_view();
  g_signal_connect(view_widget, "start_interactive_search",
		   reinterpret_cast<GCallback>(start_interactive_search_cb),
		   this);
}


Folder *
BookmarksUI::folder_new(const char *name)
{
  Folder *folder = get_root_folder()->folder_new();
  folder->set_name(name);

  new_element(get_root_folder(), folder);

  return folder;
}


void
BookmarksUI::bookmark(Folder *folder, const char *url, const char *name)
{
  g_return_if_fail(folder != NULL);
  g_return_if_fail(url != NULL);
  g_return_if_fail(name != NULL);

  new_element(folder, new BookmarkEntry(name, url));
}


void
BookmarksUI::insert_my_bookmark(BBSThread *thread, int res_num,
				const char *name)
{
  assert(thread != NULL);
  assert(name != NULL);

  remove_my_bookmark(thread);

  if (res_num == 0)
    return;

  char *url = thread->get_url(res_num);
  g_return_if_fail(url != NULL);

  new_element(my_bookmark_folder, new BookmarkEntry(name, url));

  free(url);
}


void
BookmarksUI::remove_my_bookmark(BBSThread *thread)
{
  g_return_if_fail(thread != NULL);
  g_return_if_fail(thread->get_board() != NULL);

  list<smart_ptr<TreeElement> >::const_iterator iter
    = my_bookmark_folder->get_elements_begin();
  list<smart_ptr<TreeElement> >::const_iterator end_iter
    = my_bookmark_folder->get_elements_end();

  while (iter != end_iter)
    {
      TreeElement *element = *iter++;
      if (!element->is_folder())
	{
	  BookmarkEntry *entry = dynamic_cast<BookmarkEntry *>(element);
	  assert(entry != NULL);
	  const char *url = entry->get_url();
	  BulletinBoard *board = controller.get_bbs_table().guess_board(url);
	  if (board != thread->get_board())
	    continue;

	  BBSThread *thr = board->lookup_bbs_thread_by_url(url);
	  if (thread != thr)
	    continue;

	  remove_element(my_bookmark_folder, entry);
	  return;
	}
    }
}


void
BookmarksUI::bookmark(const char *url, const char *name)
{
  g_return_if_fail(url != NULL);

  add_element(new BookmarkEntry(name != NULL ? name : _("New Bookmark"), url),
	      _("Add Bookmark"));
}


BookmarksUI::~BookmarksUI()
{
  g_object_unref(action_group);
}


void
BookmarksUI::start_interactive_search_cb(OchushaTreeSelectorView *view,
					 BookmarksUI *ui)
{
  ui->master_view->setup_search(_("Find in Bookmarks"));
}


void
BookmarksUI::save_bookmarks()
{
  bookmarks->write_bookmarks_xml(controller.get_repository());
}


GtkActionGroup *
BookmarksUI::create_action_group()
{
  static GtkActionEntry entries[] = {
    {
      "ManageBookmarks", NULL,
      N_("_Manage Bookmarks"), NULL,
      N_("Manage Bookmarks"),
      reinterpret_cast<GCallback>(manage_bookmarks_cb)
    }
  };

  action_group = gtk_action_group_new("BookmarksActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group, entries,
			       G_N_ELEMENTS(entries), this);

#if 1	// not implemented yet.
  GtkAction *action = gtk_action_group_get_action(action_group,
						  "ManageBookmarks");
  if (action != NULL)
    gtk_action_set_sensitive(action, FALSE);
#endif

  gtk_ui_manager_insert_action_group(controller.get_ui_manager(),
				     action_group, 0);
  return action_group;
}


void
BookmarksUI::manage_bookmarks_cb(GtkAction *action, BookmarksUI *ui)
{
}


const gchar *
BookmarksUI::get_element_icon_stock_id(TreeElement *element, gboolean expanded)
{
  if (element->is_folder())
    return TreeUIImpl::get_element_icon_stock_id(element, expanded);

  return OCHUSHA_STOCK_BOOKMARK;
}


bool
BookmarksUI::show_all() const
{
  return true;
}


static Folder *
find_my_bookmark_folder(Folder *folder)
{
  assert(folder != NULL);

  list<smart_ptr<TreeElement> >::const_iterator iter
    = folder->get_elements_begin();
  list<smart_ptr<TreeElement> >::const_iterator end_iter
    = folder->get_elements_end();

  while (iter != end_iter)
    {
      TreeElement *element = *iter++;
      if (element->is_folder())
	{
	  Folder *fldr = dynamic_cast<Folder *>(element);
	  assert(fldr != NULL);
	  if (fldr->get_folder_type() == FOLDER_TYPE_OCHUSHA_MY_BOOKMARK)
	    return fldr;
	  fldr = find_my_bookmark_folder(fldr);
	  if (fldr != NULL)
	    return fldr;
	}
    }

  return NULL;	// not found
}


void
BookmarksUI::reconstruct()
{
  my_bookmark_folder = find_my_bookmark_folder(get_root_folder());
  if (static_cast<Folder *>(my_bookmark_folder) == NULL)
    {
      my_bookmark_folder = bookmarks->RootFolder::folder_new();
      my_bookmark_folder->set_folder_type(FOLDER_TYPE_OCHUSHA_MY_BOOKMARK);
      my_bookmark_folder->set_name(_("My Bookmark"));
      get_root_folder()->prepend(my_bookmark_folder);

      bookmarks->write_bookmarks_xml(controller.get_repository());
    }

  TreeUIImpl::reconstruct();
}


const char *
BookmarksUI::get_popup_menu_root()
{
  return "/BookmarksPopupMenu";
}


const char *
BookmarksUI::get_folder_tree_popup_menu_root()
{
  return "/FolderTreePopupMenu";
}


GtkWidget *
BookmarksUI::get_menu_widget()
{
  GtkWidget *menu_widget
    = gtk_ui_manager_get_widget(controller.get_ui_manager(),
				"/MainMenu/BookmarksMenu");
  if (menu_widget == NULL)
    {
      fprintf(stderr, "Cannot find BookmarksMenu\n");
      return NULL;
    }

  return menu_widget;
}


void
BookmarksUI::open(TreeElement *element, bool in_tab, bool with_browser,
		  bool select_page)
{
  g_return_if_fail(!element->is_separator());

  if (!element->is_folder())
    {
      BookmarkEntry *entry = dynamic_cast<BookmarkEntry *>(element);
      g_return_if_fail(entry != NULL);

      controller.open(entry->get_url(), in_tab, with_browser, select_page);
      return;
    }

  if (!in_tab)
    return;

  Folder *folder = dynamic_cast<Folder *>(element);
  g_return_if_fail(folder != NULL);

  list<smart_ptr<TreeElement> >::const_iterator iter
    = folder->get_elements_begin();
  list<smart_ptr<TreeElement> >::const_iterator end_iter
    = folder->get_elements_end();

  while (iter != end_iter)
    {
      TreeElement *elm = *iter++;
      if (!elm->is_folder() && !elm->is_separator())
	{
	  BookmarkEntry *entry = dynamic_cast<BookmarkEntry *>(elm);
	  if (entry != NULL)
	    controller.open(entry->get_url(), true, false, select_page);
	}
    }
}


void
BookmarksUI::bookmark(TreeElement *element)
{
  BookmarkEntry *entry = dynamic_cast<BookmarkEntry *>(element);
  g_return_if_fail(entry != NULL);
  controller.bookmark(entry->get_url(), entry->get_name());
}


const char *
BookmarksUI::get_url(TreeElement *element)
{
  if (!element->is_folder())
    {
      BookmarkEntry *entry = dynamic_cast<BookmarkEntry *>(element);
      if (entry != NULL)
	return entry->get_url();
    }
  return "";	// ޤ
}


const char *
BookmarksUI::get_ui_name()
{
  return "BookmarksView";
}


class BookmarkProperties
{
private:
  BookmarksUI &ui;
  smart_ptr<BookmarkEntry> item;

  GtkEntry *name_entry;
  GtkEntry *url_entry;
  GtkTextBuffer *description_buffer;

  ~BookmarkProperties();

  static void dialog_destroyed_cb(GtkWidget *widget,
				  BookmarkProperties *properties);


public:
  BookmarkProperties(BookmarksUI &ui, GtkDialog *dialog, BookmarkEntry *item);

  inline BookmarkEntry *get_item() { return item; }

  void process_response(GtkDialog *dialog, int response_id,
			BookmarkEntry *item);
};


BookmarkProperties::BookmarkProperties(BookmarksUI &ui, GtkDialog *dialog,
				       BookmarkEntry *item)
  : ui(ui),
    item(item)
{
  GtkBox *outer_vbox = GTK_BOX(dialog->vbox);

  GtkWidget *widget = gtk_table_new(3, 2, FALSE);
  gtk_widget_show(widget);
  gtk_box_pack_start(outer_vbox, widget, TRUE, TRUE, 10);

  GtkTable *table = GTK_TABLE(widget);



  GtkWidget *label = gtk_label_new_with_mnemonic(_("_Name:"));
  gtk_widget_show(label);
  gtk_table_attach(table, label, 0, 1, 0, 1, GTK_SHRINK, GTK_SHRINK, 0, 0);


  GtkWidget *entry = gtk_entry_new();
  gtk_widget_show(entry);
  gtk_table_attach_defaults(table, entry, 1, 2, 0, 1);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), entry);

  name_entry = GTK_ENTRY(entry);

  const char *name = item->get_name();
  gtk_entry_set_text(name_entry, name != NULL ? name : _("Anonymous Item"));
  gtk_entry_set_width_chars(name_entry, 36);
  gtk_entry_set_activates_default(name_entry, TRUE);



  label = gtk_label_new_with_mnemonic(_("_URL:"));
  gtk_widget_show(label);
  gtk_table_attach(table, label, 0, 1, 1, 2, GTK_SHRINK, GTK_SHRINK, 0, 0);


  entry = gtk_entry_new();
  gtk_widget_show(entry);
  gtk_table_attach_defaults(table, entry, 1, 2, 1, 2);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), entry);

  url_entry = GTK_ENTRY(entry);

  const char *url = item->get_url();
  if (url != NULL)
    gtk_entry_set_text(url_entry, url);
  gtk_entry_set_width_chars(url_entry, 48);
  gtk_entry_set_activates_default(url_entry, TRUE);



  GtkWidget *vbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox);
  gtk_table_attach(table, vbox, 0, 1, 2, 3, GTK_SHRINK,
		   static_cast<GtkAttachOptions>(GTK_EXPAND | GTK_FILL), 0, 3);


  label = gtk_label_new_with_mnemonic(_("_Description:"));
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 3);


  description_buffer = gtk_text_buffer_new(NULL);
  const char *desc = item->get_description();
  if (desc != NULL)
    gtk_text_buffer_insert_at_cursor(description_buffer, desc, -1);

  GtkWidget *win = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  GtkWidget *view = gtk_text_view_new_with_buffer(description_buffer);
  gtk_widget_show(view);
  gtk_container_add(GTK_CONTAINER(win), view);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), view);

  gtk_widget_show(win);
  gtk_table_attach(table, win, 1, 2, 2, 3,
		static_cast<GtkAttachOptions>(GTK_EXPAND|GTK_SHRINK|GTK_FILL),
		static_cast<GtkAttachOptions>(GTK_EXPAND|GTK_SHRINK|GTK_FILL),
		2, 2);
}


BookmarkProperties::~BookmarkProperties()
{
  g_object_unref(description_buffer);
}


void
BookmarkProperties::dialog_destroyed_cb(GtkWidget *widget,
					BookmarkProperties *properties)
{
  delete properties;
}


void
BookmarkProperties::process_response(GtkDialog *dialog, int response_id,
				     BookmarkEntry *item)
{
  g_return_if_fail(get_item() == item);

  if (response_id != GTK_RESPONSE_OK)
    return;


  const char *name = gtk_entry_get_text(name_entry);
  const char *prev_name = item->get_name();
  if (prev_name == NULL || strcmp(name, prev_name) != 0)
    item->set_name(name);


  const char *url = gtk_entry_get_text(url_entry);
  const char *prev_url = item->get_url();
  if (*url != '\0' && (prev_url == NULL || strcmp(url, prev_url) != 0))
    item->set_url(url);


  GtkTextIter iter;
  GtkTextIter end_iter;
  gtk_text_buffer_get_start_iter(description_buffer, &iter);
  gtk_text_buffer_get_end_iter(description_buffer, &end_iter);
  gchar *desc = gtk_text_buffer_get_text(description_buffer,
					 &iter, &end_iter, FALSE);
  item->set_description(desc);
  g_free(desc);
}


static GQuark properties_id
  = g_quark_from_static_string("BookmarksUI::BookmarkProperties");


void
BookmarksUI::process_response(GtkDialog *dialog, int response_id,
			      TreeElement *element)
{
  if (element->is_folder())
    return;	// ȤꤢեΥץѥƥ϶̤ۤȤ

  BookmarkProperties *properties
    = static_cast<BookmarkProperties *>(g_object_get_qdata(G_OBJECT(dialog),
							   properties_id));
  g_return_if_fail(properties != NULL);
  properties->process_response(dialog, response_id,
			       dynamic_cast<BookmarkEntry *>(element));
}


bool
BookmarksUI::construct_properties_contents(GtkDialog *dialog,
					   TreeElement *element)
{
  if (element->is_folder())
    return false;	// ȤꤢեΥץѥƥ϶̤ۤȤ

  BookmarkProperties *properties
    = new BookmarkProperties(*this, dialog,
			     dynamic_cast<BookmarkEntry *>(element));
  g_object_set_qdata(G_OBJECT(dialog), properties_id, properties);
  return true;
}


void
BookmarksUI::open_properties(TreeElement *element)
{
  const char *name = element->get_name();
  if (name == NULL)
    name = _("Anonymous Item");

  char *title;
  if (element->is_folder())
    title = g_strdup_printf(_("Bookmark Folder's Properties for \"%s\""), name);
  else
    title = g_strdup_printf(_("Bookmark's Properties for \"%s\""), name);

  // ʤǸʤ
  TreeUIImpl::open_properties(title, element);

  g_free(title);
}


bool
BookmarksUI::process_drop_data(Folder *src_folder, TreeElement *src_element,
			       Folder *drop_folder, TreeElement *drop_element,
			       GtkTreeViewDropPosition position, bool move)
{
  if (src_folder == drop_folder && src_element == drop_element)
    return false;	// ͭ͡ġ

  // src_element֥åޡΤΤǤʤURLdropΤƱ
  int folder_type = src_folder->get_folder_type();
  if (!(folder_type == FOLDER_TYPE_ROOT_FOLDER
	&& src_folder == get_root_folder())
      && folder_type != FOLDER_TYPE_BOOKMARKS_FOLDER
      && folder_type != FOLDER_TYPE_OCHUSHA_MY_BOOKMARK)
    {
      return process_drop_data(get_url(src_element), drop_folder, drop_element,
			       position);
    }
  Folder *dest_folder;
  if (position == GTK_TREE_VIEW_DROP_BEFORE
      || position == GTK_TREE_VIEW_DROP_AFTER
      || !drop_element->is_folder())
    dest_folder = drop_folder;
  else
    dest_folder = dynamic_cast<Folder *>(drop_element);
  g_return_val_if_fail(dest_folder != NULL, false);

  if (src_folder == dest_folder)
    move = true;

  if (move)
    return move_element(src_folder, src_element, drop_folder, drop_element,
			position);

  // 褺src_element->is_folder()ʾ̵ȿȤ롣
  if (src_element->is_folder())
    return false;

  smart_ptr<BookmarkEntry> entry = dynamic_cast<BookmarkEntry *>(src_element);
  g_return_val_if_fail(entry != NULL, false);

  BookmarkEntry *clone_entry
    = new BookmarkEntry(entry->get_name(), entry->get_url());
  clone_entry->set_description(entry->get_description());

  new_element(dest_folder, clone_entry);

  return true;
}


bool
BookmarksUI::process_drop_data(const char *uri, Folder *drop_folder,
			       TreeElement *drop_element,
			       GtkTreeViewDropPosition position)
{
  return false;
}


}
