/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: image_manager.cc,v 1.36 2008/10/25 08:50:23 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "image_manager.h"

#include "libochusha/constants.h"
#include "libochusha/http_headers.h"
#include "libochusha/misc_utils.h"
#include "libochusha/network_agent.h"
#include "libochusha/sigslot.h"
#include "libochusha/uri_utils.h"

#include "libochushagtk/stock.h"


#include <algorithm>
#include <list>

#include <cassert>

#ifdef HAVE_MMAP
#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif

#ifndef MAP_NOCORE
#define MAP_NOCORE 0
#endif

#ifndef MAP_NOSYNC
#define MAP_NOSYNC 0
#endif
#endif	// HAVE_MMAP

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <gtk/gtk.h>


namespace Ochusha
{

// #define DISABLE_SCHEME_REWRITER	1


#define OCHUSHA_A_BONE_IMAGE_TXT	"a-bone-image.txt"
#define OCHUSHA_A_BONE_SERVER_TXT	"a-bone-server.txt"
#define IMAGE_CACHE_TIMEOUT		30000	// 30
#define LOAD_IMAGE_CHUNKSIZE		262144	// åե뤫
						// 256KBĥǡɤࡣ
						// mmap()ǤʤΤߡ

#undef CHASE_REF_COUNTS
#undef VERBOSE_DESTRUCTOR

#define IMAGE_INFO_MAGIC_NUMBER		0xbabebabe

class ImageInfo
{
private:
  ImageManager &manager;
  char *url;

  GdkPixbufLoader *loader;

  int width;
  int height;

  GdkPixbufAnimation *animation;
  GdkPixbuf *pixbuf;

  list<GtkImage *> image_list;

  guint timeout_id;

  unsigned int magic;

  void update_pixbuf(GdkPixbufLoader *loader);
  

  static gboolean timeout_cb(ImageInfo *info);
  static void destroy_cb(GtkImage *image, ImageInfo *info);

  static void size_prepared_cb(GdkPixbufLoader *loader, int width, int height,
			       ImageInfo *info);
  static void area_prepared_cb(GdkPixbufLoader *loader, ImageInfo *info);
  static void area_updated_cb(GdkPixbufLoader *loader, int x, int y,
			      int width, int height, ImageInfo *info);
  static void closed_cb(GdkPixbufLoader *loader, ImageInfo *info);

public:
  ImageInfo(ImageManager &manager, const char *url, GdkPixbufLoader *loader,
	    GtkWidget *image, int width, int height);
  ~ImageInfo();

  void add_widget(GtkWidget *image);
  void a_bone();

  inline GdkPixbufLoader *get_loader() { return loader; }
  inline int get_width() { return width; }
  inline int get_height() { return height; }
  inline GdkPixbufAnimation *get_animation() { return animation; }
  inline GdkPixbuf *get_pixbuf() { return pixbuf; }
};


ImageInfo::ImageInfo(ImageManager &manager, const char *url,
		     GdkPixbufLoader *loader, GtkWidget *image,
		     int width, int height)
  : manager(manager),
    url(strdup(url)),
    loader(loader),
    width(width),
    height(height),
    animation(NULL),
    pixbuf(NULL),
    timeout_id(0),
    magic(IMAGE_INFO_MAGIC_NUMBER)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (loader != NULL)
    {
      g_object_ref(loader);
      g_signal_connect(loader, "size_prepared",
		       reinterpret_cast<GCallback>(size_prepared_cb), this);
      g_signal_connect(loader, "area_prepared",
		       reinterpret_cast<GCallback>(area_prepared_cb), this);
      g_signal_connect(loader, "area_updated",
		       reinterpret_cast<GCallback>(area_updated_cb), this);
      g_signal_connect(loader, "closed",
		       reinterpret_cast<GCallback>(closed_cb), this);
    }

  if (image != NULL)
    add_widget(image);
}


ImageInfo::~ImageInfo()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());

  magic = 0;
  free(url);
  if (timeout_id != 0)
    g_source_remove(timeout_id);
#if CHASE_REF_COUNTS
  fprintf(stderr, "ImageInfo(@%p) destructed\n", this);
  if (loader != NULL)
    {
      fprintf(stderr, "~ImageInfo() loader(@%p)->ref_count=%d\n",
	      loader, G_OBJECT(loader)->ref_count);
      g_object_unref(loader);
    }
  if (animation != NULL)
    {
      fprintf(stderr, "~ImageInfo() animation->ref_count=%d\n",
	      G_OBJECT(animation)->ref_count);
      g_object_unref(animation);
    }
  if (pixbuf != NULL)
    {
      fprintf(stderr, "~ImageInfo() pixbuf->ref_count=%d\n",
	      G_OBJECT(pixbuf)->ref_count);
      g_object_unref(pixbuf);
    }
#else
  if (loader != NULL)
    g_object_unref(loader);
  if (animation != NULL)
    g_object_unref(animation);
  if (pixbuf != NULL)
    g_object_unref(pixbuf);
#endif
}


gboolean
ImageInfo::timeout_cb(ImageInfo *info)
{
#if CHASE_REF_COUNTS
  fprintf(stderr, "ImageInfo::timeout_cb(%p)\n", info);
#endif
  // g_timeout_add()ΥХåGLibɳʤΤGDK_THREADS_ENTER()Ǥʤ
  GDK_THREADS_ENTER();

  assert(info->loader == NULL);

  info->timeout_id = 0;

  if (!info->image_list.empty())
    {
      GDK_THREADS_LEAVE();

      return FALSE;
    }

  GSList *info_list
    = static_cast<GSList *>(info->manager.image_info_table.lookup(info->url));
  info_list = g_slist_remove(info_list, info);
  if (info_list == NULL)
    info->manager.image_info_table.remove(info->url);
  else
    {
      ImageInfo *another_info = static_cast<ImageInfo *>(info_list->data);
      info->manager.image_info_table.insert(another_info->url, info_list);
    }
  delete info;

  GDK_THREADS_LEAVE();


  return FALSE;
}


void
ImageInfo::destroy_cb(GtkImage *image, ImageInfo *info)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());

#if CHASE_REF_COUNTS
  fprintf(stderr, "ImageInfo::destroy_cb(%p, %p)\n", image, info);
#endif

  info->image_list.erase(find(info->image_list.begin(),
			      info->image_list.end(), image));
  if (info->image_list.empty() && info->loader == NULL && info->timeout_id == 0)
    info->timeout_id = g_timeout_add(IMAGE_CACHE_TIMEOUT,
				     reinterpret_cast<GSourceFunc>(timeout_cb),
				     info);
#if CHASE_REF_COUNTS
  else
    {
      fprintf(stderr, "info->image_list.empty()=%d, info->loader=%p\n",
	      info->image_list.empty(), info->loader);
    }
#endif
}


void
ImageInfo::add_widget(GtkWidget *image)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  g_return_if_fail(image != NULL && GTK_IS_IMAGE(image));

  g_signal_connect(image, "destroy",
		   reinterpret_cast<GCallback>(destroy_cb), this);
  image_list.push_back(reinterpret_cast<GtkImage *>(image));

  if (timeout_id != 0)
    {
      g_source_remove(timeout_id);
      timeout_id = 0;
    }
}


void
ImageInfo::a_bone()
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  list<GtkImage *>::iterator iter = image_list.begin();
  list<GtkImage *>::iterator end_iter = image_list.end();
  while (iter != end_iter)
    gtk_image_set_from_stock(*iter++,
			     GTK_STOCK_DIALOG_WARNING, GTK_ICON_SIZE_DIALOG);
}


struct SetAnimationFunc
{
  GdkPixbufAnimation *animation;
  SetAnimationFunc(GdkPixbufAnimation *anim) : animation(anim) {}
  inline void operator()(GtkImage *image)
  {
    gtk_image_set_from_animation(image, animation);
  }
};


struct SetPixbufFunc
{
  GdkPixbuf *pixbuf;
  SetPixbufFunc(GdkPixbuf *pixbuf) : pixbuf(pixbuf) {}
  inline void operator()(GtkImage *image)
  {
    gtk_image_set_from_pixbuf(image, pixbuf);
  }
};


struct SetMissingImage
{
  inline void operator()(GtkImage *image)
  {
    gtk_image_set_from_stock(image, GTK_STOCK_MISSING_IMAGE,
			     GTK_ICON_SIZE_DIALOG);
  }
};


struct QueueDrawFunc
{
  inline void operator()(GtkImage *image)
  {
    gtk_widget_queue_draw(GTK_WIDGET(image));
  }
};


void
ImageInfo::update_pixbuf(GdkPixbufLoader *loader)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());

  assert(magic == IMAGE_INFO_MAGIC_NUMBER);

  if (animation != NULL || pixbuf != NULL)
    {
      g_assert(this->loader == NULL);
      for_each(image_list.begin(), image_list.end(),
	       QueueDrawFunc());
      goto skip;	// 
    }

  if (manager.is_a_boned(url))
    goto finish;	// ܡ󤵤Ƥ롣

  animation = gdk_pixbuf_loader_get_animation(loader);
  if (animation != NULL)
    {
      g_object_ref(animation);
      for_each(image_list.begin(), image_list.end(),
	       SetAnimationFunc(animation));
    }
  else
    {
      pixbuf = gdk_pixbuf_loader_get_pixbuf(loader);
      if (pixbuf != NULL)
	{
	  g_object_ref(pixbuf);
	  for_each(image_list.begin(), image_list.end(),
		   SetPixbufFunc(pixbuf));
	}
      else
	{
	  for_each(image_list.begin(), image_list.end(), SetMissingImage());
	}
    }

 finish:
  if (this->loader != NULL)
    {
#if CHASE_REF_COUNTS
      fprintf(stderr, "ImageInfo::update_pixbuf(): loader(@%p)->ref_count=%d\n",
	      loader, G_OBJECT(loader)->ref_count);
#endif
      g_object_unref(this->loader);
      this->loader = NULL;
    }

 skip:
  if (image_list.empty() && timeout_id == 0)
    {
      timeout_id = g_timeout_add(IMAGE_CACHE_TIMEOUT,
				 reinterpret_cast<GSourceFunc>(timeout_cb),
				 this);
    }
}


void
ImageInfo::size_prepared_cb(GdkPixbufLoader *loader, int width, int height,
			    ImageInfo *info)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (info->manager.is_a_boned(info->url))
    return;	// ܡ󤵤Ƥ롣

  double width_request = info->width;
  double height_request = info->height;

  if (width > width_request || height > height_request)
    {
      double width_scale = width_request / width;
      double height_scale = height_request / height;
      if (width_scale > height_scale)
	{
	  width = static_cast<int>(width * height_scale);
	  height = static_cast<int>(height * height_scale);
	}
      else
	{
	  width = static_cast<int>(width * width_scale);
	  height = static_cast<int>(height * width_scale);
	}
      
      if (width < 0 || height < 0)
	return;	/* ɽ꤬󸽼Ūäꤷoverflowʾ */

      /*
       * ü˽Ĺ䲣Ĺβ̾褦Ȥ̡⤵0ˤʤ
       * GTK+ΤǤк
       */
      if (width == 0)
	width = 1;
      if (height == 0)
	height = 1;

      gdk_pixbuf_loader_set_size(loader, width, height);
    }

  if (gdk_pixbuf_loader_get_animation(loader) != NULL
      || gdk_pixbuf_loader_get_pixbuf(loader) != NULL)
    info->update_pixbuf(loader);
}


void
ImageInfo::area_prepared_cb(GdkPixbufLoader *loader, ImageInfo *info)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  info->update_pixbuf(loader);
}


void
ImageInfo::area_updated_cb(GdkPixbufLoader *loader, int x, int y,
			   int width, int height, ImageInfo *info)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  info->update_pixbuf(loader);
}


void
ImageInfo::closed_cb(GdkPixbufLoader *loader, ImageInfo *info)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  info->update_pixbuf(loader);
}


ImageManager::ImageManager(Controller &controller)
  : CacheManager(controller.get_repository(),
		 "image_cache",	// 켡å
		 "image",	// Ҹ
		 controller.get_prefs().get_capacity_of_image_cache(),
		 controller.get_prefs().get_maximum_number_of_image_cache_files()),
    controller(controller),
    thread_pool(0, 1),
    supported_format_list(gdk_pixbuf_get_formats())
{
  int fd = controller.get_repository().open_file(OCHUSHA_A_BONE_IMAGE_TXT,
						 O_RDONLY);
  if (fd >= 0)
    {
      FILE *file = fdopen(fd, "r");
      if (file != NULL)
	{
	  char line_buf[PATH_MAX];
	  line_buf[PATH_MAX - 1] = '\0';
	  while (!feof(file))
	    {
	      char *tmp_str = fgets(line_buf, PATH_MAX - 1, file);
	      if (tmp_str != NULL)
		{
		  tmp_str = strpbrk(line_buf, "\r\n");
		  if (tmp_str != NULL)
		    *tmp_str = '\0';
		  a_bone_url_table.insert(line_buf, this);
		}
	    }
	  fclose(file);
	}
      else
	close(fd);
    }

  fd = controller.get_repository().open_file(OCHUSHA_A_BONE_SERVER_TXT,
					     O_RDONLY);
  if (fd >= 0)
    {
      FILE *file = fdopen(fd, "r");
      if (file != NULL)
	{
	  char line_buf[PATH_MAX];
	  line_buf[PATH_MAX - 1] = '\0';
	  while (!feof(file))
	    {
	      char *tmp_str = fgets(line_buf, PATH_MAX - 1, file);
	      if (tmp_str != NULL)
		{
		  tmp_str = strpbrk(line_buf, "\r\n");
		  if (tmp_str != NULL)
		    *tmp_str = '\0';
		  a_bone_server_table.insert(line_buf, this);
		}
	    }
	  fclose(file);
	}
      else
	close(fd);
    }

  GSList *list = supported_format_list;
  while (list != NULL)
    {
      GdkPixbufFormat *format = static_cast<GdkPixbufFormat *>(list->data);
      gchar **mime_types = gdk_pixbuf_format_get_mime_types(format);
      gchar **a_mime_type = mime_types;
      while (*a_mime_type != NULL)
	supported_format_mime_types.push_back(strdup(*a_mime_type++));

      g_strfreev(mime_types);
      list = g_slist_next(list);
    }
}


ImageManager::~ImageManager()
{
  thread_pool.cancel_all();
  thread_pool.terminate_workers();

  g_slist_free(supported_format_list);
  vector<char *>::iterator iter = supported_format_mime_types.begin();
  vector<char *>::iterator end_iter = supported_format_mime_types.end();
  while (iter != end_iter)
    free(*iter++);
}


class ImageAgentData : public RefCount
{
private:
  ImageManager &manager;

public:
  ImageAgentData(ImageManager &manager);

  inline ImageManager &get_manager() { return manager; }
};


ImageAgentData::ImageAgentData(ImageManager &manager)
  : manager(manager)
{
}


class ImageBuffer : public Buffer, public has_slots
{
private:
  ImageManager &manager;
  GdkPixbufLoader *loader;
  char *url;
  smart_ptr<Tmpfile> tmpfile;	// ĤNULLʾϥåɤ߹
  smart_ptr<NetworkAgent> agent;

  void prepare_request_cb(NetworkAgent *agent, RefCount *data);
  void status_received_cb(NetworkAgent *agent, RefCount *data);
  void response_header_received_cb(NetworkAgent *agent, RefCount *data);
  void access_finished_cb(NetworkAgent *agent, RefCount *data);
  void access_failed_cb(NetworkAgent *agent, const char *reason,
			RefCount *data);


public:
  ImageBuffer(ImageManager &manager, GdkPixbufLoader *loader);
  ImageBuffer(ImageManager &manager, GdkPixbufLoader *loader,
	      const char *url, Tmpfile *tmpfile, NetworkAgent *agent);
  virtual ~ImageBuffer();
  virtual bool append_data(const char *data, size_t length, int lock_id);
};


ImageBuffer::ImageBuffer(ImageManager &manager, GdkPixbufLoader *loader)
  : Buffer((size_t)0),
    manager(manager),
    loader(loader),
    url(NULL),
    tmpfile(NULL),
    agent(NULL)
{
  assert(loader != NULL);
  g_object_ref(loader);
}


ImageBuffer::ImageBuffer(ImageManager &manager, GdkPixbufLoader *loader,
			 const char *url, Tmpfile *tmpfile, NetworkAgent *agent)
  : Buffer(dup(tmpfile->get_fd()), false),
    manager(manager),
    loader(loader),
    url(strdup(url)),
    tmpfile(tmpfile),
    agent(NULL)
{
  assert(loader != NULL);
  g_object_ref(loader);

  agent->prepare_request.connect(this, &ImageBuffer::prepare_request_cb);
  agent->status_received.connect(this, &ImageBuffer::status_received_cb);
  agent->response_header_received
    .connect(this, &ImageBuffer::response_header_received_cb);
  agent->access_finished.connect(this, &ImageBuffer::access_finished_cb);
  agent->access_failed.connect(this, &ImageBuffer::access_failed_cb);
}


void
ImageBuffer::prepare_request_cb(NetworkAgent *agent, RefCount *data)
{
  // sigslotʥϥɥ
  agent->set_header(HTTPHeader::Accept_Encoding, "deflate, gzip");
  agent->set_header(HTTPHeader::User_Agent, OCHUSHA_USER_AGENT);
  agent->set_non_interactive_mode(true);

  g_return_if_fail(data != NULL);

  ImageAgentData *agent_data = dynamic_cast<ImageAgentData *>(data);
  g_return_if_fail(agent_data != NULL);

  ImageManager &manager = agent_data->get_manager();
  const Preferences &prefs = manager.controller.get_prefs();
  if (prefs.get_enable_proxy_for_images())
    {
      agent->set_proxy_authinfo(NULL, NULL);
      agent->set_proxy_url(prefs.get_proxy_uri_for_images().get_uri());
    }
}


void
ImageBuffer::status_received_cb(NetworkAgent *agent, RefCount *data)
{
  // sigslotʥϥɥ
  if (agent->get_status_code() == 200)
    return;


  GDK_THREADS_ENTER();

  if (loader != NULL)
    {
      gdk_pixbuf_loader_close(loader, NULL);
      g_object_unref(loader);
      loader = NULL;
    }

  tmpfile = NULL;

  GDK_THREADS_LEAVE();
}


void
ImageBuffer::response_header_received_cb(NetworkAgent *agent, RefCount *data)
{
  // sigslotʥϥɥ
  if (agent->get_status_code() != 200)
    return;

  const char *content_type = agent->get_header(HTTPHeader::Content_Type);

  if (!manager.is_content_type_supported(content_type))
    {
      char message[512];
      snprintf(message, 512,
	       _("Unsupported Content-Type, \"%s\", for images: %s\n"),
	       content_type, url != NULL ? url : "NULL?");
      manager.output_log(message);


      GDK_THREADS_ENTER();

      if (loader != NULL)
	{
	  gdk_pixbuf_loader_close(loader, NULL);
	  g_object_unref(loader);
	  loader = NULL;
	}
      tmpfile = NULL;

      GDK_THREADS_LEAVE();
    }
}


void
ImageBuffer::access_finished_cb(NetworkAgent *agent, RefCount *data)
{
  // sigslotʥϥɥ
  GDK_THREADS_ENTER();

  if (loader != NULL)
    {
      if ((gdk_pixbuf_loader_get_animation(loader) != NULL
	   || gdk_pixbuf_loader_get_pixbuf(loader) != NULL)
	  && tmpfile != NULL)
	manager.cache(url, tmpfile);

      gdk_pixbuf_loader_close(loader, NULL);
      g_object_unref(loader);
      loader = NULL;

      tmpfile = NULL;
    }

  GDK_THREADS_LEAVE();
}


void
ImageBuffer::access_failed_cb(NetworkAgent *agent, const char *reason,
			      RefCount *data)
{
  // sigslotʥϥɥ
  GDK_THREADS_ENTER();

  tmpfile = NULL;	// Τ
#if 0
  fprintf(stderr, "ImageBuffer::access_failed_cb(): url=\"%s\" due to %s\n",
	  url != NULL ? url : "<unknown>", reason);
#endif
  if (loader != NULL)
    {
      gdk_pixbuf_loader_close(loader, NULL);
      g_object_unref(loader);
      loader = NULL;
    }

  GDK_THREADS_LEAVE();
}


ImageBuffer::~ImageBuffer()
{
  GDK_THREADS_ENTER();

  // ͥåȥλ⤷ϥåɤ߹߽λƱ
  // ˴Ϥ
#if CHASE_REF_COUNTS
  fprintf(stderr, "ImageBuffer(@%p) destructed\n", this);
#endif
  if (loader != NULL)
    {
      gdk_pixbuf_loader_close(loader, NULL);
#if CHASE_REF_COUNTS
      fprintf(stderr,
	      "ImageBuffer::~ImageBuffer(): loader(@%p)->ref_count=%d\n",
	      loader, ((GObject *)loader)->ref_count);
#endif
      g_object_unref(loader);
      loader = NULL;
    }

  if (url != NULL)
    free(url);

  GDK_THREADS_LEAVE();
}


bool
ImageBuffer::append_data(const char *data, size_t length, int lock_id)
{
  // ͥåȥ򤷤Ƥ륹åɤ⤷ϥե뤫ɤ߹
  // ԤäƤ륹åɤǼ¹Ԥ롣
  gboolean result;

  GDK_THREADS_ENTER();

  if (loader != NULL)
    {
      GError *error = NULL;
      result = gdk_pixbuf_loader_write(loader,
				       reinterpret_cast<const guchar *>(data),
				       length, &error);

      if (result)
	{
	  if (tmpfile != NULL)
	    {
	      GDK_THREADS_LEAVE();

	      return Buffer::append_data(data, length, lock_id);
	    }
	}
      else
	{ // ʾɤǤ̵̤ʤΤǻߤᡣ
	  char message[512];
	  snprintf(message, 512,
		   _("%s: gdk_pixbuf_loader_write(...) failed: %s\n"),
		   url != NULL ? url : "unknown resource",
		   error != NULL ? error->message : "unknown reason");
	  manager.output_log(message);

	  if (error != NULL)
	    {
	      g_error_free(error);
	      error = NULL;
	    }

	  if (!gdk_pixbuf_loader_close(loader, &error))
	    {
	      if (error != NULL)
		{
		  fprintf(stderr, "ImageBuffer::append_data(): gdk_pixbuf_loader_close(loader=%p) failed: %s\n",
			  loader, error->message);
		  g_error_free(error);
		  error = NULL;
		}
	      else
		fprintf(stderr, "ImageBuffer::append_data(): gdk_pixbuf_loader_close(loader=%p) failed: Unknown reason.\n",
			loader);
	    }

#if CHASE_REF_COUNTS
	  fprintf(stderr,
		  "ImageBuffer::append_data(): loader(@%p)->ref_count=%d\n",
		  loader, reinterpret_cast<GObject *>(loader)->ref_count);
#endif
	  g_object_unref(loader);
	  loader = NULL;

	  tmpfile = NULL;	// tmpfileΤ
	  // ʹɤǡɤФäƤ
	  // false֤ȸǤϥͥåȥ¦Out of Memory
	  // ˤƤޤΤɤ߼ΤƤн衣
	}
    }
  else
    result = FALSE;

  GDK_THREADS_LEAVE();


  return result != FALSE;
}


class LoadImageJob : public WorkerJob
{
private:
  smart_ptr<ImageBuffer> buffer;
  int fd;

public:
  LoadImageJob(ImageBuffer *buffer, int fd) : buffer(buffer), fd(fd)
  { assert(fd >= 0); }
  virtual void work(const WorkerThread *employee);
  virtual const char *description() const
  { return "loading image from cache file"; }
};


void
LoadImageJob::work(const WorkerThread *employee)
{
#ifdef HAVE_MMAP
  size_t buf_len = static_cast<size_t>(lseek(fd, 0, SEEK_END));
  char *buf = static_cast<char *>(mmap(NULL, buf_len, PROT_READ,
				       MAP_PRIVATE | MAP_NOCORE | MAP_NOSYNC,
				       fd, 0));
  if (buf != MAP_FAILED)
    {
      buffer->append_data(buf, buf_len, 0);
      munmap(buf, buf_len);
      close(fd);
      fd = -1;
      return;
    }

  // ԤǽλǤɤ褦ʵ⤹뤬ġġ
  buf = static_cast<char *>(malloc(LOAD_IMAGE_CHUNKSIZE));
#else
  char *buf = static_cast<char *>(malloc(LOAD_IMAGE_CHUNKSIZE));
#endif

  g_return_if_fail(buffer != NULL);	// ̵
  while (employee->get_state() != WorkerThread::DIE_NOW)
    {
      ssize_t result = read(fd, buf, LOAD_IMAGE_CHUNKSIZE);
      if (result <= 0 || !buffer->append_data(buf, result, 0))
	break;	// λǽʥ顼
    }
  free(buf);
  close(fd);
  fd = -1;
}


GtkWidget *
ImageManager::get_widget(const char *url, int width, int height)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  g_return_val_if_fail(url != NULL, NULL);
  if (*url == '\0')
    return NULL;

  if (is_a_boned(url))
    return NULL;

#ifndef DISABLE_SCHEME_REWRITER
  SchemeManager &sm = controller.get_scheme_manager();
  Cell *result = sm.get_core().evalf0(sm.rewrite_image_link.get_cell(),
				      "%s", url);
  string url_buf;
  if (result != NULL && result->is_string())
    {
      url_buf.append(result->get_string(), result->get_strlen());

      if (strcmp(url_buf.c_str(), url) != 0)
	{
	  string message_buf;
	  message_buf.append("rewrite-image-link: ").append(url).append(" => ")
	    .append(url_buf.c_str()).append("\n");
	  sm.console.output(message_buf.c_str());
	  url = url_buf.c_str();
	}
    }
#endif

  URI uri(url);
  if (!uri.is_valid()
      || uri.get_scheme() == NULL || strcmp(uri.get_scheme(), "http") != 0
      || uri.get_server() == NULL)
    return NULL;

  ImageInfo *image_info = lookup_image_info(url, width, height);
  if (image_info != NULL)
    {
      GtkWidget *image;
      if (image_info->get_animation() != NULL)
	image = gtk_image_new_from_animation(image_info->get_animation());
      else if (image_info->get_pixbuf() != NULL)
	image = gtk_image_new_from_pixbuf(image_info->get_pixbuf());
      else if (image_info->get_loader() != NULL)
	image = gtk_image_new_from_stock(OCHUSHA_STOCK_MISSING_IMAGE,
					 GTK_ICON_SIZE_DIALOG);
      else
	image = gtk_image_new_from_stock(GTK_STOCK_MISSING_IMAGE,
					 GTK_ICON_SIZE_DIALOG);
      image_info->add_widget(image);
      touch(url);
      return image;
    }

  GSList *info_list = static_cast<GSList *>(image_info_table.lookup(url));
  if (info_list != NULL)
    {
      // 㤤ɤ
      image_info = static_cast<ImageInfo *>(info_list->data);
      if(image_info->get_loader() == NULL
	 && image_info->get_animation() == NULL
	 && image_info->get_pixbuf() == NULL)
	{
	  // 󥯤Ǥ뤫ɽǤǤʤȤȽƤURL
	  GtkWidget *image = gtk_image_new_from_stock(GTK_STOCK_MISSING_IMAGE,
						      GTK_ICON_SIZE_DIALOG);
	  image_info = new ImageInfo(*this, url, NULL, image, width, height);
	  assert(info_list != NULL);	// GDK_THREADS_ENTER()ǼƤ
	  info_list = g_slist_append(info_list, image_info);
	  return image;
	}
    }

  int fd = open_cache(url, O_RDONLY);

  if (fd < 0 && controller.get_config().get_offline())
    return gtk_image_new_from_stock(GTK_STOCK_MISSING_IMAGE,
				    GTK_ICON_SIZE_DIALOG);

  GtkWidget *image = gtk_image_new_from_stock(OCHUSHA_STOCK_MISSING_IMAGE,
					      GTK_ICON_SIZE_DIALOG);
  GdkPixbufLoader *loader = gdk_pixbuf_loader_new();
  image_info = new ImageInfo(*this, url, loader, image, width, height);
  if (info_list != NULL)
    info_list = g_slist_append(info_list, image_info);
  else
    image_info_table.insert(url, g_slist_append(NULL, image_info));

  if (fd >= 0)
    {
      // å夬Ĥä
      touch(url);	// öƤ


      GDK_THREADS_LEAVE();

      smart_ptr<ImageBuffer> buffer = new ImageBuffer(*this, loader);
      thread_pool.commit(new LoadImageJob(buffer, fd));
    }
  else
    {
      smart_ptr<Tmpfile> tmpfile
	= controller.get_repository().create_tmpfile(O_WRONLY|O_CREAT|O_TRUNC);
      NetworkAgent *agent = controller.get_broker().employ_agent();


      GDK_THREADS_LEAVE();

      smart_ptr<ImageBuffer> buffer
	= new ImageBuffer(*this, loader, url, tmpfile, agent);
      agent->get(url, buffer, new ImageAgentData(*this));
    }

  // åɤΥ塼󥰶礤ͥåȥоݤξ
  // ˤäƤϡif֥åȴImageBufferdestruct롣

  // GDK_THREADS_ENTER()ǻȤäƤåϡrecursive lockˤƤ
  // 뤫ºݤˤϰImageBufferdestructorGDK_THREADS_ENTER()
  // ΤǡImageBufferdestructǤGDK_THREADS_ENTER()
  // destruct褦ˤƤrecursiveGDK_THREADS_ENTER()
  // ƤФΤϥХäƤ뤳Ȥ¿ΤǡѤθåǤ
  // recursiveGDK_THREADS_ENTER()ϵƤʤ
  GDK_THREADS_ENTER();


#if CHASE_REF_COUNTS
  fprintf(stderr, "ImageManager::get_widget(): loader(@%p)->ref_count=%d\n",
	  loader, ((GObject *)loader)->ref_count);
#endif
  g_object_unref(loader);

  return image;
}


ImageInfo *
ImageManager::lookup_image_info(const char *url, int width, int height)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());

  GSList *info_list = static_cast<GSList *>(image_info_table.lookup(url));
  while (info_list != NULL)
    {
      ImageInfo *info = static_cast<ImageInfo *>(info_list->data);
      if (info->get_width() == width && info->get_height() == height)
	return info;
      info_list = info_list->next;
    }
  return NULL;
}


static void
a_bone_image_in_table(ImageInfo *info, gpointer unused)
{
  info->a_bone();
}


static void
output_a_bone_key(void *unused, FILE *file, const char *server)
{
  fprintf(file, "%s\n", server);
}


void
ImageManager::a_bone(const char *url)
{
  assert(Lock::get_slave_lock()->is_locked_by_self());
  if (url == NULL || *url == '\0')
    return;

  if (a_bone_url_table.lookup(url) != NULL)
    return;

  URI uri(url);
  if (!uri.is_valid())
    return;

  a_bone_url_table.insert(url, this);

  GSList *info_list = static_cast<GSList *>(image_info_table.lookup(url));
  g_slist_foreach(info_list,
		  reinterpret_cast<GFunc>(a_bone_image_in_table), NULL);

  invalidate(url);

  int fd = controller.get_repository().open_file(OCHUSHA_A_BONE_IMAGE_TXT,
						 O_WRONLY | O_TRUNC | O_CREAT);
  FILE *file;
  if (fd >= 0)
    {
      file = fdopen(fd, "w");
      if (file == NULL)
	{
	  close(fd);	// 롣
	  return;
	}
      a_bone_url_table.for_each(reinterpret_cast<CStringToPointerHashtable::Scanner>(output_a_bone_key), file);
      fclose(file);
    }
}


struct ABoneServerArgs
{
  ImageManager &manager;
  const char *server;
};


static void
a_bone_server_in_table(GSList *info_list,
		       ABoneServerArgs *args, const char *url)
{
  URI uri(url);
  assert(uri.is_valid());
  assert(uri.get_server() != NULL);

  if (strcmp(uri.get_server(), args->server) != 0)
    return;

  g_slist_foreach(info_list,
		  reinterpret_cast<GFunc>(a_bone_image_in_table), NULL);

  args->manager.invalidate(url);
}


void
ImageManager::a_bone_server(const char *url)
{
  if (url == NULL || *url == '\0')
    return;

  URI uri(url);
  if (!uri.is_valid() || uri.get_server() == NULL
      || a_bone_server_table.lookup(uri.get_server()) != NULL)
    return;

  a_bone_server_table.insert(uri.get_server(), this);

  ABoneServerArgs args = { *this, uri.get_server() };
  image_info_table.for_each(reinterpret_cast<CStringToPointerHashtable::Scanner>(a_bone_server_in_table), &args);

  int fd = controller.get_repository().open_file(OCHUSHA_A_BONE_SERVER_TXT,
						 O_WRONLY | O_TRUNC | O_CREAT);
  FILE *file;
  if (fd >= 0)
    {
      file = fdopen(fd, "w");
      if (file == NULL)
	{
	  close(fd);	// 롣
	  return;
	}
      a_bone_server_table.for_each(reinterpret_cast<CStringToPointerHashtable::Scanner>(output_a_bone_key), file);
      fclose(file);
    }
}


bool
ImageManager::is_a_boned(const char *url)
{
  assert(url != NULL);
  if (a_bone_url_table.lookup(url) != NULL)
    return true;

  URI uri(url);
  if (!uri.is_valid() || uri.get_server() == NULL)
    return false;

  return a_bone_server_table.lookup(uri.get_server()) != NULL;
}


bool
ImageManager::is_image_link(const char *url)
{
  URI uri(url);
  if (!uri.is_valid())
    return false;

  const char *scheme = uri.get_scheme();
  if (scheme == NULL
      || (strcmp(scheme, "http") != 0 && strcmp(scheme, "https") != 0))
    return false;

  const char *path = uri.get_path();
  if (path != NULL
      && (g_str_has_suffix(path, ".jpg") || g_str_has_suffix(path, ".png")
	  || g_str_has_suffix(path, ".gif") || g_str_has_suffix(path, ".jpeg")
	  || g_str_has_suffix(path, ".JPG") || g_str_has_suffix(path, ".PNG")
	  || g_str_has_suffix(path, ".GIF") || g_str_has_suffix(path, ".JPEG")))
    return !is_a_boned(url);

  const char *query = uri.get_query();
  if (query != NULL
      && (strcasestr(query, ".jpg") != NULL
	  || strcasestr(query, ".png") != NULL
	  || strcasestr(query, ".gif") != NULL
	  || strcasestr(query, ".jpeg") != NULL))
    return !is_a_boned(url);

#ifndef DISABLE_SCHEME_REWRITER
  SchemeManager &sm = controller.get_scheme_manager();
  Cell *result = sm.get_core().evalf0(sm.is_image_link.get_cell(), "%s", url);

  return result != NULL && result != Cell::f();
#else
  return false;
#endif
}


bool
ImageManager::is_content_type_supported(const char *content_type)
{
  if (content_type == NULL || *content_type == '\0')
    return false;

  vector<char *>::iterator iter = supported_format_mime_types.begin();
  vector<char *>::iterator end_iter = supported_format_mime_types.end();
  while (iter != end_iter)
    {
      if (strcmp(*iter, content_type) == 0)
	return true;

      ++iter;
    }

  return false;
}

}
