/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: scheme_manager.cc,v 1.17 2008/10/25 08:49:50 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "scheme_manager.h"

#include "libochusha/ochusha_procedure.h"
#include "libochusha/misc_utils.h"

#include "controller.h"
#include "gdk_procedure.h"
#include "image_manager.h"
#include "selection_utils.h"
#include "ui_constants.h"

#include "libochushagtk/menu_action.h"
#include "libochushagtk/misc_utils.h"

#include <gtk/gtk.h>


#include <cassert>
#include <cstdio>

#include <string>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


using namespace LibTSCore;
using namespace Ochusha;
using namespace std;
namespace Ochusha
{

#define SHOW_SCHEME_OUTPUT_IN_CONSOLE	1

static const char *default_init_scm =
  "(define (decorate-list-entry-default thread)"
  "  (let ((new-responses (thread-get-number-of-new-responses thread))"
  "        (n-got (thread-get-number-of-responses-read thread)))"
  "    (cond ((thread-hidden? thread)"
  "           (cons list-entry-fg-hidden list-entry-bg-hidden))"
  "          ((= (thread-get-previous-rank thread) 0)"
  "           (cons list-entry-fg-normal list-entry-bg-strong))"
  "          ((> n-got 0)"
  "           (if (> (thread-get-number-of-responses-on-server thread) n-got)"
  "               (cons list-entry-fg-emph list-entry-bg-emph) '()))"
  "          (else '()))))"
  "(define decorate-list-entry decorate-list-entry-default)"
  "(register-threadlist-rule"
  "  \"ochusha-default-plain-all\""
  "  \"all\""
  "  \"live threads -> dead threads\""
  "  (lambda (thread)"
  "    (if (thread-alive? thread) 0 10000)))";


enum {
  EXTERNAL_TOOL_NAME = 0,
  EXTERNAL_TOOL_TYPE,
  EXTERNAL_TOOL_LABEL,
  EXTERNAL_TOOL_DESCRIPTION,
  EXTERNAL_TOOL_NICK,
  EXTERNAL_TOOL_SENSITIVE,
  EXTERNAL_TOOLS_MODEL_N_COLUMNS
};


SchemeManager::SchemeManager(Controller &controller)
  : controller(controller),
    console(_("Scheme Console"), get_output_fd(),
	    controller.get_prefs().scheme_console),
    ochusha_procedure(new OchushaProcedure(core)),
    decorate_list_entry(core, core.mk_symbol("decorate-list-entry")),
    rewrite_link(core, core.mk_symbol("rewrite-link")),
    rewrite_message(core, core.mk_symbol("rewrite-message")),
    is_image_link(core, core.mk_symbol("image-link?")),
    rewrite_image_link(core, core.mk_symbol("rewrite-image-link")),
    prepare_image_request(core, core.mk_symbol("prepare-image-request")),
    action_group(create_action_group())
{
#ifdef SHOW_SCHEME_OUTPUT_IN_CONSOLE	// ư褦ˤʤä
  int fd = get_input_fd();
  if (fd >= 0)
    {
      FILE *file = fdopen(fd, "w");
      if (file != NULL)
	core.set_interaction_output_port_file(file);
      else
	close(fd);
    }
#else
  int fd;
  core.set_interaction_output_port_file(stderr);
#endif


  core.define_procedure(const_cast<OchushaProcedure *>(ochusha_procedure));
  core.define_procedure(new GdkProcedure(core));


  Preferences &prefs = controller.get_prefs();
  define_constants(prefs);


  FILE *scm_file = fopen(PKGDATADIR "/" OCHUSHA_INIT_SCM, "r");
  if (scm_file != NULL)
    {
      core.load_file(scm_file);
      fclose(scm_file);
    }
  else
    core.load_string(default_init_scm);

  fd = controller.get_repository().open_file(OCHUSHA_INIT_SCM, O_RDONLY);
  if (fd >= 0)
    {
      scm_file = fdopen(fd, "r");
      if (scm_file != NULL)
	{
	  core.load_file(scm_file);
	  fclose(scm_file);
	}
      else
	close(fd);
    }

  update_tools_models();


  GDK_THREADS_LEAVE();

  prefs.prefs_changed.connect(this, &SchemeManager::prefs_changed_cb);

  NetworkAgentBroker &broker = controller.get_broker();
  broker.pre_prepare_request.connect(this,
				     &SchemeManager::pre_prepare_request_cb);
  broker.post_prepare_request.connect(this,
				      &SchemeManager::post_prepare_request_cb);

  GDK_THREADS_ENTER();
}


static void
set_menu_data(GtkMenuItem *menu_item, GtkLabel *menu_label,
	      GtkTreeModel *model, GtkTreeIter *iter, gpointer unused)
{
  const char *label = NULL;
  gtk_tree_model_get(model, iter, EXTERNAL_TOOL_LABEL, &label, -1);
  g_return_if_fail(label != NULL);

  gtk_label_set_text(menu_label, label);
}


static gboolean
tool_is_sensitive(GtkTreeModel *model, GtkTreeIter *iter, gpointer unused)
{
  gboolean sensitive = TRUE;
  gtk_tree_model_get(model, iter, EXTERNAL_TOOL_SENSITIVE, &sensitive, -1);

  return sensitive;
}


static gboolean
is_link_tool_iter(GtkTreeModel *model, GtkTreeIter *iter, gpointer unused)
{
  gint tool_type = 0;
  gtk_tree_model_get(model, iter, EXTERNAL_TOOL_TYPE, &tool_type, -1);

  return tool_type == ExternalToolInfo::LINK_TOOL;
}


static gboolean
is_image_tool_iter(GtkTreeModel *model, GtkTreeIter *iter, gpointer unused)
{
  gint tool_type = 0;
  gtk_tree_model_get(model, iter, EXTERNAL_TOOL_TYPE, &tool_type, -1);

  return tool_type == ExternalToolInfo::IMAGE_TOOL;
}


static gboolean
is_text_tool_iter(GtkTreeModel *model, GtkTreeIter *iter, gpointer unused)
{
  gint tool_type = 0;
  gtk_tree_model_get(model, iter, EXTERNAL_TOOL_TYPE, &tool_type, -1);

  return tool_type == ExternalToolInfo::TEXT_TOOL;
}


void
SchemeManager::menu_activate_cb(OchushaMenuAction *action,
				GtkTreeModel *model,
				GtkTreeIter *iter, GtkTreePath *path,
				SchemeManager *manager)
{
  const char *name = NULL;
  gint tool_type = 0;

  gtk_tree_model_get(model, iter,
		     EXTERNAL_TOOL_NAME, &name,
		     EXTERNAL_TOOL_TYPE, &tool_type,
		     -1);

  g_return_if_fail(tool_type >= ExternalToolInfo::LINK_TOOL
		   && tool_type <= ExternalToolInfo::TEXT_TOOL);

  manager->invoke_external_tool_helper(name, tool_type);
}


GtkActionGroup *
SchemeManager::create_action_group()
{
  GtkActionGroup *group = gtk_action_group_new("SchemeManagerActions");
  gtk_action_group_set_translation_domain(group, PACKAGE_NAME);

  external_tools_store = gtk_list_store_new(EXTERNAL_TOOLS_MODEL_N_COLUMNS,
					    G_TYPE_POINTER,
					    G_TYPE_INT,
					    G_TYPE_POINTER,
					    G_TYPE_POINTER,
					    G_TYPE_POINTER,
					    G_TYPE_BOOLEAN);

  GtkTreeModel *model = GTK_TREE_MODEL(external_tools_store);
  link_tools_model = gtk_tree_model_filter_new(model, NULL);
  gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(link_tools_model),
					 is_link_tool_iter, this, NULL);

  OchushaMenuAction *menu_action
    = ochusha_menu_action_new_with_model("ExternalLinkTools",
					 _("External Tools for Links"),
					 NULL,
					 NULL,
					 link_tools_model);
  ochusha_menu_action_set_menu_data_func(menu_action, set_menu_data,
					 this, NULL);
  ochusha_menu_action_set_row_sensitive_func(menu_action, tool_is_sensitive,
					     this, NULL);
  g_signal_connect(menu_action, "menu_activate",
		   reinterpret_cast<GCallback>(menu_activate_cb), this);

  link_tools_action = GTK_ACTION(menu_action);
  gtk_action_group_add_action(group, link_tools_action);



  image_tools_model = gtk_tree_model_filter_new(model, NULL);
  gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(image_tools_model),
					 is_image_tool_iter, this, NULL);

  menu_action
    = ochusha_menu_action_new_with_model("ExternalImageTools",
					 _("External Tools for Images"),
					 NULL,
					 NULL,
					 image_tools_model);
  ochusha_menu_action_set_menu_data_func(menu_action, set_menu_data,
					 this, NULL);
  ochusha_menu_action_set_row_sensitive_func(menu_action, tool_is_sensitive,
					     this, NULL);
  g_signal_connect(menu_action, "menu_activate",
		   reinterpret_cast<GCallback>(menu_activate_cb), this);

  image_tools_action = GTK_ACTION(menu_action);
  gtk_action_group_add_action(group, image_tools_action);



  text_tools_model = gtk_tree_model_filter_new(model, NULL);
  gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(text_tools_model),
					 is_text_tool_iter, this, NULL);

  menu_action
    = ochusha_menu_action_new_with_model("ExternalTextTools",
					 _("External Tools for Texts"),
					 NULL,
					 NULL,
					 text_tools_model);
  ochusha_menu_action_set_menu_data_func(menu_action, set_menu_data,
					 this, NULL);
  ochusha_menu_action_set_row_sensitive_func(menu_action, tool_is_sensitive,
					     this, NULL);
  g_signal_connect(menu_action, "menu_activate",
		   reinterpret_cast<GCallback>(menu_activate_cb), this);

  text_tools_action = GTK_ACTION(menu_action);
  gtk_action_group_add_action(group, text_tools_action);

  gtk_ui_manager_insert_action_group(controller.get_ui_manager(), group, 0);

  g_signal_connect(controller.get_internal_clipboard(), "owner-change",
		   reinterpret_cast<GCallback>(owner_change_cb), this);

  return group;
}


void
SchemeManager::update_tools_models()
{
  gtk_list_store_clear(external_tools_store);

  vector<ExternalToolInfo *>::const_iterator iter
    = ochusha_procedure->external_tool_list_begin();
  vector<ExternalToolInfo *>::const_iterator end_iter
    = ochusha_procedure->external_tool_list_end();

  while (iter != end_iter)
    {
      ExternalToolInfo *info = *iter++;
      gtk_list_store_insert_with_values(external_tools_store, NULL, G_MAXINT32,
			EXTERNAL_TOOL_NAME, info->get_name(),
			EXTERNAL_TOOL_TYPE, info->get_tool_type(),
			EXTERNAL_TOOL_LABEL, info->get_label(),
			EXTERNAL_TOOL_DESCRIPTION, info->get_description(),
			EXTERNAL_TOOL_NICK, "Reserved",
			EXTERNAL_TOOL_SENSITIVE, TRUE,
			-1);
    }
}


void
SchemeManager::owner_change_cb(GtkClipboard *clipboard, GdkEvent *event,
			       SchemeManager *manager)
{
  gboolean text_available = gtk_clipboard_wait_is_text_available(clipboard);
  manager->tools_set_sensitive(ExternalToolInfo::TEXT_TOOL, text_available);

  gboolean uri_available = ochusha_clipboard_wait_is_uri_available(clipboard);

  manager->tools_set_sensitive(ExternalToolInfo::LINK_TOOL,
			       uri_available != FALSE);
  if (uri_available)
    {
      gchar **uri_list = ochusha_clipboard_wait_for_uri(clipboard);
      g_return_if_fail(uri_list != NULL);

      gchar **uri_pos = uri_list;
      ImageManager &image_manager = manager->controller.get_image_manager();
      while (*uri_pos != NULL)
	{
	  if (image_manager.cache_exist(*uri_pos))
	    break;
	  ++uri_pos;
	}

      manager->tools_set_sensitive(ExternalToolInfo::IMAGE_TOOL,
				   *uri_pos != NULL);

      g_free(uri_list);
    }
}


void
SchemeManager::tools_set_sensitive(int tool_type, bool value)
{
  GtkListStore *store = external_tools_store;
  GtkTreeModel *model = GTK_TREE_MODEL(store);
  GtkTreeIter iter;
  if (gtk_tree_model_get_iter_first(model, &iter))
    {
      do
	{
	  int tt = ExternalToolInfo::UNKNOWN_TOOL;
	  gtk_tree_model_get(model, &iter, EXTERNAL_TOOL_TYPE, &tt, -1);
	  if (tt == tool_type)
	    gtk_list_store_set(store, &iter,
			       EXTERNAL_TOOL_SENSITIVE, value ? TRUE : FALSE,
			       -1);
	}
      while (gtk_tree_model_iter_next(model, &iter));
    }
}


void
SchemeManager::link_tools_set_sensitive(bool value)
{
  tools_set_sensitive(ExternalToolInfo::LINK_TOOL, value);
}


void
SchemeManager::image_tools_set_sensitive(bool value)
{
  tools_set_sensitive(ExternalToolInfo::IMAGE_TOOL, value);
}


void
SchemeManager::text_tools_set_sensitive(bool value)
{
  tools_set_sensitive(ExternalToolInfo::TEXT_TOOL, value);
}


void
SchemeManager::define_constants(Preferences &prefs)
{
  core.define_global_foreign_object("list-entry-fg-normal",
				    prefs.get_fg_normal(), NULL);
  core.define_global_foreign_object("list-entry-bg-normal",
				    prefs.get_bg_normal(), NULL);
  core.define_global_foreign_object("list-entry-fg-emph",
				    prefs.get_fg_emph(), NULL);
  core.define_global_foreign_object("list-entry-bg-emph",
				    prefs.get_bg_emph(), NULL);
  core.define_global_foreign_object("list-entry-fg-strong",
				    prefs.get_fg_strong(), NULL);
  core.define_global_foreign_object("list-entry-bg-strong",
				    prefs.get_bg_strong(), NULL);
  core.define_global_foreign_object("list-entry-fg-hidden",
				    prefs.get_fg_hidden(), NULL);
  core.define_global_foreign_object("list-entry-bg-hidden",
				    prefs.get_bg_hidden(), NULL);
}


void
SchemeManager::prefs_changed_cb(Preferences &prefs)
{
  GDK_THREADS_ENTER();

  define_constants(prefs);

  GDK_THREADS_LEAVE();
}


void
SchemeManager::pre_prepare_request_cb(NetworkAgent *agent)
{
}


void
SchemeManager::post_prepare_request_cb(NetworkAgent *agent)
{
}


int
SchemeManager::get_output_fd()
{
  comm[0] = -1;
  comm[1] = -1;
  if (pipe(comm) == 0)
    return comm[0];
  return -1;
}


int
SchemeManager::get_input_fd()
{
  // get_output_fd()ƤФƤȲꤷƤ롣
  g_return_val_if_fail(comm[1] != -1, -1);

  fcntl(comm[1], F_SETFL,
	fcntl(comm[1], F_GETFL) | O_NONBLOCK);
  return comm[1];
}


SchemeManager::~SchemeManager()
{
  g_object_unref(text_tools_action);
  g_object_unref(text_tools_model);
  g_object_unref(image_tools_action);
  g_object_unref(image_tools_model);
  g_object_unref(link_tools_action);
  g_object_unref(link_tools_model);
  g_object_unref(external_tools_store);
  g_object_unref(action_group);
}


void
SchemeManager::prepare_shutdown()
{
  console.prepare_shutdown();
}


void
SchemeManager::invoke_external_tool(const char *name, const char *argument)
{
  if (name == NULL || argument == NULL)
    return;

  const ExternalToolInfo *info
    = ochusha_procedure->lookup_external_tool_info(name);

  if (info == NULL)
    return;

  Cell *templ = info->get_template();
  if (templ->is_procedure())
    {
      templ = core.applyf(templ, "%s", argument);
      if (templ == NULL || !templ->is_string())
	return;
    }

  assert(templ->is_string());
  const char *templ_start = templ->get_string();
  const char *templ_tail = templ_start + templ->get_strlen();

  string command_line_buf;
  const char *cur_pos = templ_start;
  while (cur_pos < templ_tail)
    {
      const char *key = strnstr(cur_pos, "%ARG%", templ_tail - cur_pos);
      if (key != NULL)
	{
	  if (cur_pos != key)
	    command_line_buf.append(cur_pos, key - cur_pos);
	  command_line_buf.append(argument);
	  cur_pos = key + 5;
	}
      else
	{
	  command_line_buf.append(cur_pos, templ_tail - cur_pos);
	  break;
	}
    }

  if (command_line_buf.c_str()[0] != '\0')
    {
      if (strncmp(command_line_buf.c_str(), "%WEBBROWSER%", 12) == 0)
	{
	  const char *link = command_line_buf.c_str() + 12;
	  while (*link != '\0' && *link == ' ')
	    ++link;
	  if (*link != '\0')
	    controller.open(link, true, true);
	}
      else
	controller.spawn_command(command_line_buf.c_str());
    }
}


void
SchemeManager::invoke_external_tool_helper(const char *name, int tool_type)
{
  GtkClipboard *clipboard = controller.get_internal_clipboard();

  switch (tool_type)
    {
    case ExternalToolInfo::LINK_TOOL:
      {
	if (!ochusha_clipboard_wait_is_uri_available(clipboard))
	  return;

	gchar **uri_list = ochusha_clipboard_wait_for_uri(clipboard);
	g_return_if_fail(uri_list != NULL);

	gchar **uri_pos = uri_list;
	while (*uri_pos != NULL)
	  invoke_external_tool(name, *uri_pos++);

	g_strfreev(uri_list);
	return;
      }

    case ExternalToolInfo::IMAGE_TOOL:
      {
	if (!ochusha_clipboard_wait_is_uri_available(clipboard))
	  return;

	gchar **uri_list = ochusha_clipboard_wait_for_uri(clipboard);
	g_return_if_fail(uri_list != NULL);

	gchar **uri_pos = uri_list;
	while (*uri_pos != NULL)
	  {
	    const char *uri = *uri_pos++;
	    char *filename = controller.get_image_manager().find_cache(uri);
	    if (filename != NULL)
	      {
		invoke_external_tool(name, filename);
		free(filename);
	      }
	  }

	g_strfreev(uri_list);
	return;
      }

    case ExternalToolInfo::TEXT_TOOL:
      {
	if (!gtk_clipboard_wait_is_text_available(clipboard))
	  return;

	gchar *text = gtk_clipboard_wait_for_text(clipboard);
	g_return_if_fail(text != NULL);
	invoke_external_tool(name, text);
	g_free(text);
      }

    default:
      break;
    }
}


void
SchemeManager::attach_external_text_tools_menu(GtkTextView *view,
					       GtkMenu *popup_menu)
{
  GtkWidget *menu = gtk_action_create_menu_item(text_tools_action);
  g_return_if_fail(menu != NULL);
  g_return_if_fail(GTK_IS_MENU_ITEM(menu));

  GtkWidget *separator = gtk_separator_menu_item_new();
  gtk_widget_show(separator);
  gtk_menu_shell_prepend(GTK_MENU_SHELL(popup_menu), separator);
  gtk_widget_show(menu);
  gtk_menu_shell_prepend(GTK_MENU_SHELL(popup_menu), menu);

  GtkTextBuffer *buffer = gtk_text_view_get_buffer(view);
  controller.update_internal_clipboard(new TextBufferSelectionData(buffer));
}

}
