/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: selection_utils.cc,v 1.3 2008/09/24 11:56:59 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "selection_utils.h"

#include <cstdlib>
#include <cstring>


namespace Ochusha
{


static GQuark data_id
  = g_quark_from_static_string("SelectionUtils::SelectionData");


SelectionData::~SelectionData()
{
}


void
SelectionData::get_data(GtkClipboard *clipboard,
			GtkSelectionData *data,
			guint info, gpointer user_data)
{
  GObject *owner = G_OBJECT(user_data);
  g_return_if_fail(owner != NULL);

  SelectionData *selection_data
    = static_cast<SelectionData *>(g_object_get_qdata(owner, data_id));
  g_return_if_fail(selection_data != NULL);

  selection_data->get_data(clipboard, data, info);
}


void
SelectionData::clear_data(GtkClipboard *clipboard, gpointer user_data)
{
  GObject *owner = G_OBJECT(user_data);
  g_return_if_fail(owner != NULL);

  SelectionData *selection_data
    = static_cast<SelectionData *>(g_object_get_qdata(owner, data_id));
  g_return_if_fail(selection_data != NULL);

  if (selection_data->clear_data(clipboard))
    delete selection_data;

  g_object_unref(owner);
}


gboolean
SelectionData::set_to(GtkClipboard *clipboard)
{
  g_return_val_if_fail(GTK_IS_CLIPBOARD(clipboard), FALSE);

  GObject *owner = G_OBJECT(g_object_new(G_TYPE_OBJECT, NULL));
  g_object_set_qdata(owner, data_id, this);

  return gtk_clipboard_set_with_owner(clipboard, get_targets(), get_n_targets(),
				      get_data, clear_data, owner);
}


gboolean
SelectionData::set_uris(GtkSelectionData *data,
			const vector<const char *> &uris)
{
  if (uris.size() == 0)
    return FALSE;

  gchar **uri_list
    = static_cast<gchar **>(malloc(sizeof(gchar *) * (uris.size() + 1)));

  int i = 0;
  vector<const char *>::const_iterator iter = uris.begin();
  vector<const char *>::const_iterator end_iter = uris.end();
  while (iter != end_iter)
    uri_list[i++] = const_cast<char *>(*iter++);
  uri_list[i] = NULL;

  gboolean result = gtk_selection_data_set_uris(data, uri_list);
  free(uri_list);

  return result;
}


gboolean
SelectionData::set_uris(GtkSelectionData *data,
			const vector<char *> &uris)
{
  if (uris.size() == 0)
    return FALSE;

  gchar **uri_list
    = static_cast<gchar **>(malloc(sizeof(gchar *) * (uris.size() + 1)));

  int i = 0;
  vector<char *>::const_iterator iter = uris.begin();
  vector<char *>::const_iterator end_iter = uris.end();
  while (iter != end_iter)
    uri_list[i++] = *iter++;
  uri_list[i] = NULL;

  gboolean result = gtk_selection_data_set_uris(data, uri_list);
  free(uri_list);

  return result;
}


TextBufferSelectionData::TextBufferSelectionData(GtkTextBuffer *buffer)
  : buffer(buffer)
{
  g_object_ref(buffer);
}


TextBufferSelectionData::~TextBufferSelectionData()
{
  g_object_unref(buffer);
}


static const GtkTargetEntry text_targets[] = {
  { const_cast<char *>("UTF8_STRING"), 0, 0 },
  { const_cast<char *>("STRING"), 0, 0 },
  { const_cast<char *>("TEXT"), 0, 0 },
  { const_cast<char *>("COMPOUND_TEXT"), 0, 0 }
};


static const GtkTargetEntry empty_targets[] = {
};


const GtkTargetEntry *
TextBufferSelectionData::get_targets()
{
#if GTK_CHECK_VERSION(2,10,0)
  return gtk_text_buffer_get_has_selection(buffer)
    ? text_targets : empty_targets;
#else
  GtkTextIter start_iter;
  GtkTextIter end_iter;
  return gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter)
    ? text_targets : empty_targets;
#endif
}


guint
TextBufferSelectionData::get_n_targets()
{
#if GTK_CHECK_VERSION(2,10,0)
  return gtk_text_buffer_get_has_selection(buffer)
    ? G_N_ELEMENTS(text_targets) : 0;
#else
  GtkTextIter start_iter;
  GtkTextIter end_iter;
  return gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter)
    ? G_N_ELEMENTS(text_targets) : 0;
#endif
}


void
TextBufferSelectionData::get_data(GtkClipboard *clipboard,
				  GtkSelectionData *data, guint info)
{
  g_return_if_fail(info == 0);

  GtkTextIter start_iter;
  GtkTextIter end_iter;
  if (gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter))
    {
      gchar *text
	= gtk_text_buffer_get_text(buffer, &start_iter, &end_iter, FALSE);
      gtk_selection_data_set_text(data, text, -1);
      g_free(text);
    }
}


bool
TextBufferSelectionData::clear_data(GtkClipboard *clipboard)
{
  return true;
}


SelectionDataProxy::SelectionDataProxy(SelectionData *data, GtkWidget *widget)
  : data(data),
    widget(widget)
{
  destroy_id = g_signal_connect(widget, "destroy",
				reinterpret_cast<GCallback>(destroy_cb), this);
}


SelectionDataProxy::~SelectionDataProxy()
{
  if (widget != NULL)
    g_signal_handler_disconnect(widget, destroy_id);
}


void
SelectionDataProxy::destroy_cb(GtkWidget *widget, SelectionDataProxy *proxy)
{
  proxy->data = NULL;
  proxy->widget = NULL;
}


const GtkTargetEntry *
SelectionDataProxy::get_targets()
{
  return data != NULL ? data->get_targets() : empty_targets;
}


guint
SelectionDataProxy::get_n_targets()
{
  return data != NULL ? data->get_n_targets() : 0;
}


void
SelectionDataProxy::get_data(GtkClipboard *clipboard,
			     GtkSelectionData *selection_data, guint info)
{
  if (data != NULL)
    data->get_data(clipboard, selection_data, info);
}


bool
SelectionDataProxy::clear_data(GtkClipboard *clipboard)
{
  if (widget != NULL)
    {
      g_signal_handler_disconnect(widget, destroy_id);
      widget = NULL;
      destroy_id = 0;
    }
  return true;
}


}
