/* -*- c++ -*-
 * Copyright (c) 2003-2008 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: view_manager.cc,v 1.38 2008/10/25 08:45:56 fuyu Exp $
 */

#include "config.h"

#include "libochusha/ochusha_intl.h"

#if defined(__GNUG__) && !defined(__ICC)
#pragma implementation
#endif
#include "view_manager.h"

#include "libochusha/bbs_thread.h"
#include "libochusha/bulletin_board.h"

#include "bbs_thread_view.h"
#include "controller.h"
#include "preferences.h"
#include "session_client.h"

#include "libochushagtk/arrow_action.h"
#include "libochushagtk/combo_action.h"
#include "libochushagtk/hvpaned.h"
#include "libochushagtk/misc_utils.h"
#include "libochushagtk/searchable.h"
#include "libochushagtk/searchbar.h"
#include "libochushagtk/stock.h"


#include <gtk/gtk.h>
#include <glib.h>

#include <cstdlib>

#include <unistd.h>


namespace Ochusha
{

GQuark ViewManager::bulletin_board_id
  = g_quark_from_static_string("ViewManager::BulletinBoard");

GQuark ViewManager::bbs_thread_id
  = g_quark_from_static_string("ViewManager::BBSThread");


enum {
  VIEW_WIDGET = 0,
  NOTEBOOK_ITEM_N_COLUMNS
};


static void
combo_changed_cb(OchushaComboAction *action,
		 GtkTreeModel *model, GtkTreeIter *iter, GtkTreePath *path,
		 ModNotebook *notebook)
{
  GtkWidget *view = NULL;
  gint page;

  gtk_tree_model_get(model, iter, VIEW_WIDGET, &view, -1);
  g_return_if_fail(view != NULL);

  page = mod_notebook_page_num(notebook, view);
  g_return_if_fail(page != -1);
  mod_notebook_set_current_page(notebook, page);
}


static void
set_label_text(GtkMenuItem *item, GtkLabel *label,
	       GtkTreeModel *model, GtkTreeIter *iter,
	       gpointer unused)
{
  GtkWidget *widget = NULL;
  const gchar *menu_title = NULL;

  gtk_tree_model_get(model, iter, VIEW_WIDGET, &widget, -1);
  if (widget != NULL)
    {
      BBSThread *thread = ViewManager::get_thread(widget);
      if (thread != NULL)
	menu_title = thread->get_title();
      else
	{
	  BulletinBoard *board = ViewManager::get_board(widget);
	  if (board != NULL)
	    menu_title = board->get_name();
	}
    }

  if (menu_title == NULL)
    menu_title = _("Anonymous");

  gtk_label_set_text(label, menu_title);
}


static void
set_combo_text(GtkCellLayout *layout, GtkCellRenderer *renderer,
	       GtkTreeModel *model, GtkTreeIter *iter, gpointer unused)
{
  GtkWidget *widget = NULL;
  const gchar *menu_title = NULL;

  gtk_tree_model_get(model, iter, VIEW_WIDGET, &widget, -1);
  if (widget != NULL)
    {
      BBSThread *thread = ViewManager::get_thread(widget);
      if (thread != NULL)
	menu_title = thread->get_title();
      else
	{
	  BulletinBoard *board = ViewManager::get_board(widget);
	  if (board != NULL)
	    menu_title = board->get_name();
	}
    }

  if (menu_title == NULL)
    menu_title = _("Anonymous");

  g_object_set(renderer, "text", menu_title, NULL);
}


ViewManager::ViewManager(Controller &controller)
  : controller(controller),
    prefs(controller.get_prefs()),
    contents_pane(NULL),
    number_of_boards(0),
    number_of_threads(0),
    editor(controller),
    current_focus_widget(NULL),
    last_event_widget(NULL),
    board_search_message_id(0),
    board_search_response_id(0),
    board_search_request_id(0),
    thread_search_message_id(0),
    thread_search_response_id(0),
    thread_search_request_id(0)
{
  static GtkActionEntry entries[] = {
    {
      "CloseThisPage", GTK_STOCK_CLOSE,
      N_("_Close This Page"), "<ctrl>w",
      N_("Close This Page"),
      reinterpret_cast<GCallback>(close_this_page_cb)
    },
    {
      "BookmarkThisPage", OCHUSHA_STOCK_ADD_BOOKMARK,
      N_("_Bookmark This Page"), "",
      N_("Bookmark This Page"),
      reinterpret_cast<GCallback>(bookmark_this_page_cb)
    },
    {
      "FindInThisPage", GTK_STOCK_FIND,
      N_("_Find in This Page"), "",
      N_("Find in This Page"),
      reinterpret_cast<GCallback>(find_in_this_page_cb)
    },
    {
      "RefreshThisPage", GTK_STOCK_REFRESH,
      N_("_Refresh This Page"), "",
      N_("Refresh This Page"),
      reinterpret_cast<GCallback>(refresh_this_page_cb)
    },
    {
#ifdef GTK_STOCK_INFO
      "PageInfo", GTK_STOCK_INFO,
#else
      "PageInfo", GTK_STOCK_PROPERTIES,
#endif
      N_("Page _Info"), NULL,
      N_("Page Info"),
      reinterpret_cast<GCallback>(page_info_cb)
    },
    {
      "CloseCurrentThreadlist", GTK_STOCK_CLOSE,
      N_("Close Current Thread_list"), "",
      N_("Close Current Threadlist"),
      reinterpret_cast<GCallback>(close_board_cb)
    },
    {
      "CloseCurrentThread", GTK_STOCK_CLOSE,
      N_("Close Current _Thread"), "",
      N_("Close Current Thread"),
      reinterpret_cast<GCallback>(close_thread_cb)
    },
    {
      "SearchThreadlist", GTK_STOCK_FIND,
      N_("Search Thread_list"), "",
      N_("Search Threadlist"),
      reinterpret_cast<GCallback>(search_board_cb)
    },
    {
      "SearchThread", GTK_STOCK_FIND,
      N_("Search _Thread"), "",
      N_("Search Thread"),
      reinterpret_cast<GCallback>(search_thread_cb)
    },
    {
      "RefreshThreadlist", GTK_STOCK_REFRESH,
      N_("Refresh Thread_list"), "",
      N_("Refresh Threadlist"),
      reinterpret_cast<GCallback>(refresh_board_cb)
    },
    {
      "RefreshThread", GTK_STOCK_REFRESH,
      N_("Refresh _Thread"), "",
      N_("Refresh Thread"),
      reinterpret_cast<GCallback>(refresh_thread_cb)
    },
    {
      "AutoRefreshThread", NULL,
      N_("Auto Refresh Thread"), NULL,
      NULL,
      reinterpret_cast<GCallback>(auto_refresh_thread_cb)
    },
    {
      "BookmarkCurrentBoard", OCHUSHA_STOCK_ADD_BOOKMARK,
      N_("Bookmark Current Boar_d"), NULL,
      N_("Bookmark Current Board"),
      reinterpret_cast<GCallback>(bookmark_board_cb)
    },
    {
      "BookmarkCurrentThread", OCHUSHA_STOCK_ADD_BOOKMARK,
      N_("Bookmark Current _Thread"), NULL,
      N_("Bookmark Current Thread"),
      reinterpret_cast<GCallback>(bookmark_thread_cb)
    },
    {
#ifdef GTK_STOCK_INFO
      "BoardInfo", GTK_STOCK_INFO,
#else
      "BoardInfo", GTK_STOCK_PROPERTIES,
#endif
      N_("_Board Information"), NULL,
      N_("Board Information"),
      reinterpret_cast<GCallback>(board_info_cb)
    },
    {
#ifdef GTK_STOCK_INFO
      "ThreadInfo", GTK_STOCK_INFO,
#else
      "ThreadInfo", GTK_STOCK_PROPERTIES,
#endif
      N_("_Thread Information"), NULL,
      N_("Thread Information"),
      reinterpret_cast<GCallback>(thread_info_cb)
    },
    {
      "BookmarkTab", OCHUSHA_STOCK_ADD_BOOKMARK,
      N_("Bookmark This _Tab"), NULL,
      N_("Bookmark This Tab"),
      reinterpret_cast<GCallback>(bookmark_tab_cb)
    },
    {
      "RefreshTab", GTK_STOCK_REFRESH,
      N_("_Refresh This Tab"), "",
      N_("Refresh This Tab"),
      reinterpret_cast<GCallback>(refresh_tab_cb)
    },
    {
      "CloseTab", GTK_STOCK_CLOSE,
      N_("_Close This Tab"), "",
      N_("Close This Tab"),
      reinterpret_cast<GCallback>(close_tab_cb)
    },
    {
      "AutoRefreshThreadTab", NULL,
      N_("Auto Refresh This Tab"), NULL,
      NULL,
      reinterpret_cast<GCallback>(auto_refresh_thread_tab_cb)
    },
#if 1	// ʲdeplicated
    {
      "BookmarkBoardTab", OCHUSHA_STOCK_ADD_BOOKMARK,
      N_("Bookmark This _Tab"), NULL,
      N_("Bookmark This Tab"),
      reinterpret_cast<GCallback>(bookmark_board_tab_cb)
    },
    {
      "RefreshBoardTab", GTK_STOCK_REFRESH,
      N_("_Refresh This Tab"), "",
      N_("Refresh This Tab"),
      reinterpret_cast<GCallback>(refresh_board_tab_cb)
    },
    {
      "CloseBoardTab", GTK_STOCK_CLOSE,
      N_("_Close This Tab"), "",
      N_("Close This Tab"),
      reinterpret_cast<GCallback>(close_board_tab_cb)
    },
    {
      "BookmarkThreadTab", OCHUSHA_STOCK_ADD_BOOKMARK,
      N_("Bookmark This _Tab"), NULL,
      N_("Bookmark This Tab"),
      reinterpret_cast<GCallback>(bookmark_thread_tab_cb)
    },
    {
      "RefreshThreadTab", GTK_STOCK_REFRESH,
      N_("_Refresh This Tab"), "",
      N_("Refresh This Tab"),
      reinterpret_cast<GCallback>(refresh_thread_tab_cb)
    },
    {
      "CloseThreadTab", GTK_STOCK_CLOSE,
      N_("_Close This Tab"), "",
      N_("Close This Tab"),
      reinterpret_cast<GCallback>(close_thread_tab_cb)
    }
#endif
  };
  action_group = gtk_action_group_new("ViewManagerActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group, entries,
			       G_N_ELEMENTS(entries), this);

  board_left
    = GTK_ACTION(ochusha_arrow_action_new("BoardArrowLeft",
					  _("Left Board Page"),
					  _("Go to the Left Board Page"),
					  GTK_ARROW_LEFT,
					  GTK_SHADOW_ETCHED_OUT));
  gtk_action_group_add_action(action_group, board_left);

  board_right
    = GTK_ACTION(ochusha_arrow_action_new("BoardArrowRight",
					  _("Right Board Page"),
					  _("Go to the Right Board Page"),
					  GTK_ARROW_RIGHT,
					  GTK_SHADOW_ETCHED_OUT));
  gtk_action_group_add_action(action_group, board_right);

  thread_left
    = GTK_ACTION(ochusha_arrow_action_new("ThreadArrowLeft",
					  _("Left Thread Page"),
					  _("Go to the Left Thread Page"),
					  GTK_ARROW_LEFT,
					  GTK_SHADOW_ETCHED_OUT));
  gtk_action_group_add_action(action_group, thread_left);

  thread_right
    = GTK_ACTION(ochusha_arrow_action_new("ThreadArrowRight",
					  _("Right Thread Page"),
					  _("Go to the Right Thread Page"),
					  GTK_ARROW_RIGHT,
					  GTK_SHADOW_ETCHED_OUT));
  gtk_action_group_add_action(action_group, thread_right);


  board_entries = gtk_list_store_new(NOTEBOOK_ITEM_N_COLUMNS, G_TYPE_POINTER);

  OchushaComboAction *combo_action
    = ochusha_combo_action_new_with_model("ThreadlistTabs",
					  _("Thread_list Tabs"),
					  _("Threadlist Tabs"),
					  NULL,
					  GTK_TREE_MODEL(board_entries));
  ochusha_combo_action_set_menu_data_func(combo_action,
					  set_label_text, this, NULL);
  ochusha_combo_action_set_cell_data_func(combo_action,
					  set_combo_text, this, NULL);

  board_combo = GTK_ACTION(combo_action);
  gtk_action_group_add_action(action_group, board_combo);


  thread_entries = gtk_list_store_new(NOTEBOOK_ITEM_N_COLUMNS, G_TYPE_POINTER);

  combo_action
    = ochusha_combo_action_new_with_model("ThreadTabs",
					  _("_Thread Tabs"),
					  _("Thread Tabs"),
					  NULL,
					  GTK_TREE_MODEL(thread_entries));
  ochusha_combo_action_set_menu_data_func(combo_action,
					  set_label_text, this, NULL);
  ochusha_combo_action_set_cell_data_func(combo_action,
					  set_combo_text, this, NULL);

  thread_combo = GTK_ACTION(combo_action);
  gtk_action_group_add_action(action_group, thread_combo);


  gtk_ui_manager_insert_action_group(controller.get_ui_manager(),
				     action_group, 0);

  close_this_page = gtk_action_group_get_action(action_group, "CloseThisPage");
  bookmark_this_page = gtk_action_group_get_action(action_group,
						   "BookmarkThisPage");
  find_in_this_page = gtk_action_group_get_action(action_group,
						  "FindInThisPage");
  refresh_this_page = gtk_action_group_get_action(action_group,
						  "RefreshThisPage");
  page_info = gtk_action_group_get_action(action_group, "PageInfo");

  close_board = gtk_action_group_get_action(action_group,
					    "CloseCurrentThreadlist");
  close_thread = gtk_action_group_get_action(action_group,
					     "CloseCurrentThread");
  search_board = gtk_action_group_get_action(action_group,
					     "SearchThreadlist");
  search_thread = gtk_action_group_get_action(action_group,
					      "SearchThread");
  refresh_board = gtk_action_group_get_action(action_group,
					      "RefreshThreadlist");
  refresh_thread = gtk_action_group_get_action(action_group,
					       "RefreshThread");
  auto_refresh_thread = gtk_action_group_get_action(action_group,
						    "AutoRefreshThread");
  bookmark_board = gtk_action_group_get_action(action_group,
					       "BookmarkCurrentBoard");
  bookmark_thread = gtk_action_group_get_action(action_group,
						"BookmarkCurrentThread");
  board_info = gtk_action_group_get_action(action_group, "BoardInfo");
  thread_info = gtk_action_group_get_action(action_group, "ThreadInfo");
}


ViewManager::~ViewManager()
{
  g_object_unref(thread_combo);
  g_object_unref(thread_entries);

  g_object_unref(board_combo);
  g_object_unref(board_entries);

  g_object_unref(thread_right);
  g_object_unref(thread_left);
  g_object_unref(board_right);
  g_object_unref(board_left);

  g_object_unref(action_group);
}


void
ViewManager::notify_position_cb(GtkPaned *paned, GParamSpec *arg1,
				ViewManager *view_manager)
{
#if 0
  fprintf(stderr, "ViewManager::notify_position_cb: position=%d\n",
	  gtk_paned_get_position(paned));
#endif
}


void
ViewManager::notify_position_set_cb(GtkPaned *paned, GParamSpec *arg1,
				    ViewManager *view_manager)
{
#if 0
  fprintf(stderr, "ViewManager::notify_position_set_cb: position=%d\n",
	  gtk_paned_get_position(paned));
#endif
  char value[64];
  snprintf(value, 64, "%d", gtk_paned_get_position(paned));
  SessionClient &session = view_manager->controller.get_session();
  if (static_cast<gpointer>(paned) == static_cast<gpointer>(view_manager->contents_pane))
    session.set_state("ViewManager::board_pane_size", value);
  else
    session.set_state("ViewManager::thread_pane_size", value);
}


GtkWidget *
ViewManager::create_contents_pane()
{
  g_return_val_if_fail(contents_pane == NULL, contents_pane);


  contents_pane = ochusha_hvpaned_new(prefs.get_vertical_placement()
				      ? OCHUSHA_VPANED_STYLE
				      : OCHUSHA_HPANED_STYLE);
  g_signal_connect(contents_pane, "destroy",
		   reinterpret_cast<GCallback>(gtk_widget_destroyed),
		   &contents_pane);
  g_signal_connect(contents_pane, "notify::position",
		   reinterpret_cast<GCallback>(notify_position_cb), this);
  g_signal_connect(contents_pane, "notify::position-set",
		   reinterpret_cast<GCallback>(notify_position_set_cb), this);


  GtkWidget *open_url_tool_item
    = gtk_ui_manager_get_widget(controller.get_ui_manager(),
				"/NavigationToolbar/OpenURLEntry");
  current_url_entry = NULL;
  if (open_url_tool_item != NULL && GTK_IS_TOOL_ITEM(open_url_tool_item))
    {
      GtkWidget *child = gtk_bin_get_child(GTK_BIN(open_url_tool_item));
      if (GTK_IS_COMBO_BOX_ENTRY(child))
	{
	  current_url_entry = GTK_ENTRY(gtk_bin_get_child(GTK_BIN(child)));
	  g_signal_connect(current_url_entry, "activate",
			   reinterpret_cast<GCallback>(current_url_entry_activate_cb),
			   this);
	}
    }


  // ()ڥν
  board_box = gtk_vbox_new(FALSE, 0);
  // gtk_container_set_border_width(GTK_CONTAINER(board_box), 0);

  board_toolbar = gtk_ui_manager_get_widget(controller.get_ui_manager(),
					    "/BoardViewToolbar");
  if (board_toolbar != NULL)
    {
      GtkToolbar *toolbar = GTK_TOOLBAR(board_toolbar);
      gtk_toolbar_set_style(toolbar, GTK_TOOLBAR_ICONS);

      gtk_box_pack_start(GTK_BOX(board_box), board_toolbar, FALSE, TRUE, 0);
    }


  GtkWidget *tmp_widget = mod_notebook_new();
  g_signal_connect(tmp_widget, "switch_page",
		   reinterpret_cast<GCallback>(board_switch_page_cb), this);
  g_signal_connect(tmp_widget, "button_press_event",
		   reinterpret_cast<GCallback>(board_button_press_event_cb),
		   this);
  g_signal_connect(tmp_widget, "close_button_clicked",
		   reinterpret_cast<GCallback>(close_board_cb),
		   this);
  board_book = MOD_NOTEBOOK(tmp_widget);
  mod_notebook_set_scrollable(board_book, TRUE);
  if (board_toolbar == NULL
      || (gtk_ui_manager_get_widget(controller.get_ui_manager(),
				    "/BoardViewToolbar/CloseCurrentThreadlist")
	  == NULL))
    mod_notebook_close_button_enable(board_book);
  else
    mod_notebook_close_button_disable(board_book);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(board_box), tmp_widget, TRUE, TRUE, 0);
  g_signal_connect(board_book, "destroy",
		   reinterpret_cast<GCallback>(gtk_widget_destroyed),
		   &board_book);


  g_signal_connect_swapped(board_left, "activate",
			   reinterpret_cast<GCallback>(mod_notebook_prev_page),
			   board_book);
  g_signal_connect_swapped(board_right, "activate",
			   reinterpret_cast<GCallback>(mod_notebook_next_page),
			   board_book);

  g_signal_connect(board_combo, "changed",
		   reinterpret_cast<GCallback>(combo_changed_cb), board_book);

  gtk_widget_show(board_box);
  gtk_paned_add1(GTK_PANED(contents_pane), board_box);



  /* ɽڥν */
  thread_pane = gtk_vpaned_new();
  gtk_container_set_border_width(GTK_CONTAINER(thread_pane), 0);

  g_signal_connect(contents_pane, "notify::position",
		   reinterpret_cast<GCallback>(notify_position_cb), this);
  g_signal_connect(contents_pane, "notify::position-set",
		   reinterpret_cast<GCallback>(notify_position_set_cb), this);

  thread_box = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(thread_box), 0);

  thread_toolbar = gtk_ui_manager_get_widget(controller.get_ui_manager(),
					     "/ThreadViewToolbar");
  if (thread_toolbar != NULL)
    {
      gtk_container_set_border_width(GTK_CONTAINER(thread_toolbar), 0);
      GtkToolbar *toolbar = GTK_TOOLBAR(thread_toolbar);
      gtk_toolbar_set_style(toolbar, GTK_TOOLBAR_ICONS);
      gtk_box_pack_start(GTK_BOX(thread_box), thread_toolbar, FALSE, TRUE, 0);
    }


  tmp_widget = mod_notebook_new();
  g_signal_connect(tmp_widget, "switch_page",
		   reinterpret_cast<GCallback>(thread_switch_page_cb), this);
  g_signal_connect(tmp_widget, "button_press_event",
		   reinterpret_cast<GCallback>(thread_button_press_event_cb),
		   this);
  g_signal_connect(tmp_widget, "close_button_clicked",
		   reinterpret_cast<GCallback>(close_thread_cb),
		   this);
  thread_book = MOD_NOTEBOOK(tmp_widget);
  mod_notebook_set_scrollable(thread_book, TRUE);
  if (thread_toolbar == NULL
      || (gtk_ui_manager_get_widget(controller.get_ui_manager(),
				    "/ThreadViewToolbar/CloseCurrentThread")
	  == NULL))
    mod_notebook_close_button_enable(thread_book);
  else
    mod_notebook_close_button_disable(thread_book);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(thread_box), tmp_widget, TRUE, TRUE, 0);
  g_signal_connect(thread_book, "destroy",
		   reinterpret_cast<GCallback>(gtk_widget_destroyed),
		   &thread_book);


  g_signal_connect_swapped(thread_left, "activate",
			   reinterpret_cast<GCallback>(mod_notebook_prev_page),
			   thread_book);
  g_signal_connect_swapped(thread_right, "activate",
			   reinterpret_cast<GCallback>(mod_notebook_next_page),
			   thread_book);

  g_signal_connect(thread_combo, "changed",
		   reinterpret_cast<GCallback>(combo_changed_cb), thread_book);

  gtk_widget_show(thread_box);
  gtk_paned_pack1(GTK_PANED(thread_pane), thread_box, TRUE, TRUE);

  tmp_widget = editor.get_widget();
  if (prefs.get_enable_embeded_response_editor())
    gtk_widget_show(tmp_widget);
  gtk_paned_pack2(GTK_PANED(thread_pane), tmp_widget, FALSE, TRUE);


  gtk_paned_add2(GTK_PANED(contents_pane), thread_pane);


  // 
  mod_notebook_set_tab_shrinkable(board_book,
				  prefs.get_board_tab_shrinkable());
  mod_notebook_set_minimum_tab_label_size(board_book,
					  prefs.get_board_tab_minimum_size());
  if (prefs.get_board_tab_enable_tooltips())
    mod_notebook_tooltips_enable(board_book);
  else
    mod_notebook_tooltips_disable(board_book);

  mod_notebook_set_tab_shrinkable(thread_book,
				  prefs.get_thread_tab_shrinkable());
  mod_notebook_set_minimum_tab_label_size(thread_book,
					  prefs.get_thread_tab_minimum_size());
  if (prefs.get_thread_tab_enable_tooltips())
    mod_notebook_tooltips_enable(thread_book);
  else
    mod_notebook_tooltips_disable(thread_book);

  const char *pane_size
    = controller.get_session().get_state("ViewManager::board_pane_size");
  if (pane_size != NULL)
    {
      int size;
      if (sscanf(pane_size, "%d", &size) == 1)
	gtk_paned_set_position(GTK_PANED(contents_pane), size);
    }
  pane_size
    = controller.get_session().get_state("ViewManager::thread_pane_size");
  if (pane_size != NULL)
    {
      int size;
      if (sscanf(pane_size, "%d", &size) == 1)
	gtk_paned_set_position(GTK_PANED(thread_pane), size);
    }

  g_signal_connect(prefs.vertical_placement, "toggled",
		   reinterpret_cast<GCallback>(vertical_placement_toggled_cb),
		   this);
  g_signal_connect(prefs.board_tab_shrinkable, "toggled",
		   reinterpret_cast<GCallback>(board_tab_shrinkable_toggled_cb),
		   this);
  g_signal_connect(prefs.board_tab_enable_tooltips, "toggled",
		   reinterpret_cast<GCallback>(board_tab_enable_tooltips_toggled_cb), this);
  g_signal_connect(prefs.board_tab_always_show, "toggled",
		   reinterpret_cast<GCallback>(board_tab_always_show_toggled_cb), this);
  g_signal_connect(prefs.board_pane_always_show, "toggled",
		   reinterpret_cast<GCallback>(board_pane_always_show_toggled_cb), this);
  g_signal_connect(prefs.thread_tab_shrinkable, "toggled",
		   reinterpret_cast<GCallback>(thread_tab_shrinkable_toggled_cb), this);
  g_signal_connect(prefs.thread_tab_enable_tooltips, "toggled",
		   reinterpret_cast<GCallback>(thread_tab_enable_tooltips_toggled_cb), this);
  g_signal_connect(prefs.thread_tab_always_show, "toggled",
		   reinterpret_cast<GCallback>(thread_tab_always_show_toggled_cb), this);
  g_signal_connect(prefs.thread_pane_always_show, "toggled",
		   reinterpret_cast<GCallback>(thread_pane_always_show_toggled_cb), this);

  g_signal_connect(prefs.show_threadlist_toolbar, "toggled",
		   reinterpret_cast<GCallback>(show_threadlist_toolbar_toggled_cb), this);
  gtk_toggle_action_toggled(prefs.show_threadlist_toolbar);
  g_signal_connect(prefs.show_thread_toolbar, "toggled",
		   reinterpret_cast<GCallback>(show_thread_toolbar_toggled_cb),
		   this);
  gtk_toggle_action_toggled(prefs.show_thread_toolbar);

  g_signal_connect(prefs.enable_embeded_response_editor, "toggled",
		   reinterpret_cast<GCallback>(enable_embeded_response_editor_toggled_cb), this);


  GDK_THREADS_LEAVE();

  prefs.prefs_changed.connect(this, &ViewManager::prefs_changed_cb);

  GDK_THREADS_ENTER();


  update_board_pane();
  update_board_tab();
  update_board_navigator();
  update_thread_pane();
  update_thread_tab();
  update_thread_navigator();

  return contents_pane;
}


GtkWidget *
ViewManager::get_widget()
{
  if (contents_pane != NULL)
    return contents_pane;
  return create_contents_pane();
}


void
ViewManager::current_url_entry_activate_cb(GtkEntry *entry,
					   ViewManager *view_manager)
{
  const gchar *url = gtk_entry_get_text(entry);
  bool in_tab = view_manager->controller.get_prefs().get_default_open_in_tab();
  view_manager->controller.open(url, in_tab, false);
}


void
ViewManager::update_board_pane()
{
  if (number_of_boards == 0 && !prefs.get_board_pane_always_show())
    {
      gtk_widget_set_child_visible(thread_pane, TRUE);
      gtk_widget_set_sensitive(board_box, FALSE);
      gtk_widget_hide(board_box);
    }
  else
    {
      gtk_widget_set_sensitive(board_box, TRUE);
      gtk_widget_show(board_box);
    }

  if (number_of_boards == 0)
    {
      gtk_action_set_sensitive(close_board, FALSE);
      gtk_action_set_sensitive(search_board, FALSE);
      gtk_action_set_sensitive(refresh_board, FALSE);
      gtk_action_set_sensitive(bookmark_board, FALSE);
      gtk_action_set_sensitive(board_info, FALSE);
    }
  else
    {
      gtk_action_set_sensitive(close_board, TRUE);
      gtk_action_set_sensitive(search_board, TRUE);
      gtk_action_set_sensitive(refresh_board, TRUE);
      gtk_action_set_sensitive(bookmark_board, TRUE);
      gtk_action_set_sensitive(board_info, TRUE);
    }

  gboolean sensitive = number_of_boards + number_of_threads > 0;
  gtk_action_set_sensitive(close_this_page, sensitive);
  gtk_action_set_sensitive(bookmark_this_page, sensitive);
  gtk_action_set_sensitive(find_in_this_page, sensitive);
  gtk_action_set_sensitive(refresh_this_page, sensitive);
  gtk_action_set_sensitive(page_info, sensitive);
}


void
ViewManager::update_board_tab()
{
  if (number_of_boards < 2 && !prefs.get_board_tab_always_show())
    mod_notebook_set_show_tabs(board_book, FALSE);
  else
    mod_notebook_set_show_tabs(board_book, TRUE);
}


void
ViewManager::update_board_navigator()
{
  gint page = mod_notebook_get_current_page(board_book);

  gtk_action_set_sensitive(board_left, page > 0);
  gtk_action_set_sensitive(board_right,
			   page != -1
			   && static_cast<guint>(page) < number_of_boards - 1);
}


void
ViewManager::update_thread_pane()
{
  if (number_of_threads == 0 && !prefs.get_thread_pane_always_show())
    {
      gtk_widget_set_child_visible(board_box, TRUE);
      gtk_widget_set_sensitive(thread_pane, FALSE);
      gtk_widget_hide(thread_pane);
    }
  else
    {
      gtk_widget_set_sensitive(thread_pane, TRUE);
      gtk_widget_show(thread_pane);
    }

  if (number_of_threads == 0)
    {
      gtk_action_set_sensitive(close_thread, FALSE);
      gtk_action_set_sensitive(search_thread, FALSE);
      gtk_action_set_sensitive(refresh_thread, FALSE);
      gtk_action_set_sensitive(bookmark_thread, FALSE);
      gtk_action_set_sensitive(thread_info, FALSE);
      gtk_action_set_sensitive(auto_refresh_thread, FALSE);
      editor.set_thread(NULL);
    }
  else
    {
      gtk_action_set_sensitive(close_thread, TRUE);
      gtk_action_set_sensitive(search_thread, TRUE);
      gtk_action_set_sensitive(refresh_thread, TRUE);
      gtk_action_set_sensitive(bookmark_thread, TRUE);
      gtk_action_set_sensitive(thread_info, TRUE);
      gtk_action_set_sensitive(auto_refresh_thread, TRUE);
      if (prefs.get_enable_embeded_response_editor())
	editor.set_thread(get_current_thread());
    }

  gboolean sensitive = number_of_boards + number_of_threads > 0;
  gtk_action_set_sensitive(close_this_page, sensitive);
  gtk_action_set_sensitive(bookmark_this_page, sensitive);
  gtk_action_set_sensitive(find_in_this_page, sensitive);
  gtk_action_set_sensitive(refresh_this_page, sensitive);
  gtk_action_set_sensitive(page_info, sensitive);
}


void
ViewManager::update_thread_tab()
{
  if (number_of_threads < 2 && !prefs.get_thread_tab_always_show())
    mod_notebook_set_show_tabs(thread_book, FALSE);
  else
    mod_notebook_set_show_tabs(thread_book, TRUE);
}


void
ViewManager::update_thread_navigator()
{
  gint page = mod_notebook_get_current_page(thread_book);

  gtk_action_set_sensitive(thread_left, page > 0);
  gtk_action_set_sensitive(thread_right,
			   page != -1
			   && static_cast<guint>(page) < number_of_threads - 1);
}


void
ViewManager::close_this_page_cb(GtkAction *action, ViewManager *view_manager)
{
  view_manager->close_current_focus_tab();
}


void
ViewManager::bookmark_this_page_cb(GtkAction *action, ViewManager *view_manager)
{
  if (view_manager->current_focus_widget == NULL)
    return;

  BBSThread *thread = get_thread(view_manager->current_focus_widget);
  if (thread != NULL)
    {
      view_manager->controller.bookmark(thread);
      return;
    }

  BulletinBoard *board = get_board(view_manager->current_focus_widget);
  if (board != NULL)
    view_manager->controller.bookmark(board);
}


void
ViewManager::find_in_this_page_cb(GtkAction *action, ViewManager *view_manager)
{
  if (view_manager->current_focus_widget == NULL)
    return;

  BBSThread *thread = get_thread(view_manager->current_focus_widget);
  if (thread != NULL)
    {
      if (view_manager->thread_search_response_id == 0)
	view_manager->setup_thread_search();
      return;
    }

  BulletinBoard *board = get_board(view_manager->current_focus_widget);
  if (board != NULL && view_manager->board_search_response_id == 0)
    view_manager->setup_board_search();
}


void
ViewManager::refresh_this_page_cb(GtkAction *action, ViewManager *view_manager)
{
  if (view_manager->current_focus_widget == NULL)
    return;

  BBSThread *thread = get_thread(view_manager->current_focus_widget);
  if (thread != NULL)
    {
      view_manager->controller.refresh(thread);
      return;
    }

  BulletinBoard *board = get_board(view_manager->current_focus_widget);
  if (board != NULL)
    view_manager->controller.refresh(board, true);
}


void
ViewManager::page_info_cb(GtkAction *action, ViewManager *view_manager)
{
  if (view_manager->current_focus_widget == NULL)
    return;

  BBSThread *thread = get_thread(view_manager->current_focus_widget);
  if (thread != NULL)
    {
      view_manager->controller.get_thread_ui().open_properties(thread);
      return;
    }

  BulletinBoard *board = get_board(view_manager->current_focus_widget);
  if (board != NULL)
    view_manager->controller.get_boardlist_ui().open_properties(board);
}


void
ViewManager::close_board_cb(GtkAction *action, ViewManager *view_manager)
{
  gint page = mod_notebook_get_current_page(view_manager->board_book);
  if (page == -1)
    return;

  view_manager->close_nth_board_page(page);

  view_manager->update_board_pane();
  view_manager->update_board_tab();
  view_manager->update_board_navigator();
}


void
ViewManager::close_thread_cb(GtkAction *action, ViewManager *view_manager)
{
  gint page = mod_notebook_get_current_page(view_manager->thread_book);
  if (page == -1)
    return;

  view_manager->close_nth_thread_page(page);

  view_manager->update_thread_pane();
  view_manager->update_thread_tab();
  view_manager->update_thread_navigator();
}


void
ViewManager::board_search_response_cb(GtkWidget *widget, int response_id,
				      ViewManager *view_manager)
{
  g_assert(view_manager->board_search_message_id != 0);
  g_assert(view_manager->board_search_response_id != 0);
  g_assert(view_manager->board_search_request_id != 0);

  switch (response_id)
    {
    case GTK_RESPONSE_CANCEL:
      gtk_widget_hide(widget);
      gtk_statusbar_remove(view_manager->controller.get_statusbar(),
			   view_manager->controller.get_searchbar_context_id(),
			   view_manager->board_search_message_id);
      g_signal_handler_disconnect(widget,
				  view_manager->board_search_response_id);
      g_signal_handler_disconnect(widget,
				  view_manager->board_search_request_id);
      view_manager->board_search_message_id = 0;
      view_manager->board_search_response_id = 0;
      view_manager->board_search_request_id = 0;
      return;
    }
}


OchushaSearchable *
ViewManager::board_search_request_cb(gpointer unused, ViewManager *view_manager)
{
  GtkWidget *bin = view_manager->get_current_board_view();
  if (bin == NULL)
    return NULL;

  // ad-hocǾ𤱤ʤݤʤΤGtkScrolledWindowäƤȷǤ
  return OCHUSHA_SEARCHABLE(gtk_bin_get_child(GTK_BIN(bin)));
}


void
ViewManager::setup_board_search()
{
  GtkWidget *bar_widget = controller.get_searchbar();
  ochusha_searchbar_response(OCHUSHA_SEARCHBAR(bar_widget),
			     GTK_RESPONSE_CANCEL);

  g_assert(board_search_response_id == 0);
  g_assert(board_search_request_id == 0);
  g_assert(board_search_message_id == 0);

  board_search_response_id
    = g_signal_connect(bar_widget, "response",
		       reinterpret_cast<GCallback>(board_search_response_cb),
		       this);
  board_search_request_id
    = g_signal_connect(bar_widget, "search_request",
		       reinterpret_cast<GCallback>(board_search_request_cb),
		       this);
  board_search_message_id
    = gtk_statusbar_push(controller.get_statusbar(),
			 controller.get_searchbar_context_id(),
			 _("Find in Threadlist"));
  gtk_widget_show(bar_widget);
  ochusha_searchbar_set_key_selected(OCHUSHA_SEARCHBAR(bar_widget));
}


void
ViewManager::search_board_cb(GtkAction *action, ViewManager *view_manager)
{
  if (view_manager->board_search_response_id != 0)
    return;
  view_manager->setup_board_search();
}


void
ViewManager::thread_search_response_cb(GtkWidget *widget, int response_id,
				       ViewManager *view_manager)
{
  g_assert(view_manager->thread_search_message_id != 0);
  g_assert(view_manager->thread_search_response_id != 0);
  g_assert(view_manager->thread_search_request_id != 0);

  switch (response_id)
    {
    case GTK_RESPONSE_CANCEL:
      gtk_widget_hide(widget);
      gtk_statusbar_remove(view_manager->controller.get_statusbar(),
			   view_manager->controller.get_searchbar_context_id(),
			   view_manager->thread_search_message_id);
      g_signal_handler_disconnect(widget,
				  view_manager->thread_search_response_id);
      g_signal_handler_disconnect(widget,
				  view_manager->thread_search_request_id);
      view_manager->thread_search_message_id = 0;
      view_manager->thread_search_response_id = 0;
      view_manager->thread_search_request_id = 0;
      return;
    }
}


OchushaSearchable *
ViewManager::thread_search_request_cb(gpointer unused, ViewManager *view_manager)
{
  GtkWidget *bin = view_manager->get_current_thread_view();
  if (bin == NULL)
    return NULL;

  // ad-hocǾ𤱤ʤݤʤΤGtkScrolledWindowäƤȷǤ
  return OCHUSHA_SEARCHABLE(gtk_bin_get_child(GTK_BIN(bin)));
}


void
ViewManager::setup_thread_search()
{
  GtkWidget *bar_widget = controller.get_searchbar();
  ochusha_searchbar_response(OCHUSHA_SEARCHBAR(bar_widget),
			     GTK_RESPONSE_CANCEL);

  g_assert(thread_search_response_id == 0);
  g_assert(thread_search_request_id == 0);
  g_assert(thread_search_message_id == 0);

  thread_search_response_id
    = g_signal_connect(bar_widget, "response",
		       reinterpret_cast<GCallback>(thread_search_response_cb),
		       this);
  thread_search_request_id
    = g_signal_connect(bar_widget, "search_request",
		       reinterpret_cast<GCallback>(thread_search_request_cb),
		       this);
  thread_search_message_id
    = gtk_statusbar_push(controller.get_statusbar(),
			 controller.get_searchbar_context_id(),
			 _("Find in Thread"));
  gtk_widget_show(bar_widget);
  ochusha_searchbar_set_key_selected(OCHUSHA_SEARCHBAR(bar_widget));
}


void
ViewManager::search_thread_cb(GtkAction *action, ViewManager *view_manager)
{
  if (view_manager->thread_search_response_id != 0)
    return;
  view_manager->setup_thread_search();
}


void
ViewManager::refresh_board_cb(GtkAction *action, ViewManager *view_manager)
{
  BulletinBoard *board = view_manager->get_current_board();
  if (board == NULL)
    return;
  view_manager->controller.refresh(board, true);
}


void
ViewManager::refresh_thread_cb(GtkAction *action, ViewManager *view_manager)
{
  BBSThread *thread = view_manager->get_current_thread();
  if (thread == NULL)
    return;
  view_manager->controller.refresh(thread);
}


void
ViewManager::auto_refresh_thread_cb(GtkAction *action,
				    ViewManager *view_manager)
{
  BBSThread *thread = view_manager->get_current_thread();
  if (thread == NULL)
    return;
  view_manager->controller.refresh(thread, true);
}


void
ViewManager::bookmark_board_cb(GtkAction *action, ViewManager *view_manager)
{
  BulletinBoard *board = view_manager->get_current_board();
  if (board == NULL)
    return;
  view_manager->controller.bookmark(board);
}


void
ViewManager::bookmark_thread_cb(GtkAction *action, ViewManager *view_manager)
{
  BBSThread *thread = view_manager->get_current_thread();
  if (thread == NULL)
    return;
  view_manager->controller.bookmark(thread);
}


void
ViewManager::board_info_cb(GtkAction *action, ViewManager *view_manager)
{
  BulletinBoard *board = view_manager->get_current_board();
  if (board == NULL)
    return;
  view_manager->controller.get_boardlist_ui().open_properties(board);
}


void
ViewManager::thread_info_cb(GtkAction *action, ViewManager *view_manager)
{
  BBSThread *thread = view_manager->get_current_thread();
  if (thread == NULL)
    return;
  view_manager->controller.get_thread_ui().open_properties(thread);
}


void
ViewManager::bookmark_tab_cb(GtkAction *action, ViewManager *view_manager)
{
  g_return_if_fail(view_manager->last_event_widget != NULL);
  BulletinBoard *board = get_board(view_manager->last_event_widget);
  if (board != NULL)
    {
      view_manager->controller.bookmark(board);
      return;
    }
  BBSThread *thread = get_thread(view_manager->last_event_widget);
  if (thread != NULL)
    {
      view_manager->controller.bookmark(thread);
      return;
    }
  return;
}


void
ViewManager::refresh_tab_cb(GtkAction *action, ViewManager *view_manager)
{
  g_return_if_fail(view_manager->last_event_widget != NULL);
  BulletinBoard *board = get_board(view_manager->last_event_widget);
  if (board != NULL)
    {
      view_manager->controller.refresh(board, true);
      return;
    }
  BBSThread *thread = get_thread(view_manager->last_event_widget);
  if (thread != NULL)
    {
      view_manager->controller.refresh(thread);
      return;
    }
  return;
}


void
ViewManager::close_tab_cb(GtkAction *action, ViewManager *view_manager)
{
  g_return_if_fail(view_manager->last_event_widget != NULL);
  gint page = mod_notebook_page_num(view_manager->board_book,
				    view_manager->last_event_widget);
  if (page != -1)
    {
      view_manager->close_nth_board_page(page);
      view_manager->update_board_pane();
      view_manager->update_board_tab();
      view_manager->update_board_navigator();
      return;
    }

  page = mod_notebook_page_num(view_manager->thread_book,
			       view_manager->last_event_widget);
  if (page != -1)
    {
      view_manager->close_nth_thread_page(page);
      view_manager->update_thread_pane();
      view_manager->update_thread_tab();
      view_manager->update_thread_navigator();
      return;
    }

  return;
}


void
ViewManager::auto_refresh_thread_tab_cb(GtkAction *action,
					ViewManager *view_manager)
{
  g_return_if_fail(view_manager->last_event_widget != NULL);
  BBSThread *thread = get_thread(view_manager->last_event_widget);
  if (thread != NULL)
    {
      view_manager->controller.refresh(thread, true);
      return;
    }
  return;
}


#if 1	// ʲdeplicated
void
ViewManager::bookmark_board_tab_cb(GtkAction *action, ViewManager *view_manager)
{
  fprintf(stderr, _("The action `BookmarkBoardTab' is obsoleted by `BookmarkTab'.  Please update your UI definition.\n"));
  bookmark_tab_cb(action, view_manager);
}


void
ViewManager::refresh_board_tab_cb(GtkAction *action, ViewManager *view_manager)
{
  fprintf(stderr, _("The action `RefreshBoardTab' is obsoleted by `RefreshTab'.  Please update your UI definition.\n"));
  refresh_tab_cb(action, view_manager);
}


void
ViewManager::close_board_tab_cb(GtkAction *action, ViewManager *view_manager)
{
  fprintf(stderr, _("The action `CloseBoardTab' is obsoleted by `CloseTab'.  Please update your UI definition.\n"));
  close_tab_cb(action, view_manager);
}


void
ViewManager::bookmark_thread_tab_cb(GtkAction *action,
				    ViewManager *view_manager)
{
  fprintf(stderr, _("The action `BookmarkThreadTab' is obsoleted by `BookmarkTab'.  Please update your UI definition.\n"));
  bookmark_tab_cb(action, view_manager);
}


void
ViewManager::refresh_thread_tab_cb(GtkAction *action, ViewManager *view_manager)
{
  fprintf(stderr, _("The action `RefreshThreadTab' is obsoleted by `RefreshTab'.  Please update your UI definition.\n"));
  refresh_tab_cb(action, view_manager);
}


void
ViewManager::close_thread_tab_cb(GtkAction *action, ViewManager *view_manager)
{
  fprintf(stderr, _("The action `CloseThreadTab' is obsoleted by `CloseTab'.  Please update your UI definition.\n"));
  close_tab_cb(action, view_manager);
}
#endif


void
ViewManager::vertical_placement_toggled_cb(GtkToggleAction *action,
					   ViewManager *view_manager)
{
  ochusha_hvpaned_set_pane_style(OCHUSHA_HVPANED(view_manager->contents_pane),
				 gtk_toggle_action_get_active(action)
				 ? OCHUSHA_VPANED_STYLE : OCHUSHA_HPANED_STYLE);
}


void
ViewManager::board_tab_shrinkable_toggled_cb(GtkToggleAction *action,
					     ViewManager *view_manager)
{
  mod_notebook_set_tab_shrinkable(view_manager->board_book,
				  gtk_toggle_action_get_active(action));
}


void
ViewManager::board_tab_enable_tooltips_toggled_cb(GtkToggleAction *action,
						  ViewManager *view_manager)
{
  if (gtk_toggle_action_get_active(action))
    mod_notebook_tooltips_enable(view_manager->board_book);
  else
    mod_notebook_tooltips_disable(view_manager->board_book);
}


void
ViewManager::board_tab_always_show_toggled_cb(GtkToggleAction *action,
					      ViewManager *view_manager)
{
  view_manager->update_board_tab();
}


void
ViewManager::board_pane_always_show_toggled_cb(GtkToggleAction *action,
					       ViewManager *view_manager)
{
  view_manager->update_board_pane();
}


void
ViewManager::thread_tab_shrinkable_toggled_cb(GtkToggleAction *action,
					      ViewManager *view_manager)
{
  mod_notebook_set_tab_shrinkable(view_manager->thread_book,
				  gtk_toggle_action_get_active(action));
}


void
ViewManager::thread_tab_enable_tooltips_toggled_cb(GtkToggleAction *action,
						   ViewManager *view_manager)
{
  if (gtk_toggle_action_get_active(action))
    mod_notebook_tooltips_enable(view_manager->thread_book);
  else
    mod_notebook_tooltips_disable(view_manager->thread_book);
}


void
ViewManager::thread_tab_always_show_toggled_cb(GtkToggleAction *action,
					       ViewManager *view_manager)
{
  view_manager->update_thread_tab();
}


void
ViewManager::thread_pane_always_show_toggled_cb(GtkToggleAction *action,
						ViewManager *view_manager)
{
  view_manager->update_thread_pane();
}


void
ViewManager::show_threadlist_toolbar_toggled_cb(GtkToggleAction *action,
						ViewManager *view_manager)
{
  if (view_manager->board_toolbar == NULL)
    return;

  gboolean active = gtk_toggle_action_get_active(action);
  if (active)
    gtk_widget_show(view_manager->board_toolbar);
  else
    gtk_widget_hide(view_manager->board_toolbar);

  if (!active
      || gtk_ui_manager_get_widget(view_manager->controller
				   .get_ui_manager(),
				   "/BoardViewToolbar/CloseCurrentThreadlist")
      == NULL)
    mod_notebook_close_button_enable(view_manager->board_book);
  else
    mod_notebook_close_button_disable(view_manager->board_book);
}


void
ViewManager::show_thread_toolbar_toggled_cb(GtkToggleAction *action,
					    ViewManager *view_manager)
{
  if (view_manager->thread_toolbar == NULL)
    return;

  gboolean active = gtk_toggle_action_get_active(action);
  if (active)
    gtk_widget_show(view_manager->thread_toolbar);
  else
    gtk_widget_hide(view_manager->thread_toolbar);

  if (!active
      || gtk_ui_manager_get_widget(view_manager->controller
				   .get_ui_manager(),
				   "/ThreadViewToolbar/CloseCurrentThread")
      == NULL)
    mod_notebook_close_button_enable(view_manager->thread_book);
  else
    mod_notebook_close_button_disable(view_manager->thread_book);
}


void
ViewManager::enable_embeded_response_editor_toggled_cb(GtkToggleAction *action,
						       ViewManager *view_manager)
{
  if (gtk_toggle_action_get_active(action))
    {
      view_manager->editor.set_thread(view_manager->get_current_thread());
      gtk_widget_show(view_manager->editor.get_widget());
    }
  else
    {
      view_manager->editor.set_thread(NULL);
      gtk_widget_hide(view_manager->editor.get_widget());
    }
}


void
ViewManager::prefs_changed_cb(Preferences &prefs)
{
  // sigslotʥϥɥ
  GDK_THREADS_ENTER();

  mod_notebook_set_minimum_tab_label_size(board_book,
					  prefs.get_board_tab_minimum_size());
  mod_notebook_set_minimum_tab_label_size(thread_book,
					  prefs.get_thread_tab_minimum_size());

  GDK_THREADS_LEAVE();
}


void
ViewManager::follow_focus()
{
  if (prefs.get_enable_embeded_response_editor())
    editor.set_thread(get_current_thread());
  if (ochusha_find_focus_widget(board_box) != NULL)
    {
      GtkWidget *widget = get_current_board_view();
      set_current_focus_widget(widget);
      return;
    }
  if (ochusha_find_focus_widget(thread_box) != NULL)
    {
      GtkWidget *widget	= get_current_thread_view();
      set_current_focus_widget(widget);
      return;
    }
}


void
ViewManager::close_current_focus_tab()
{
  GtkWidget *widget = get_current_board_view();
  if (gtk_widget_is_focus(GTK_WIDGET(board_book))
      || (widget != NULL && widget == current_focus_widget)
      || ochusha_find_focus_widget(widget) != NULL)
    {
      close_board_cb(close_board, this);
      return;
    }

  widget = get_current_thread_view();
  if (gtk_widget_is_focus(GTK_WIDGET(thread_book))
      || (widget != NULL && widget == current_focus_widget)
      || ochusha_find_focus_widget(widget) != NULL)
    {
      close_thread_cb(close_thread, this);
      return;
    }
}


GtkWidget *
ViewManager::get_current_board_view()
{
  if (board_book == NULL)
    return NULL;

  gint page = mod_notebook_get_current_page(board_book);
  if (page == -1)
    return NULL;

  return mod_notebook_get_nth_page(board_book, page);
}


BulletinBoard *
ViewManager::get_current_board()
{
  if (board_book == NULL)
    return NULL;

  gint page = mod_notebook_get_current_page(board_book);
  if (page == -1)
    return NULL;

  GtkWidget *widget = mod_notebook_get_nth_page(board_book, page);
  return static_cast<BulletinBoard *>(g_object_get_qdata(G_OBJECT(widget),
							 bulletin_board_id));
}
  

GtkWidget *
ViewManager::get_current_thread_view()
{
  if (thread_book == NULL)
    return NULL;

  gint page = mod_notebook_get_current_page(thread_book);
  if (page == -1)
    return NULL;

  return mod_notebook_get_nth_page(thread_book, page);
}


BBSThread *
ViewManager::get_current_thread()
{
  if (thread_book == NULL)
    return NULL;

  gint page = mod_notebook_get_current_page(thread_book);
  if (page == -1)
    return NULL;

  GtkWidget *widget = mod_notebook_get_nth_page(thread_book, page);
  return static_cast<BBSThread *>(g_object_get_qdata(G_OBJECT(widget),
						     bbs_thread_id));
}


struct RemoveEntryArgs
{
  GtkWidget *view;
  gboolean result;
};


static gboolean
remove_entry(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
	     RemoveEntryArgs *args)
{
  GtkWidget *widget = NULL;
  gtk_tree_model_get(model, iter, VIEW_WIDGET, &widget, -1);
  if (widget == args->view)
    {
      g_assert(!args->result);
      gtk_list_store_remove(GTK_LIST_STORE(model), iter);
      args->result = TRUE;
      return TRUE;
    }
  return FALSE;
}


static void
remove_list_entry(GtkListStore *store, GtkWidget *view)
{
  RemoveEntryArgs args = { view, FALSE };
  gtk_tree_model_foreach(GTK_TREE_MODEL(store),
		reinterpret_cast<GtkTreeModelForeachFunc>(remove_entry), &args);
  g_assert(args.result);
}


struct ActivateComboEntryArgs
{
  GtkWidget *view;
  OchushaComboAction *combo;
};


static gboolean
activate_entry(GtkTreeModel *model,
	       GtkTreePath *path, GtkTreeIter *iter,
	       ActivateComboEntryArgs *args)
{
  GtkWidget *widget = NULL;
  gtk_tree_model_get(model, iter, VIEW_WIDGET, &widget, -1);
  g_assert(widget != NULL);

  if (widget == args->view)
    {
      ochusha_combo_action_set_active_iter(args->combo, iter);
      return TRUE;
    }
  return FALSE;
}


static void
activate_combo_entry(GtkListStore *store,
		     OchushaComboAction *combo, GtkWidget *view)
{
  ActivateComboEntryArgs args = { view, combo };
  gtk_tree_model_foreach(GTK_TREE_MODEL(store),
		reinterpret_cast<GtkTreeModelForeachFunc>(activate_entry),
		&args);
}


void
ViewManager::board_switch_page_cb(ModNotebook *notebook, ModNotebookPage *page,
				  guint page_num, ViewManager *view_manager)
{
  GtkWidget *new_page = mod_notebook_get_nth_page(notebook, page_num);

  gtk_action_set_sensitive(view_manager->board_left, page_num > 0);
  gtk_action_set_sensitive(view_manager->board_right,
			   page_num < view_manager->number_of_boards - 1);

  activate_combo_entry(view_manager->board_entries,
		       OCHUSHA_COMBO_ACTION(view_manager->board_combo),
		       new_page);

  view_manager->notify_view_changed(notebook, new_page);
}


void
ViewManager::notify_view_changed(ModNotebook *notebook, GtkWidget *new_page)
{
  if (!gtk_widget_is_focus(GTK_WIDGET(notebook)))
    {
      /* ʤad-hocǷʴǤ⤷ʤȻȤŤ餤 */
      if (GTK_IS_SCROLLED_WINDOW(new_page))
	gtk_widget_grab_focus(gtk_bin_get_child(GTK_BIN(new_page)));
      else
	gtk_widget_grab_focus(new_page);
    }
  follow_focus();

  BulletinBoard *board = get_board(new_page);
  if (board != NULL)
    {
      if (get_current_board_view() == new_page)
	controller.get_board_ui().notify_current_board_view(new_page, board);
      return;
    }

  BBSThread *thread = get_thread(new_page);
  g_return_if_fail(thread != NULL);

  if (get_current_thread_view() == new_page)
    controller.get_thread_ui().notify_current_thread_view(new_page, thread);

  // 1ڥ󥵥ݡȤΤ
  if (get_current_board() == NULL)
    controller.get_board_ui().notify_current_board_view(NULL, NULL);
}


gboolean
ViewManager::board_button_press_event_cb(ModNotebook *notebook,
					 GdkEventButton *event,
					 ViewManager *view_manager)
{
  g_return_val_if_fail(notebook == view_manager->board_book, FALSE);

  if (event->button == 2)
    {
      gint page = mod_notebook_get_mouse_event_page_num(notebook,
					reinterpret_cast<GdkEventAny *>(event));
      if (page != -1)
	{
	  view_manager->close_nth_board_page(page);
	  view_manager->update_board_pane();
	  view_manager->update_board_tab();
	  view_manager->update_board_navigator();
	  return TRUE;
	}
      return FALSE;
    }

  if (event->button == 1)
    {
      if (event->type == GDK_2BUTTON_PRESS)
	{
	  // ֥륯å
	  gint page = mod_notebook_get_mouse_event_page_num(notebook,
					reinterpret_cast<GdkEventAny *>(event));
	  if (page == -1)
	    return FALSE;

	  GtkWidget *view = mod_notebook_get_nth_page(notebook, page);
	  g_return_val_if_fail(view != NULL, FALSE);

	  BulletinBoard *board = get_board(view);
	  if (board != NULL)
	    {
	      view_manager->controller.refresh(board, true);
	      return TRUE;
	    }
	  // 1ڥ󥵥ݡȤΤ
	  BBSThread *thread = get_thread(view);
	  if (thread != NULL)
	    {
	      view_manager->controller.refresh(thread);
	      return TRUE;
	    }
	}
      return FALSE;
    }

  if (event->button != 3)
    return FALSE;

  gint page = mod_notebook_get_mouse_event_page_num(notebook,
					reinterpret_cast<GdkEventAny *>(event));
  if (page == -1)
    return FALSE;

  GtkWidget *view = mod_notebook_get_nth_page(notebook, page);
  g_return_val_if_fail(view != NULL, FALSE);

  if (view_manager->show_popup_menu(view, event))
    return TRUE;

  return FALSE;
}


void
ViewManager::board_start_interactive_search_cb(OchushaTreeSelectorView *view,
					       ViewManager *view_manager)
{
  g_return_if_fail(view_manager->current_focus_widget != NULL);
  BulletinBoard *board = get_board(view_manager->current_focus_widget);
  g_return_if_fail(board != NULL);

  if (view_manager->board_search_response_id == 0)
    view_manager->setup_board_search();
}


void
ViewManager::popup_menu_destroy_cb(gpointer data)
{
  ViewManager *view_manager = static_cast<ViewManager *>(data);
  view_manager->last_event_widget = NULL;
}


gboolean
ViewManager::show_popup_menu(GtkWidget *target, GdkEventButton *event)
{
  const gchar *menu_key = NULL;

  BulletinBoard *board = get_board(target);
  if (board != NULL)
    menu_key = "/BoardTabPopupMenu";

  BBSThread *thread = get_thread(target);
  if (thread != NULL)
    menu_key = "/ThreadTabPopupMenu";

  if (menu_key != NULL)
    {
      last_event_widget = target;
      ochusha_ui_manager_popup_with_data(controller.get_ui_manager(),
					 menu_key, this,
					 popup_menu_destroy_cb,
					 NULL, NULL,
					 event->button,
					 event->time);
      return TRUE;
    }

  last_event_widget = NULL;
  return FALSE;
}


void
ViewManager::close_nth_board_page(gint page)
{
  g_return_if_fail(page >= 0);

  GtkWidget *view = mod_notebook_get_nth_page(board_book, page);
  g_return_if_fail(view != NULL);

  BulletinBoard *board = get_board(view);
  BBSThread *thread = get_thread(view);

  remove_list_entry(board_entries, view);
  if (--number_of_boards == 0 && ochusha_find_focus_widget(view) != NULL)
    {
      GtkWidget *thread_view = get_current_thread_view();
      if (thread_view != NULL)
	{
	  set_current_focus_widget(thread_view);
	  /* ʤad-hocǷʴǤ⤷ʤȻȤŤ餤 */
	  if (GTK_IS_SCROLLED_WINDOW(thread_view))
	    gtk_widget_grab_focus(gtk_bin_get_child(GTK_BIN(thread_view)));
	  else
	    gtk_widget_grab_focus(thread_view);
	}
    }

  mod_notebook_remove_page(board_book, page);

  page = mod_notebook_get_current_page(board_book);
  if (current_focus_widget == view)
    {
      if (page != -1)
	view = mod_notebook_get_nth_page(board_book, page);
      else
	view = NULL;
      set_current_focus_widget(view);
    }

  if (page != -1)
    {
      view = mod_notebook_get_nth_page(board_book, page);
      board = get_board(view);
      if (board != NULL)
	{
	  controller.get_board_ui().notify_current_board_view(view, board);
	  return;
	}

      // 1ڥ󥵥ݡȤΤ
      thread = get_thread(view);
      g_return_if_fail(thread != NULL);
      controller.get_thread_ui().notify_current_thread_view(view, thread);
    }
  else
    {
      if (board != NULL)
	controller.get_board_ui().notify_current_board_view(NULL, NULL);
      else	// 1ڥ󥵥ݡȤΤ
	controller.get_thread_ui().notify_current_thread_view(NULL, NULL);
    }
}


void
ViewManager::thread_switch_page_cb(ModNotebook *notebook,
				   ModNotebookPage *page, guint page_num,
				   ViewManager *view_manager)
{
  GtkWidget *new_page = mod_notebook_get_nth_page(notebook, page_num);

  gtk_action_set_sensitive(view_manager->thread_left, page_num > 0);
  gtk_action_set_sensitive(view_manager->thread_right,
			   page_num < view_manager->number_of_threads - 1);

  activate_combo_entry(view_manager->thread_entries,
		       OCHUSHA_COMBO_ACTION(view_manager->thread_combo),
		       new_page);

  view_manager->notify_view_changed(notebook, new_page);
}


gboolean
ViewManager::thread_button_press_event_cb(ModNotebook *notebook,
					  GdkEventButton *event,
					  ViewManager *view_manager)
{
  g_return_val_if_fail(notebook == view_manager->thread_book, FALSE);

  if (event->button == 2)
    {
      gint page = mod_notebook_get_mouse_event_page_num(notebook,
					reinterpret_cast<GdkEventAny *>(event));
      if (page != -1)
	{
	  view_manager->close_nth_thread_page(page);
	  view_manager->update_thread_pane();
	  view_manager->update_thread_tab();
	  view_manager->update_thread_navigator();
	  return TRUE;
	}
      return FALSE;
    }

  if (event->button == 1)
    {
      if (event->type == GDK_2BUTTON_PRESS)
	{
	  // ֥륯å
	  gint page = mod_notebook_get_mouse_event_page_num(notebook,
					reinterpret_cast<GdkEventAny *>(event));
	  if (page == -1)
	    return FALSE;

	  GtkWidget *view = mod_notebook_get_nth_page(notebook, page);
	  g_return_val_if_fail(view != NULL, FALSE);

	  BBSThread *thread = get_thread(view);
	  if (thread != NULL)
	    {
	      view_manager->controller.refresh(thread);
	      return TRUE;
	    }
	}
      return FALSE;
    }

  if (event->button != 3)
    return FALSE;

  gint page = mod_notebook_get_mouse_event_page_num(notebook,
					reinterpret_cast<GdkEventAny *>(event));
  if (page == -1)
    return FALSE;

  GtkWidget *view = mod_notebook_get_nth_page(notebook, page);
  g_return_val_if_fail(view != NULL, FALSE);

  if (view_manager->show_popup_menu(view, event))
    return TRUE;

  return FALSE;
}


void
ViewManager::thread_start_interactive_search_cb(OchushaHypertextView *view,
						OchushaSearchActionType action,
						ViewManager *view_manager)
{
  g_return_if_fail(view_manager->current_focus_widget != NULL);
  BBSThread *thread = get_thread(view_manager->current_focus_widget);
  g_return_if_fail(thread != NULL);
  /*
   * action˽search barνĴ٤
   */
  if (view_manager->thread_search_response_id == 0)
    view_manager->setup_thread_search();
}


void
ViewManager::close_nth_thread_page(gint page)
{
  g_return_if_fail(page >= 0);

  GtkWidget *view = mod_notebook_get_nth_page(thread_book, page);
  g_return_if_fail(view != NULL);

  remove_list_entry(thread_entries, view);
  if (--number_of_threads == 0 && ochusha_find_focus_widget(view) != NULL)
    {
      GtkWidget *board_view = get_current_board_view();
      if (board_view != NULL)
	{
	  set_current_focus_widget(board_view);
	  /* ʤad-hocǷʴǤ⤷ʤȻȤŤ餤 */
	  if (GTK_IS_SCROLLED_WINDOW(board_view))
	    gtk_widget_grab_focus(gtk_bin_get_child(GTK_BIN(board_view)));
	  else
	    gtk_widget_grab_focus(board_view);
	}
    }

  /* ʤad-hocǷʴġ */
  if (GTK_IS_SCROLLED_WINDOW(view))
    {
      GtkWidget *text_view = gtk_bin_get_child(GTK_BIN(view));
      BBSThreadView *thread_view
	= dynamic_cast<BBSThreadView *>(BBSThreadView::get_view(text_view));
      if (thread_view != NULL)
	thread_view->remember_current_view_position();
    }

  mod_notebook_remove_page(thread_book, page);

  page = mod_notebook_get_current_page(thread_book);
  if (current_focus_widget == view)
    {
      if (page != -1)
	view = mod_notebook_get_nth_page(thread_book, page);
      else
	view = NULL;
      set_current_focus_widget(view);
    }

  if (page != -1)
    {
      view = mod_notebook_get_nth_page(thread_book, page);
      BBSThread *thread = get_thread(view);

      g_return_if_fail(thread != NULL);
      controller.get_thread_ui().notify_current_thread_view(view, thread);
      return;
    }

  controller.get_thread_ui().notify_current_thread_view(NULL, NULL);
}


gboolean
ViewManager::focus_in_event_cb(GtkWidget *widget, GdkEventFocus *event,
			       ViewManager *view_manager)
{
  view_manager->set_current_focus_widget(gtk_widget_get_parent(widget));
  return FALSE;
}


gboolean
ViewManager::enter_notify_event_cb(GtkWidget *widget, GdkEventCrossing *event,
				   ViewManager *view_manager)
{
  if (view_manager->prefs.get_focus_follow_mouse())
    gtk_widget_grab_focus(widget);
  return FALSE;
}


void
ViewManager::append_board_view(BulletinBoard *board,
			       GtkWidget *view, GtkWidget *tab_label,
			       bool in_tab, bool select_page)
{
  ModNotebook *notebook = board_book;

  g_object_set_qdata(G_OBJECT(view), bulletin_board_id, board);

  if (GTK_IS_SCROLLED_WINDOW(view))
    {
      // ʡad-hocǷʴġ
      GtkWidget *widget = gtk_bin_get_child(GTK_BIN(view));
      g_signal_connect(widget, "focus_in_event",
		       reinterpret_cast<GCallback>(focus_in_event_cb), this);
      g_signal_connect(widget, "enter_notify_event",
		       reinterpret_cast<GCallback>(enter_notify_event_cb),
		       this);
      g_signal_connect(widget, "start_interactive_search",
		       reinterpret_cast<GCallback>(board_start_interactive_search_cb),
		       this);
    }

  GtkTreeIter dummy_iter;
  gtk_list_store_insert_with_values(board_entries, &dummy_iter, G_MAXINT32,
				    VIEW_WIDGET, view, -1);

  if (number_of_boards == 0)
    gtk_widget_show(board_box);

  if (number_of_boards == 0 || in_tab)
    {
      mod_notebook_append_page(notebook, view, tab_label);
      mod_notebook_set_tab_reorderable(notebook, view, TRUE);
      mod_notebook_set_tab_label_packing(notebook,
					 view, FALSE, FALSE, GTK_PACK_START);

      if (++number_of_boards > 1 && select_page)
	{
	  gint page = mod_notebook_page_num(notebook, view);
	  g_assert(page != -1);
	  mod_notebook_set_current_page(notebook, page);
	}

      if (number_of_boards == 1)
	controller.get_board_ui().notify_current_board_view(view, board);
    }
  else
    {
      GtkWidget *prev_view;
      gint page = mod_notebook_get_current_page(notebook);
      g_assert(page != -1);

      prev_view = mod_notebook_get_nth_page(notebook, page);
      remove_list_entry(board_entries, prev_view);
      mod_notebook_remove_page(notebook, page);

      if (current_focus_widget == prev_view)
	set_current_focus_widget(view);

      mod_notebook_insert_page(notebook, view, tab_label, page);
      mod_notebook_set_tab_reorderable(notebook, view, TRUE);
      mod_notebook_set_tab_label_packing(notebook,
					 view, FALSE, FALSE, GTK_PACK_START);
      mod_notebook_set_current_page(notebook, page);
    }

  if (current_focus_widget == NULL)
    {
      gtk_container_set_focus_child(GTK_CONTAINER(notebook), view);
      set_current_focus_widget(view);
    }

  update_board_pane();
  update_board_tab();
  update_board_navigator();
}


void
ViewManager::append_thread_view(BBSThread *thread,
				GtkWidget *view, GtkWidget *tab_label,
				bool in_tab, bool select_page)
{
  ModNotebook *notebook = thread_book;

  g_object_set_qdata(G_OBJECT(view), bbs_thread_id, thread);

  if (GTK_IS_SCROLLED_WINDOW(view))
    {
      // ʡad-hocǷʴġ
      GtkWidget *widget = gtk_bin_get_child(GTK_BIN(view));
      g_signal_connect(widget, "focus_in_event",
		       reinterpret_cast<GCallback>(focus_in_event_cb), this);
      g_signal_connect(widget, "enter_notify_event",
		       reinterpret_cast<GCallback>(enter_notify_event_cb),
		       this);
      g_signal_connect(widget, "start_interactive_search",
		       reinterpret_cast<GCallback>(thread_start_interactive_search_cb),
		       this);
    }

  GtkTreeIter dummy_iter;
  gtk_list_store_insert_with_values(thread_entries, &dummy_iter, G_MAXINT32,
				    VIEW_WIDGET, view, -1);

  if (number_of_threads == 0)
    gtk_widget_show(thread_pane);

  if (number_of_threads == 0 || in_tab)
    {
      mod_notebook_append_page(notebook, view, tab_label);
      mod_notebook_set_tab_reorderable(notebook, view, TRUE);
      mod_notebook_set_tab_label_packing(notebook,
					 view, FALSE, FALSE, GTK_PACK_START);

      if (++number_of_threads > 1 && select_page)
	{
	  gint page = mod_notebook_page_num(notebook, view);
	  g_assert(page != -1);
	  mod_notebook_set_current_page(notebook, page);
	}
      if (number_of_threads == 1)
	controller.get_thread_ui().notify_current_thread_view(view, thread);
    }
  else
    {
      GtkWidget *prev_view;
      gint page = mod_notebook_get_current_page(notebook);
      g_assert(page != -1);

      prev_view = mod_notebook_get_nth_page(notebook, page);
      remove_list_entry(thread_entries, prev_view);
      mod_notebook_remove_page(notebook, page);

      if (current_focus_widget == prev_view)
	set_current_focus_widget(view);

      mod_notebook_insert_page(notebook, view, tab_label, page);
      mod_notebook_set_tab_reorderable(notebook, view, TRUE);
      mod_notebook_set_tab_label_packing(notebook,
					 view, FALSE, FALSE, GTK_PACK_START);
      mod_notebook_set_current_page(notebook, page);
    }

  if (current_focus_widget == NULL)
    {
      gtk_container_set_focus_child(GTK_CONTAINER(notebook), view);
      set_current_focus_widget(view);
    }

  update_thread_pane();
  update_thread_tab();
  update_thread_navigator();
}


struct FindBoardArgs
{
  const BulletinBoard *board;
  GtkWidget *view;
};


static gboolean
compare_board_entry(GtkTreeModel *model,
		    GtkTreePath *path, GtkTreeIter *iter,
		    FindBoardArgs *args)
{
  GtkWidget *widget = NULL;
  gtk_tree_model_get(model, iter, VIEW_WIDGET, &widget, -1);
  g_assert(widget != NULL);

  if (ViewManager::get_board(widget) == args->board)
    {
      args->view = widget;
      return TRUE;
    }
  return FALSE;
}


static GtkWidget *
find_board_entry(GtkListStore *store, const BulletinBoard *board)
{
  FindBoardArgs args = { board, NULL };
  gtk_tree_model_foreach(GTK_TREE_MODEL(store),
		reinterpret_cast<GtkTreeModelForeachFunc>(compare_board_entry),
		&args);
  return args.view;
}


GtkWidget *
ViewManager::find_board_view(const BulletinBoard *board, bool select_page)
{
  GtkWidget *view = find_board_entry(board_entries, board);
  if (select_page && view != NULL)
    {
      gint page = mod_notebook_page_num(board_book, view);
      g_assert(page != -1);
      mod_notebook_set_current_page(board_book, page);
    }
  return view;
}


GtkWidget *
ViewManager::get_board_tab_label(GtkWidget *view)
{
  return mod_notebook_get_tab_label(board_book, view);
}


struct FindThreadArgs
{
  const BBSThread *thread;
  GtkWidget *view;
};


static gboolean
compare_thread_entry(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
		     FindThreadArgs *args)
{
  GtkWidget *widget = NULL;
  gtk_tree_model_get(model, iter, VIEW_WIDGET, &widget, -1);
  g_assert(widget != NULL);

  if (ViewManager::get_thread(widget) == args->thread)
    {
      args->view = widget;
      return TRUE;
    }
  return FALSE;
}


static GtkWidget *
find_thread_entry(GtkListStore *store, const BBSThread *thread)
{
  FindThreadArgs args = { thread, NULL };
  gtk_tree_model_foreach(GTK_TREE_MODEL(store),
		reinterpret_cast<GtkTreeModelForeachFunc>(compare_thread_entry),
		&args);
  return args.view;
}


GtkWidget *
ViewManager::find_thread_view(const BBSThread *thread, bool select_page)
{
  GtkWidget *view = find_thread_entry(thread_entries, thread);
  if (select_page && view != NULL)
    {
      gint page = mod_notebook_page_num(thread_book, view);
      g_assert(page != -1);
      mod_notebook_set_current_page(thread_book, page);
    }
  return view;
}


GtkWidget *
ViewManager::get_thread_tab_label(GtkWidget *view)
{
  return mod_notebook_get_tab_label(thread_book, view);
}


GtkWidget *
ViewManager::get_tab_label(const BBSThread *thread)
{
  GtkWidget *view = find_thread_entry(thread_entries, thread);
  if (view != NULL)
    return get_thread_tab_label(view);

  return NULL;
}


GtkWidget *
ViewManager::get_tab_label(const BulletinBoard *board)
{
  GtkWidget *view = find_board_entry(board_entries, board);
  if (view != NULL)
    return get_board_tab_label(view);

  return NULL;
}


static GtkWindow *
get_toplevel_window(GtkWidget *widget)
{
  g_assert(widget != NULL);
  GtkWidget *toplevel = gtk_widget_get_toplevel(widget);
  if (GTK_WIDGET_TOPLEVEL(toplevel))
    return GTK_WINDOW(toplevel);
  return NULL;
}


void
ViewManager::set_current_focus_widget(GtkWidget *widget)
{
  gchar title_buf[1024];

  if (current_focus_widget == widget)
    return;

  current_focus_widget = widget;
  GtkWindow *toplevel = get_toplevel_window(contents_pane);
  if (toplevel == NULL)
    return;

  if (widget == NULL)
    {
      gtk_window_set_title(toplevel, _("Ochusha"));
      if (current_url_entry != NULL)
	gtk_entry_set_text(current_url_entry, "");
      return;
    }

  char *url_copy = NULL;
  const char *url = NULL;
  BBSThread *thread
    = static_cast<BBSThread *>(g_object_get_qdata(G_OBJECT(widget),
						  bbs_thread_id));
  if (thread != NULL)
    {
      BulletinBoard *board = thread->get_board();
      snprintf(title_buf, 1024, _("%s @ %s board - Ochusha"),
	       thread->get_title(), board->get_name());
      url_copy = thread->get_url();
      url = url_copy;
    }
  else
    {
      BulletinBoard *board
	= static_cast<BulletinBoard *>(g_object_get_qdata(G_OBJECT(widget),
							  bulletin_board_id));
      g_assert(board != NULL);
      snprintf(title_buf, 1024, _("%s board - Ochusha"), board->get_name());
      url = board->get_base_url();
    }
  gtk_window_set_title(toplevel, title_buf);

  if (current_url_entry != NULL)
    gtk_entry_set_text(current_url_entry, url);

  if (url_copy != NULL)
    free(url_copy);
}


struct ForeachBoardArgs
{
  ViewManager::ForeachBoardFunc callback;
  void *callback_data;
};


static void
foreach_board_helper(GtkWidget *widget, gpointer data)
{
  BulletinBoard *board = ViewManager::get_board(widget);
  if (board != NULL)
    {
      ForeachBoardArgs *args = static_cast<ForeachBoardArgs *>(data);
      (*args->callback)(widget, board, args->callback_data);
    }
}


void
ViewManager::foreach_board(ForeachBoardFunc callback, void *callback_data)
{
  ForeachBoardArgs args = { callback, callback_data };

  gtk_container_foreach(GTK_CONTAINER(board_book), foreach_board_helper,
			&args);
}


struct ForeachThreadArgs
{
  ViewManager::ForeachThreadFunc callback;
  void *callback_data;
};


static void
foreach_thread_helper(GtkWidget *widget, gpointer data)
{
  BBSThread *thread = ViewManager::get_thread(widget);
  if (thread != NULL)
    {
      ForeachThreadArgs *args = static_cast<ForeachThreadArgs *>(data);
      (*args->callback)(widget, thread, args->callback_data);
    }
}


void
ViewManager::foreach_thread(ForeachThreadFunc callback, void *callback_data)
{
  ForeachThreadArgs args = { callback, callback_data };

  gtk_container_foreach(GTK_CONTAINER(thread_book), foreach_thread_helper,
			&args);
}


BulletinBoard *
ViewManager::get_board(GtkWidget *view)
{
  return static_cast<BulletinBoard *>(g_object_get_qdata(G_OBJECT(view),
							 bulletin_board_id));
}


BBSThread *
ViewManager::get_thread(GtkWidget *view)
{
  return static_cast<BBSThread *>(g_object_get_qdata(G_OBJECT(view),
						     bbs_thread_id));
}

}
